#!/usr/local/bin/perl
#
#   CGIProxy 1.5.1
#
#   nph-proxy.cgi-- CGIProxy 1.5.1: an proxy in the form of a CGI script.
#     Retrieves the resource at any HTTP or FTP URL, updating embedded URLs
#     in HTML resources to point back through this script.  By default, no
#     user info is sent to the server.  Options include text-only proxying
#     to save bandwidth, cookie support, ad-filtering, script removal,
#     user-defined encoding of the target URL, and more.  Runs in Perl 4 or 5.
#
#   Copyright (C) 1996, 1998-2002 by James Marshall, james@jmarshall.com
#   All rights reserved.
#
#   For the latest, see http://www.jmarshall.com/tools/cgiproxy/
#
#
#   IMPORTANT NOTE ABOUT ANONYMOUS BROWSING:
#     CGIProxy was originally made for indirect browsing more than
#       anonymity, but since people are using it for anonymity, I've tried
#       to make it as anonymous as possible.  Suggestions welcome.  For best
#       anonymity, browse with JavaScript turned off.
#     Anonymity is pretty good, but may not be bulletproof.  For example,
#       if even a single JavaScript statement can be run, your anonymity can
#       be compromised.  I've tried to remove JS from every place it can
#       exist, but please tell me if I missed any.  Also, browser plugins or
#       other executable extensions may be able to reveal you to a server.
#     If you find any way your anonymity can be compromised even with scripts
#       turned off, please let me know.
#
#
#   CONFIGURATION:
#
#     None required in most situations.  On some servers, these might be
#       required (all in the "user configuration" section):
#       . If this is running on an SSL server, set $RUNNING_ON_SSL_SERVER=1.
#       . If this is running on Windows, set $RUNNING_ON_WINDOWS=1 to work
#           around a couple of Windows and IIS bugs.
#       . If you're using another HTTP proxy, set $ENV{'http_proxy'} and
#           $ENV{'no_proxy'}.  If that proxy uses authentication, set
#           $PROXY_AUTH.
#       . To make this run in Perl 4, follow the instructions around the
#           "use Socket" line.
#
#     Options include:
#       . Set $TEXT_ONLY, $REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS,
#           $HIDE_REFERER, and $INSERT_ENTRY_FORM as desired.  Set
#           $REMOVE_SCRIPTS if anonymity is important.
#       . To let the user choose all of those settings (except $TEXT_ONLY),
#           set $ALLOW_USER_CONFIG=1.
#       . To change the encoding format of the URL, modify the
#           proxy_encode() and proxy_decode() routines.  The default
#           routines are suitable for simple PATH_INFO compliance.
#       . To encode cookies, modify the cookie_encode() and cookie_decode()
#           routines.
#       . You can restrict which servers this proxy will access, with
#           @ALLOWED_SERVERS and @BANNED_SERVERS.
#       . Similarly, you can specify allowed and denied server lists for
#           both cookies and scripts.
#       . For security, you can ban access to private IP ranges, with
#           @BANNED_NETWORKS.
#       . If filtering ads, you can customize this with a few settings.
#       . To insert your own block of HTML into each page, set $INSERT_HTML
#           or $INSERT_FILE.
#       . As a last resort, if you really can't run this script as NPH,
#           you can try to run it as non-NPH by setting $NOT_RUNNING_AS_NPH=1.
#           BUT, read the notes and warnings above that line.  Caveat surfor.
#       . For crude load-balancing among a set of proxies, set @PROXY_GROUP.
#       . Other minor config is possible; see the user configuration section.
#       . If heavy use of this proxy puts a load on your server, see the
#           "NOTES ON PERFORMANCE" section below.
#
#     For more info, read the comments regarding any config options you set.
#
#     This script MUST be installed as a non-parsed header (NPH) script.
#       In Apache and many other servers, this is done by simply starting the
#       filename with "nph-".  It MAY be possible to fake it as a non-NPH
#       script, MOST of the time, by using the $NOT_RUNNING_AS_NPH feature.
#       This is not advised.  See the comments by that option for warnings.
#
#
#   TO USE:
#     Start a browsing session by calling the script with no parameters.
#       You can bookmark pages you browse to through the proxy, or link to
#       the URLs that are generated.
#
#
#   NOTES ON PERFORMANCE:
#     Unfortunately, this has gotten slower through the versions, mostly
#       because of optional new features.  Configured equally, version 1.3
#       takes 25% longer to run than 1.0 or 1.1 (based on *cough* highly
#       abbreviated testing).  Compiling takes about 50% longer.
#     Leaving $REMOVE_SCRIPTS=1 adds 25-50% to the running time.
#     Leaving $FASTER_HTML_LESS_PRIVACY=0 increases the running time by 10%.
#       (Previous versions had a privacy hole which this plugs.)
#     For a gain in compile speed, follow the instructions near the
#       "use Socket" line-- that module takes quite a while to load.
#     Remember that we're talking about tenths of a second here.  Most of
#       the delay experienced by the user is from waiting on two network
#       connections.  These performance issues only matter if your server
#       CPU is getting overloaded.  Also, these only matter when retrieving
#       HTML, because it's the HTML modification that takes all the time.
#     If you can, use mod_perl.  Starting with version 1.3.1, this should
#       work under mod_perl, but in this case you need at least Perl 5.004.
#       If you use mod_perl, be careful to install this as an NPH script,
#       i.e. set the "PerlSendHeader Off" configuration directive.  For more
#       info, see the mod_perl documentation.
#     If you use mod_perl and modify this script, see the note near the
#       "reset 'a-z'" line below, regarding UPPER_CASE and lower_case
#       variables.
#
#
#   TO DO:
#     What I want to hear about:
#       . Any HTML tags not being converted here.
#       . Any method of introducing JavaScript or other script, that's not
#           being filtered out here.
#       . Any script MIME types other than those already in @SCRIPT_MIME_TYPES.
#       . Any MIME types other than text/html that have links that need to
#           be converted.
#
#     plug any other script holes (e.g. MSIE-proprietary, other MIME types?)
#     This could use cleaner URL-encoding all over ($base_url, etc.)
#     more error checking?
#     find a simple encryption technique for proxy_encode()
#     support more protocols, like SSL or gopher
#     For ad filtering, add option to disable images from servers other than
#       that of the containing HTML page?  Is it worth it?
#
#
#   BUGS:
#     Anonymity is not perfect.  In particular, there may be some remaining
#       JavaScript holes.
#     URLs generated by JavaScript or similar mechanisms won't be re-proxy'ed
#       correctly.  JavaScript in general may not work as expected.
#     Since ALL of your cookies are sent to this script (which then chooses
#       the relevant ones), some cookies could conceivably be dropped if
#       you accumulate a whole lot.  I haven't seen this happen yet.
#     The regex that converts URLs is not perfect-- it always matches what's
#       needed, but sometimes it matches more.  A perfect regex would be
#       much slower.  See the notes above that block for more details.
#     When converted, URLs are not HTML-unescaped then re-escaped as they
#       should be.  To fix this would cost a lot of CPU time and would seldom
#       matter, but it's easy enough to do if there's demand.
#     Comments and script contents aren't parsed correctly, so sometimes
#       URLs within those will be mistakenly converted.  In reality, this
#       sometimes approximates what you want anyway.
#
#
#   I first wrote this in 1996 as an experiment to allow indirect browsing.
#     The original seed was a program I wrote for Rich Morin's article
#     in the June 1996 issue of Unix Review, online at
#     http://www.cfcl.com/tin/P/199606.shtml.
#
#   Confession: I didn't originally write this with the spec for HTTP
#     proxies in mind, and there are probably some violations of the protocol
#     (at least for proxies).  This whole thing is one big violation of the
#     proxy model anyway, so I hereby rationalize that the spec can be widely
#     interpreted here.  If there is demand, I can make it more conformant.
#
#--------------------------------------------------------------------------
#
# Here are all global variables used in this script.  Some could be
#   rewritten as local variables, possibly with a performance loss.  With
#   these declared and all local() statements converted to my(), you can
#   (almost) compile with "use strict", which is always a good idea when
#   programming in Perl 5.
#
# use vars qw(
#    $TEXT_ONLY
#    $REMOVE_COOKIES  $REMOVE_SCRIPTS  $FILTER_ADS  $INSERT_ENTRY_FORM
#    $HIDE_REFERER  $ALLOW_USER_CONFIG
#    $NO_COOKIE_WITH_IMAGE  @ALLOWED_COOKIE_SERVERS  @BANNED_COOKIE_SERVERS
#    @ALLOWED_SCRIPT_SERVERS  @BANNED_SCRIPT_SERVERS
#    @ALLOWED_SERVERS  @BANNED_SERVERS  @BANNED_NETWORKS  @BANNED_NETWORK_ADDRS
#    @BANNED_IMAGE_URL_PATTERNS  $RETURN_EMPTY_GIF
#    $INSERT_HTML  $INSERT_FILE  $INSERTION_FRAME_HEIGHT  $ANONYMIZE_INSERTION
#    $FORM_AFTER_INSERTION  $DOING_INSERT
#    $RUNNING_ON_WINDOWS  $RUNNING_ON_SSL_SERVER
#    $NOT_RUNNING_AS_NPH  $NO_BROWSE_THROUGH_SELF  $HTTP_1_X  $HTTP_VERSION
#    $USER_AGENT  $SHOW_FTP_WELCOME  $MAX_REQUEST_SIZE  $PROXY_AUTH
#    $USE_POST_ON_START  $FASTER_HTML_LESS_PRIVACY  $REMOVE_TITLES
#    $NO_LINK_TO_START  @PROXY_GROUP
#    @SCRIPT_MIME_TYPES  $SCRIPT_TYPE_REGEX  $PROXIFY_SCRIPTS
#    @OTHER_TYPES_TO_REGISTER  @ALL_TYPES  %MIME_TYPE_ID
#    @TYPES_TO_HANDLE  $TYPES_TO_HANDLE_REGEX  $NON_TEXT_EXTENSIONS
#    $AF_INET  $SOCK_STREAM  $THIS_HOST
#    $PROXY_VERSION  $FLAGS  @FLAGS  $encoded_URL  $env_accept
#    $images_are_banned_here  $scripts_are_banned_here  $cookies_are_banned_here
#    $URL  $scheme  $authority  $path  $host  $port  $username  $password
#    $cookie_to_server  %auth
#    $script_url  $this_url  $this_url_inframe  $this_url_noframe
#    $is_in_frame  $expected_type
#    $base_url  $base_scheme  $base_host  $base_path
#    $default_style_type  $default_script_type
#    $status  $headers  $body  @body  @welcome  $is_html  $response_sent
#    $tag_name  $att_init
#    $dummy ) ;
#
#--------------------------------------------------------------------------

#--------------------------------------------------------------------------
#    user configuration
#--------------------------------------------------------------------------

# If set, then proxy traffic will be restricted to text data only, to save
#   bandwidth (though it can still be circumvented with uuencode, etc.).

$TEXT_ONLY= 0 ;      # set to 1 to allow only text data, 0 to allow all


# If set, then prevent all cookies from passing through the proxy.  To allow
#   cookies from some servers, set this to 0 and see @ALLOWED_COOKIE_SERVERS
#   and @BANNED_COOKIE_SERVERS below.  You can also prevent cookies with
#   images by setting $NO_COOKIE_WITH_IMAGE below.
# Note that this only affects cookies from the target server.   The proxy
#   script sends its own cookies for other reasons too, like to support
#   authentication.  This flag does not stop these cookies from being sent.

$REMOVE_COOKIES= 0 ;


# If set, then remove as much scripting as possible.  On the HTTP level:
#   . prevent transmission of script MIME types (which only works if the server
#       marks them as such, so a malicious server could get around this, but
#       then the browser probably wouldn't execute the script).
#   . remove Link: headers that link to a resource of a script MIME type.
# Within HTML resources:
#   . remove <script>...</script> .
#   . remove intrinsic event attributes from tags, i.e. attributes whose names
#       begin with "on".
#   . remove <style>...</style> where "type" attribute is a script MIME type.
#   . remove various HTML tags that appear to link to a script MIME type.
#   . remove script macros (aka Netscape-specific "JavaScript entities"),
#       i.e. any attributes containing the string "&{" .
#   . remove "JavaScript conditional comments".
#   . remove MSIE-specific "dynamic properties".
# To allow scripts from some sites but not from others, set this to 0 and
#   see @ALLOWED_SCRIPT_SERVERS and @BANNED_SCRIPT_SERVERS below.
# The list of script MIME types is surely incomplete.  See @SCRIPT_MIME_TYPES
#   below.
# I do NOT know for certain that this removes all script content!  It removes
#   all that I know of, but I don't have a definitive list of places scripts
#   can exist.  If you do, please send it to me.  EVEN RUNNING A SINGLE
#   JAVASCRIPT STATEMENT CAN COMPROMISE YOUR ANONYMITY!  Just so you know.
# Richard Smith has a good test site for anonymizing proxies, at
#   http://users.rcn.com/rms2000/anon/test.htm
# Note that turning this on removes most popup ads!  :)

$REMOVE_SCRIPTS= 1 ;


# If set, then filter out images that match one of @BANNED_IMAGE_URL_PATTERNS,
#   below.  Also removes cookies attached to images, by setting
#   $NO_COOKIE_WITH_IMAGE.
# To remove most popup advertisements, also set $REMOVE_SCRIPTS=1 above.

$FILTER_ADS= 0 ;


# If set, then don't send a Referer: [sic] header with each request
#   (i.e. something that tells the server which page you're coming from
#   that linked to it).  This is a minor privacy issue, but a few sites
#   won't send you pages or images if the Referer: is not what they're
#   expecting.  If a page is loading without images or a link seem to be
#   refused, then try turning this off, and a correct Referer: header will
#   be sent.
# This is only a problem in a VERY small percentage of sites, so few that
#   I'm kinda hesitant to put this in the entry form.  Other arrangements
#   have their own problems, though.

$HIDE_REFERER= 1 ;


# If set, insert a compact version of the URL entry form at the top of each
#   page.  This will also display the URL currently being viewed.
# When viewing a page with frames, then a new top frame is created and the
#   insertion goes there.
# If you want to customize the appearance of the form, modify the routine
#   mini_start_form() near the end of the script.
# If you want to insert something other than this form, see $INSERT_HTML and
#   $INSERT_FILE below.
# Users should realize that options changed via the form only take affect when
#   the form is submitted by entering a new URL or pressing the "Go" button.
#   Selecting an option, then following a link on the page, will not cause
#   the option to take effect.
# Users should also realize that anything inserted into a page may throw
#   off any precise layout.  The insertion will also be subject to
#   background colors and images, and any other page-wide settings.

$INSERT_ENTRY_FORM= 1 ;


# If set, then allow the user to control $REMOVE_COOKIES, $REMOVE_SCRIPTS,
#   $FILTER_ADS, $HIDE_REFERER, and $INSERT_ENTRY_FORM.  Note that they
#   can't fine-tune any related options, such as the various @ALLOWED... and
#   @BANNED... lists.

$ALLOW_USER_CONFIG= 1 ;



# Create your own proxy_encode() and proxy_decode() to tranform the target
#   URL to and from the format that will be stored in PATH_INFO.  The encoded
#   form should only contain characters that are legal in PATH_INFO.  This
#   varies by server, but using only printable chars, no "?" or "#", and no
#   two adjacent slashes ("//") works on most servers.  Don't let PATH_INFO
#   contain the strings "./", "/.", "../", or "/..", or else it may get
#   compressed like a pathname somewhere.  Try not to make the resulting
#   string too long, either.
# Of course, proxy_decode() must exactly undo whatever proxy_encode() does.
# Make proxy_encode() as fast as possible-- it's a major bottleneck for the
#   whole program.
# Because of the simplified absolute URL resolution in full_url(), there may
#   be ".." segments in the default encoding here, notably in the first path
#   segment.  Normally, that's just an HTML mistake, but please tell me if
#   you see any privacy exploit with it.
# Note that a few sites have embedded applications (like applets or Shockwave)
#   that expect to access URLs relative to the page's URL.  This means they
#   may not work if the encoded target URL can't be treated like a base URL,
#   e.g. that it can't be appended with something like "../data/foo.data"
#   to get that expected data file.  In such cases, the default encoding below
#   should let these sites work fine, as should any other encoding that can
#   support URLs relative to it.

sub proxy_encode {
    local($URL)= @_ ;
    $URL=~ s#^([\w+.-]+)://#$1/# ;                 # http://xxx -> http/xxx
#    $URL=~ s/(.)/ sprintf('%02x',ord($1)) /ge ;   # each char -> 2-hex
#    $URL=~ tr/a-zA-Z/n-za-mN-ZA-M/ ;              # rot-13
    return $URL ;
}

sub proxy_decode {
    local($enc_URL)= @_ ;
#    $enc_URL=~ tr/a-zA-Z/n-za-mN-ZA-M/ ;        # rot-13
#    $enc_URL=~ s/([0-9A-Fa-f]{2})/ sprintf("%c",hex($1)) /ge ;
    $enc_URL=~ s#^([\w+.-]+)/#$1://# ;           # http/xxx -> http://xxx
    return $enc_URL ;
}


# Encode cookies before they're sent back to the user.
# The return value must only contain characters that are legal in cookie
#   names and values, i.e. only printable characters, and no ";", ",", "=",
#   or white space.
# cookie_encode() is called twice for each cookie: once to encode the cookie
#   name, and once to encode the cookie value.  The two are then joined with
#   "=" and sent to the user.
# cookie_decode() must exactly undo whatever cookie_encode() does.
# Also, cookie_encode() must always encode a given input string into the
#   same output string.  This is because browsers need the cookie name to
#   identify and manage a cookie, so the name must be consistent.
# This is not a bottleneck like proxy_encode() is, so speed is not critical.

sub cookie_encode {
    local($cookie)= @_ ;
#    $cookie=~ s/(.)/ sprintf('%02x',ord($1)) /ge ;   # each char -> 2-hex
#    $cookie=~ tr/a-zA-Z/n-za-mN-ZA-M/ ;              # rot-13
    $cookie=~ s/(\W)/ '%' . sprintf('%02x',ord($1)) /ge ; # simple URL-encoding
    return $cookie ;
}

sub cookie_decode {
    local($enc_cookie)= @_ ;
    $enc_cookie=~ s/%([\da-fA-F]{2})/ pack('C', hex($1)) /ge ;  # URL-decode
#    $enc_cookie=~ tr/a-zA-Z/n-za-mN-ZA-M/ ;          # rot-13
#    $enc_cookie=~ s/([0-9A-Fa-f]{2})/ sprintf("%c",hex($1)) /ge ;
    return $enc_cookie ;
}



# Use these to restrict which servers a user can visit through this proxy.
# They work like @ALLOWED_COOKIE_SERVERS and @BANNED_COOKIE_SERVERS below
#   (see those notes for more details):  They're a list of patterns; if
#   @ALLOWED_SERVERS is set, then only allow access to those servers; always
#   prevent access to any servers in @BANNED_SERVERS.
# Again, these are patterns, not literal strings-- if you want to specify
#   one server like "www.example.com", use "^www\.example\.com$", not
#   the pattern "www.example.com"; otherwise, you may match more than that
#   one server.
@ALLOWED_SERVERS= () ;
@BANNED_SERVERS= () ;


# If @BANNED_NETWORKS is set, then forbid access to these hosts or networks.
# This is done by IP address, not name, so it provides more certain security
#   than @BANNED_SERVERS above.
# Specify each element as a decimal IP address-- all four integers for a host,
#   or one to three integers for a network.  For example, '127.0.0.1' bans
#   access to the local host, and '192.168' bans access to all IP addresses
#   in the 192.168 network.  Sorry, no banning yet for subnets other than
#   8, 16, or 24 bits.
# IF YOU'RE RUNNING THIS ON OR INSIDE A FIREWALL, THIS SETTING IS STRONGLY
#   RECOMMENDED!!  In particular, you should ban access to other machines
#   inside the firewall that the firewall machine itself may have access to.
# In addition to the recommended defaults below, add all IP addresses of your
#   server machine if you want to protect it like this.
# After you set this, YOU SHOULD TEST to verify that the proxy can't access
#   the IP addresses you're banning!
# This feature is simple now but will be more complete in future releases.
@BANNED_NETWORKS= ('127.0.0.1', '192.168', '10') ;


# Settings to fine-tune cookie behavior, if $REMOVE_COOKIES is not set above.
# Note that @ALLOWED_COOKIE_SERVERS and @BANNED_COOKIE_SERVERS use Perl
#   regular expressions (aka patterns or regexes), not literal host names.
#   To turn a hostname into a pattern, replace every "." with "\.", add "^"
#   to the beginning, and add "$" to the end.  For example, "www.example.com"
#   becomes "^www\.example\.com$".  To match *every* host ending in
#   something, leave out the "^".  For example, "\.example\.com$" matches
#   every host ending in ".example.com".

# Set this to reject cookies attached to images
$NO_COOKIE_WITH_IMAGE= 1 ;

# If non-empty, only allow cookies from servers matching one of these patterns.
# Comment this out to allow all cookies (subject to @BANNED_COOKIE_SERVERS).
#@ALLOWED_COOKIE_SERVERS= ('\bslashdot\.org$') ;

# Reject cookies from servers matching these patterns.
@BANNED_COOKIE_SERVERS= (
    '\.doubleclick\.net$',
    '\.preferences\.com$',
    '\.imgis\.com$',
    '\.adforce\.com$',
    '\.focalink\.com$',
    '\.flycast\.com$',
    '\.go\.com$',
    '\.avenuea\.com$',
    '\.linkexchange\.com$',
    '\.pathfinder\.com$',
    '\.burstnet\.com$',
    '\btripod\.com$',
    '\bgeocities\.yahoo\.com$',
    '\.mediaplex\.com$',
    ) ;


# If $REMOVE_SCRIPTS is set, then all scripts will be removed.  Otherwise,
#   set @ALLOWED_SCRIPT_SERVERS and @BANNED_SCRIPT_SERVERS to select which
#   sites you'll allow scripts from.  They work like @ALLOWED_COOKIE_SERVERS
#   and @BANNED_COOKIE_SERVERS above:  They're a list of patterns; always
#   reject scripts from @BANNED_SCRIPT_SERVERS, then if @ALLOWED_SCRIPT_SERVERS
#   is set, restrict scripts to those servers.  Empty lists will allow all
#   scripts to pass (unless $REMOVE_SCRIPTS is set above).
@ALLOWED_SCRIPT_SERVERS= () ;
@BANNED_SCRIPT_SERVERS= () ;



# Various options to help filter ads and stop cookie-based privacy invasion.
# These are only effective if $FILTER_ADS is set above.
# The default settings are OK, but the sample pattern lists are very simple
#   and could be expanded.  Improvements welcome!
# The champs of ad-filtering are JunkBusters, at http://www.junkbusters.com.
#   If ad-filtering is your primary motive, go there, dey da best.
# @BANNED_IMAGE_URL_PATTERNS uses Perl patterns.  If an image's URL
#   matches one of the patterns, it will not be downloaded (typically for
#   ad-filtering).  For more information on Perl regular expressions, see
#   the Perl documentation.
# Note that most popup ads can be removed by setting $REMOVE_SCRIPTS=1, above.

# If set, replace banned images with 1x1 transparent GIF
$RETURN_EMPTY_GIF= 1 ;

# Reject images whose URL matches any of these patterns.  This is just a
#   sample list; add more depending on which sites you visit.
@BANNED_IMAGE_URL_PATTERNS= (
    '\b[a-z]\.doubleclick\.net(:\d*)?/',
    '\.imgis\.com\b',
    '\.adforce\.com\b',
    '\.avenuea\.com\b',
    '\.go\.com(:\d*)?/ad/',
    '\.eimg\.com\b',
    '\bexcite\.netscape\.com(:\d*)?/.*/promo/',
    '/excitenetscapepromos/',
    '\.yimg\.com(:\d*)?.*/promo/',
    '\bus\.yimg\.com/[a-z]/(\w\w)/\1',
    '\bpromotions\.yahoo\.com(:\d*)?/promotions/',
    '\bcnn\.com(:\d*)?/ads/',
    'ads\.msn\.com\b',
    '\blinkexchange\.com\b',
    '\badknowledge\.com\b',
    '/SmartBanner/',
    '\bdeja\.com/ads/',
    '\bimage\.pathfinder\.com/sponsors',
    'ads\.tripod\.com',
    ) ;



# If either $INSERT_HTML or $INSERT_FILE is set, then that HTML text or the
#   contents of that named file (respectively) will be inserted into any
#   downloaded HTML page.  $INSERT_HTML takes precedence over $INSERT_FILE.
# When viewing a page with frames, then a new top frame is created and the
#   insertions go there.  (But note that if you set $ANONYMIZE_INSERTION
#   below, these insertions will not go in that top frame.)
# The frame avoidance means that it's fairly easy for a surfer to bypass
#   this, by pretending in effect to be in a frame.  There's not much we
#   can do about that, since a page is retrieved the same way regardless of
#   whether it's in a frame.  This script uses a parameter in the URL to
#   communicate to itself between calls, but the user can merely change
#   that URL to make the script think it's retrieving a page for a frame.
# If you're using $INSERT_FILE and mod_perl, do NOT set $INSERT_HTML here,
#   not even to ''.  The way it's set up, the file in $INSERT_FILE will
#   only have to be read once*, and it's stored in $INSERT_HTML.  If you
#   clear the value of $INSERT_HTML here, the program will have to read the
#   file every time it's run-- very inefficient.
# [*Well, once for every new mod_perl context, which may be more than once
#   but is less than every time the script is run.]
# As with $INSERT_ENTRY_FORM, note that any insertion may throw off any
#   precise layout, and the insertion is subject to background colors and
#   other page-wide settings.

#$INSERT_HTML= "<h1>This is an inserted header</h1><hr>" ;
#$INSERT_FILE= 'insert_file_name' ;


# If your insertion has links that you want anonymized along with the rest
#   of the downloaded HTML, then set this to 1.  Otherwise leave it at 0.
# NOTE:  Currently, if this is set, then insertions done with $INSERT_HTML or
#   $INSERT_FILE will not show up when viewing a page with frames.  This
#   will likely change in future versions.  In the meantime, if you want your
#   insertions to show up in framed pages, then leave this off.
$ANONYMIZE_INSERTION= 0 ;

# If there's both a URL entry form and an insertion via $INSERT_HTML or
#   $INSERT_FILE on the same page, the entry form normally goes at the top.
#   Set this to put it after the other insertion.
$FORM_AFTER_INSERTION= 0 ;


# If the insertion is put in a top frame, then this is how many pixels high
#   the frame is.  If the default of 80 pixels is too big or too small for
#   your insertion, change this.  You can use percentage of screen height if
#   you prefer, e.g. "20%".  (Unfortunately, you can't just tell the browser
#   to "make it as high as it needs to be", but at least the frame will be
#   resizable by the user.)
# This affects insertions by $INSERT_ENTRY_FORM, $INSERT_HTML, and $INSERT_FILE.
$INSERTION_FRAME_HEIGHT= 80 ;



# Depending on whether you're using Perl 5 or Perl 4 (or want a slight speed
#   advantage), use ONE of the following two blocks.

# Use this for Perl 5 (slows it down a little).
# You can also remove the "local($1,$2)" line in full_url() for slight speedup.
use Socket qw(AF_INET SOCK_STREAM) ;
$AF_INET= AF_INET ; $SOCK_STREAM= SOCK_STREAM ;

# Use this line for Perl 4, or to get a little more speed.  The values of
#   $AF_INET and $SOCK_STREAM must be set correctly for your system; the
#   correct values are usually in either /usr/include/sys/socket.h
#   or /usr/include/linux/socket.h .
#$AF_INET= 2 ; $SOCK_STREAM= 1 ;  # these are GUESSES; set correct values



# Set this if the script is running on an SSL server, i.e. it is accessed
#   through a URL starting with "https:".  This is needed to know how to
#   route URLs back through the proxy.  Regrettably, standard CGI does not
#   yet provide a way for scripts to determine this without help.
# This has nothing to do with retrieving pages that are on SSL servers.
$RUNNING_ON_SSL_SERVER= 0 ;


# If you're running this on a Windows machine, set this.  Windows lacks some
#   features that exist on Unix, and there are bugs in the IIS server.  If
#   this variable is set, the script tries to work around these limitations.
$RUNNING_ON_WINDOWS= 0 ;


# If your server doesn't support NPH scripts, then set this variable to true
#   and try running the script as a normal non-NPH script.  HOWEVER, this
#   won't work as well as running it as NPH; there may be bugs, maybe some
#   privacy holes, and results may not be consistent.  It's a hack.
# Try to install the script as NPH before you use this option, because
#   this may not work.  NPH is supported on almost all servers, and it's
#   usually very easy to install a script as NPH (on Apache, for example,
#   you just need to name the script something starting with "nph-").
# One example of a problem is that Location: headers may get messed up,
#   because they mean different things in an NPH and a non-NPH script.
#   You have been warned.
# For this to work, your server MUST support the "Status:" CGI response
#   header.
$NOT_RUNNING_AS_NPH= 0 ;


# Here's an experimental feature that may or may not be useful.  It's trivial
#   to add, so I added it.  It was inspired in part by Mike Reiter's and Avi
#   Rubin's "Crowds", at http://www.research.att.com/projects/crowds/ .
#   Let me know if you find a use for it.
# The idea is that you have a number of mutually-trusting, cooperating
#   proxies that you list in @PROXY_GROUP().  If that is set, then instead
#   of rerouting all URLs back through this proxy, the script will choose
#   one of these proxies at random to reroute all URLs through, for each
#   run.  This could be used to balance the load among several proxies, for
#   example.  Under certain conditions it could conceivably help privacy by
#   making it harder to track a user's session, but under certain other
#   conditions it could make it easier, depending on how many people,
#   proxies, and proxy servers are involved.  For each page, both its
#   included images and followed links will go through the same proxy, so a
#   clever target server could determine which proxy servers are in each
#   group.
# proxy_encode() and proxy_decode() must be the same for all proxies in the
#   group.  Same goes for pack_flags() and unpack_flags() if you modified them,
#   and probably certain other routines and configuration options.
# Cookies and Basic authentication can't be supported with this, sorry, since
#   cookies can only be sent back to the proxy that created them.
# Set this to a list of absolute URLs of proxies, ending with "nph-proxy.cgi"
#   (or whatever you named the script).  Be sure to include the URL of this
#   proxy, or it will never redirect back through here.  Each proxy in the
#   group should have the same @PROXY_GROUP.
# Alternately, you could set each proxy's @PROXY_GROUP differently for more
#   creative configuration, such as to balance the load unevenly, or to send
#   users through a "round-robin" cycle of proxies.

#@PROXY_GROUP= ('http://www.example.com/~grommit/proxy/nph-proxy.cgi',
#	        'http://www.fnord.mil/langley/bavaria/atlantis/nph-proxy.cgi',
#	        'http://www.nothinghere.gov/No/Such/Agency/nph-proxy.cgi',
#	        ) ;


# Set HTTP proxy if needed
#$ENV{'http_proxy'}= '' ;
#$ENV{'no_proxy'}= '' ;

# If your HTTP proxy requires authentication, this script supports it in a
#   limited way: you can have a single username/password pair to authenticate
#   with, regardless of realm.  In other words, multiple realms aren't
#   supported for proxy authentication (though they are for normal server
#   authentication, elsewhere).
# Set $PROXY_AUTH either in the form of "username:password", or to the actual
#   base64 string that gets sent in the Proxy-Authorization: header.
#$PROXY_AUTH= 'Aladdin:open sesame' ;


# There's a would-be privacy hole:  If an HTML tag has more than one of the
#   same URL-containing attribute, this proxy only converts the last of that
#   attribute to have its URL point back through this script.  For example,
#   with <img src="evil.cgi" src="decoy.gif">, only decoy.gif will be
#   converted, and evil.cgi will most likely be called directly from your
#   browser.  It's not valid HTML, but it's exploitable.
# So, this program removes duplicate attributes in each tag.  Unfortunately,
#   that action slows down the main URL-converting block by about 10%, and I
#   can't find any faster way.  But if you're not worried about this privacy
#   hole, and performance is a major issue for you, you can set this flag to
#   disable the protection.
$FASTER_HTML_LESS_PRIVACY= 0 ;


# Set $USER_AGENT to something generic like this if you want to be extra
#   careful.  Conceivably, revealing which browser you're using may be a
#   slight privacy or security risk.
# However, note that some URLs serve different pages depending on which
#   browser you're using, so some pages will change if you set this.
# This defaults to the user's HTTP_USER_AGENT.
#$USER_AGENT= 'Mozilla/4.05 [en] (X11; I; Linux 2.0.34 i586)' ;


# Unlike a normal browser which can keep an FTP session open between requests,
#   this script must make a new connection with each request.  Thus, the
#   FTP welcome message (e.g. the README file) will be received every time;
#   there's no way for this script to know if you've been here before.  Set
#   $SHOW_FTP_WELCOME to true to always show the welcome message, or false
#   to never show it.
$SHOW_FTP_WELCOME= 1 ;


# For the obscure case when a POST must be repeated because of user
#   authentication, this is the max size of the request body that this
#   script will store locally.  If CONTENT_LENGTH is bigger than this,
#   the body's not saved at all-- the first POST will be correct, but
#   the second will not happen at all (since a partial POST is worse than
#   nothing).
$MAX_REQUEST_SIZE= 4194304 ;  # that's 4 Meg to you and me


# Apparently, some censoring filters search outgoing request URIs, but not
#   POST request bodies.  Set this to make the initial input form submit
#   using POST instead of GET.
$USE_POST_ON_START= 1 ;


# Apparently, some censoring filters look at titles on HTML pages.  Set this
#   to remove HTML page titles.
$REMOVE_TITLES= 0 ;


# If set, this option prevents a user from calling the proxy through the
#   proxy itself, i.e. looping.  It's normally a mistake on the user's part,
#   and a waste of resources.
# This isn't foolproof; it just catches the obvious mistakes.  It's probably
#   pretty easy for a malicious user to make the script call itself, or s/he
#   can always use two proxies to call each other in a loop.  This doesn't
#   account for IP addresses or multiple hostnames for the same server.
$NO_BROWSE_THROUGH_SELF= 0 ;


# Set this to leave out the "Restart" link at the bottom of error pages, etc.
# In some situations this could make it harder for search engines to find the
#   start page.
$NO_LINK_TO_START= 0 ;



# Stuff below here you probably shouldn't modify unless you're messing with
#   the code.


# This lists all MIME types that could identify a script, and which will be
#   filtered out as well as possible if removing scripts.  HTTP responses with
#   Content-Type: set to one of these will be nixed, certain HTML which links
#   to one of these types will be removed, style sheets with a type here will
#   be disabled, and other odds and ends.
# These are used in matching, so can't contain special regex characters.
# This list is also used for the the experimental $PROXIFY_SCRIPTS function.
# This list is largely taken from the examples at
#     http://www.robinlionheart.com/stds/html4/scripts.html
#   That page describes only the first four below as valid.
# The page at ftp://ftp.isi.edu/in-notes/iana/assignments/media-types/media-types
#   lists all media (MIME) types registered with the IANA, but unfortunately
#   many script types (especially proprietary ones) have not registered with
#   them, and that list doesn't specify which types are script content anyway.
@SCRIPT_MIME_TYPES= ('application/x-javascript', 'application/x-ecmascript',
		     'application/x-vbscript',   'application/x-perlscript',
		     'application/javascript',   'application/ecmascript',
		     'text/javascript',  'text/ecmascript', 'text/jscript',
		     'text/livescript',  'text/vbscript',   'text/vbs',
		     'text/perlscript',  'text/tcl',
		     'text/x-scriptlet', 'text/scriptlet',
		     'application/hta',
		    ) ;



# All MIME types in @SCRIPT_MIME_TYPES and @OTHER_TYPES_TO_REGISTER will be
#   "registered".  Registration helps the script remember which MIME type is
#   expected by a page when downloading embedded URLs, e.g. style sheets.  Any
#   MIME types that need special treatment should be listed here if they're not
#   already in @SCRIPT_MIME_TYPES.
# If you write a handler for a new MIME type in proxify_block(), and that type
#   isn't already listed in @SCRIPT_MIME_TYPES, then add it here.
@OTHER_TYPES_TO_REGISTER= ('text/css') ;


# These are MIME types that we *may* try to rewrite in proxify_block(), e.g.
#   to send all URLs back through this script.  If a type isn't on this list,
#   then we know for certain it should be sent back to the user unchanged,
#   which saves time.
# If you write a handler for a new MIME type in proxify_block(), then add the
#   type here.
# NOT all the types here are actually supported at this time!
# text/html is not on this list because currently it's handled specially.
@TYPES_TO_HANDLE= ('text/css',
		   'application/x-javascript', 'application/x-ecmascript',
		   'application/javascript',   'application/ecmascript',
		   'text/javascript',          'text/ecmascript',
		   'text/livescript',          'text/jscript',
		  ) ;


# The framework is in place to modify script content to pass back through the
#   proxy, though the actual code that modifies a single script block of a
#   given type are not done.  If you want to, say, modify JavaScript in
#   certain ways that work for your purpose, then see the routine
#   proxify_block().  If you set this $PROXIFY_SCRIPTS flag to true, then
#   proxify_block() will be called for every piece of script that comes
#   through this proxy.
# So, to modify script content like this:  a) set this flag to true, and b) go
#   write some code in proxify_block() that modifies the script content the
#   way you want.  You probably want to use the routine full_url(); go read
#   what it does.  Also see @TYPES_TO_HANDLE and @SCRIPT_MIME_TYPES above.
# DON'T set this unless you actually do that programming!  Without any added
#   code, it won't do anything but slow down the program-- dealing with the
#   script-modifying framework takes longer than merely removing scripts, and
#   both take a LOT longer than leaving scripts intact.
# Limited testing shows this adds 20-30% to the running time for script-heavy
#   sites, and very little for script-free sites.  However, this number varies
#   greatly from page to page.  This is ONLY the overhead involved in
#   separating out the script content to call proxify_block(); this does not
#   include anything that is actually done in that routine.
# NOTE:  THIS IS STILL HIGHLY EXPERIMENTAL!!  The framework should work fine,
#   but what goes in proxify_block() is up to you.
# NOTE TOO:  You will almost certainly not be able to anonymize JavaScript
#   completely.  It's not hard to do "mostly", but it turns out to be a very
#   complex problem to do completely; there will almost certainly be exploits
#   that a malicious server can use to get a user's identity.  The purpose of
#   this feature is more to allow scripts to function through the proxy, than
#   to provide bulletproof anonymity.  You may be able to get better anonymity
#   if you remove certain script statements altogether rather than try to
#   modify them, and accept that doing so may break a few scripts.
# The best advice remains:  FOR BEST ANONYMITY, BROWSE WITH SCRIPTS TURNED OFF.
$PROXIFY_SCRIPTS= 0 ;


# This is a list of all file extensions that will be disallowed if
#   $TEXT_ONLY is set.  It's an inexact science.  If you want to ban
#   other file extensions, you can add more to this list.  Note that
#   removing extensions from this list won't necessarily allow those
#   files through, since there are other ways $TEXT_ONLY is implemented,
#   such as only allowing MIME types of text/* .
# The format of this list is one long string, with the extensions
#   separated by "|".  This is because the string is actually used as
#   a regular expression.  Don't worry if you don't know what that means.
# Extensions are roughly taken from Netscape's "Helper Preferences" screen
#   (but that was in 1996).
# Boy, I really need that list of binary file extensions.  Write me if
#   you have one (August 1998).
$NON_TEXT_EXTENSIONS=
	  'gif|jpeg|jpe|jpg|tiff|tif|png|bmp|xbm'   # images
	. '|mp2|mp3|wav|aif|aiff|au|snd'            # audios
	. '|avi|qt|mov|mpeg|mpg|mpe'                # videos
	. '|gz|Z|exe|gtar|tar|zip|sit|hqx|pdf'      # applications
	. '|ram|rm|ra|swf' ;                        # others



#--------------------------------------------------------------------------
#    end of (useful) user configuration
#--------------------------------------------------------------------------

# This is now set directly in footer(), the only place it's used.
# $PROXY_VERSION= '1.5.1' ;


# This is needed to run an NPH script under mod_perl.
# Other stuff needed for mod_perl:
#   must use at least Perl 5.004, or STDIN and STDOUT won't behave correctly;
#   cannot use exit();
#   must initialize or reset all vars;
#   regex's with /o option retain state between calls, so be careful;
#   typeglobbing of *STDIN doesn't work, so must pass filehandles as strings.
local($|)= 1 ;

# For mod_perl, global variables are retained between calls, so they must
#   be initialized correctly.  Here, we assume all UPPER_CASE variables are
#   initialized correctly elsewhere, and no lower_case variables are set
#   before here.  It's a little hacky, deprecated, and possibly error-prone
#   if the assumption is broken elsewhere, but it's fast and simple.
reset 'a-z' ;


# Save $now, and create the Date: header required for all HTTP 1.1 responses.
# ... except, we don't need this when transmitting a server response, so
#   only create this in error routines etc.
#$now= time ;    # for (@goodmen)
#$date_header= &rfc1123date($now, 0) ;


# This script uses whatever version of HTTP the client is using.  So far
#   only 1.0 and 1.1 are supported.
# jsm-- restore this when HTTP 1.1 support is complete.
#($HTTP_VERSION)= $ENV{'SERVER_PROTOCOL'}=~ m#^HTTP/(\d+\.\d+)#i ;
#$HTTP_VERSION= '1.0' unless $HTTP_VERSION=~ /^1\.[01]$/ ;

$HTTP_VERSION= '1.0' ;


# Hack to support non-NPH installation-- luckily, the format of a
#   non-NPH response is almost exactly the same as an NPH response.
#   The main difference is the first word in the status line-- something
#   like "HTTP/1.x 200 OK" can be simulated with "Status: 200 OK", as
#   long as the server supports the Status: CGI response header.  So,
#   we set that first word to either "HTTP/1.x" or "Status:", and use
#   it for all responses throughout the script.
# NOTE:  This is not the only difference between an NPH and a non-NPH
#   response.  For example, the Location: header has different semantics
#   between the two types of responses.  This hack is only an approximation
#   that we hope works most of the time.  It's better to install the script
#   as an NPH script if possible (which it almost always is).
$HTTP_1_X=  $NOT_RUNNING_AS_NPH   ? 'Status:'   : "HTTP/$HTTP_VERSION" ;


# Create @BANNED_NETWORK_ADDRS from @BANNED_NETWORKS.
# No error checking; assumes the proxy owner set @BANNED_NETWORKS correctly.
for (@BANNED_NETWORKS) {
    push(@BANNED_NETWORK_ADDRS, pack('C*', /(\d+)/g)) ;
}


# Canonicalize all MIME types to lowercase.
for (@SCRIPT_MIME_TYPES)        { tr/A-Z/a-z/ }
for (@OTHER_TYPES_TO_REGISTER)  { tr/A-Z/a-z/ }

# Create @ALL_TYPES and %MIME_TYPE_ID, which are inverses of each other.
# This is useful e.g. to identify the MIME type expected in a given download,
#   in a one-character flag.  That's why we limit this to 64 types for now.
# $ALL_TYPES[0] is '', so we can test e.g. "if $MIME_TYPE_ID{$id} ..." .

@ALL_TYPES= ('', @SCRIPT_MIME_TYPES, @OTHER_TYPES_TO_REGISTER) ;
&HTMLdie("Too many MIME types to register.")  if @ALL_TYPES > 64 ;
@MIME_TYPE_ID{@ALL_TYPES}=  0..$#ALL_TYPES ;


# Create a regex that tells us whether we handle a given MIME type.
$TYPES_TO_HANDLE_REGEX= '(' . join("|", @TYPES_TO_HANDLE) . ')' ;


# Fix submitted by Alex Freed:  Under some unidentified conditions,
#   instances of nph-proxy.cgi can hang around for many hours and drag the
#   system.  So until we figure out why that is, here's a 10-minute timeout.
#   Please write me with any insight into this, since I can't reproduce the
#   problem.  Under what conditions, on what systems, does it happen?
# 9-9-1999: One theory is that it's a bug in older Apaches, and is fixed by
#   upgrading to Apache 1.3.6 or better.  Julian Haight reports seeing the
#   same problem with other scripts on Apache 1.3.3, and it cleared up when
#   he upgraded to Apache 1.3.6.  Let me know if you can confirm this.
# Windows doesn't seem to support alarms, so don't do this on Windows.

unless ($RUNNING_ON_WINDOWS) {
    $SIG{'ALRM'} = 'timeexit' ;
    alarm(600);
}

# Exit upon timeout.  If you wish, add code to clean up and log an error.
sub timeexit { $ENV{'MOD_PERL'}  ? goto EXIT  : exit 1 }


# Set $THIS_HOST to the best guess how this script was called-- use the
#   Host: request header if available; otherwise, use SERVER_NAME.
# We don't bother with a $THIS_PORT, since it's more reliably set to the port
#   through which the script was called.  SERVER_NAME is much more likely to
#   be different from the hostname that the user sees, since one server may
#   handle many domains or have many hostnames.
if ($ENV{'HTTP_HOST'} ne '') {
    ($dummy, $THIS_HOST)= $ENV{'HTTP_HOST'}=~ m#^([\w+.-]+://)?([^:/?]*)# ;
    $THIS_HOST= $ENV{'SERVER_NAME'}   if $THIS_HOST eq '' ;
} else {
    $THIS_HOST= $ENV{'SERVER_NAME'} ;
}


# SCRIPT_NAME should have a leading slash, but the old CGI "standard" from
#   NCSA was unclear on that, so some servers didn't give it a leading
#   slash.  Here we ensure it has a leading slash.
$ENV{'SCRIPT_NAME'}=~ s#^/?#/# ;


# The IIS server doesn't set PATH_INFO correctly-- it sets it to the entire
#   request URI, rather than just the part after the script name.  So fix it
#   here if we're running on IIS.  Thanks to Dave Moscovitz for the info!
if ( $RUNNING_ON_WINDOWS && ($ENV{'SERVER_SOFTWARE'}=~ /IIS/) ) {
    $ENV{'PATH_INFO'} =~ s/^$ENV{'SCRIPT_NAME'}// ;
}


# PATH_INFO may or may not be URL-encoded when we get it; it seems to vary
#   by server.  This script assumes it's still encoded.  Thus, if it's not,
#   we need to re-encode it.
# The only time this seems to come up is when spaces are in URLs, correctly
#   represented in the URL as %20 but decoded to " " in PATH_INFO.  Thus,
#   this hack only focuses on space characters.  It's a hack that I'm not at
#   all comfortable with.  :P
# Very yucky business, this encoding thing.
if ($ENV{'PATH_INFO'}=~ / /) {
    $ENV{'PATH_INFO'} =~ s/%/%25/g ;
    $ENV{'PATH_INFO'} =~ s/ /%20/g ;
}


# PATH_INFO consists of a path segment of flags, followed by the encoded
#   target URL.  For example, PATH_INFO might be something like
#   "/010100/http/www.example.com".  The actual format of the flag segment
#   is defined in the routine pack_flags().
# Thanks to Mike Harding for the idea of using another flag for the
#   $is_in_frame parameter, instead of using two parallel scripts.

# Extract flags and encoded URL from PATH_INFO.
($FLAGS, $encoded_URL)= $ENV{'PATH_INFO'}=~ m#/([^/]*)/?(.*)# ;

# Set all flags based on flag segment from PATH_INFO.
# Only set user options if allowed, and if flag segment was in PATH_INFO.
if ( $ALLOW_USER_CONFIG && ($FLAGS ne '') ) {
    ($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS, $HIDE_REFERER,
     $INSERT_ENTRY_FORM, $is_in_frame, $expected_type)=
	 &unpack_flags($FLAGS) ;

} else {
    # $is_in_frame is set in any case.  It indicates whether the current
    #   request will be placed in a frame.  It is almost the same as indicating
    #   whether or not an insertion is allowed to happen here, and it's
    #   sometimes used that way.  At some point the semantics of this and
    #   $DOING_INSERT could be cleaned up and clarified.
    ($is_in_frame, $expected_type)= (&unpack_flags($FLAGS))[5..6] ;
}


# Flags are now set, and $encoded_URL now contains only the encoded target URL.



# Create a one-flag test for whether we're inserting anything.
# This must happen after user flags are read, just above.
# This flag generally indicates whether this proxy is currently configured
#   to insert something, not whether the current run of the program needs
#   an insertion.  That would be closer to $is_in_frame.
$DOING_INSERT= ($INSERT_FILE ne '') || ($INSERT_HTML ne '')
	     || $INSERT_ENTRY_FORM ;


# Base64-encode $PROXY_AUTH if it's not done so already.
$PROXY_AUTH= &base64($PROXY_AUTH)  if $PROXY_AUTH=~ /:/ ;


# Default to $ENV{'HTTP_USER_AGENT'}
$USER_AGENT= $ENV{'HTTP_USER_AGENT'} if $USER_AGENT eq '' ;


# Copy often-used environment vars into scalars, for efficiency
$env_accept= $ENV{'HTTP_ACCEPT'} || '*/*' ;     # may be modified later


# One user reported problems with binary files on certain other OS's, and
#   this seemed to fix it.  Supposedly, either this or the "binmode S"
#   statements below the newsocketto() calls work, or all; I'm putting all in.
#   Tell me anything new you figure out about this.
binmode STDOUT ;


# %urlsin was an associative array that listed tags and attributes that
#   may contain URLs, declared like:
# %urlsin= ('a', 'href',
#           'applet', 'codebase',
#           'fig', 'src|imagemap',
#           'form', 'action|script',  ...
#          ) ;


# OK, let's time this thing
#$starttime= time ;
#($sutime,$sstime)= (times)[0,1] ;

#--------------------------------------------------------------------------
#    parse URL, make checks, and set various globals
#--------------------------------------------------------------------------

# Calculate $this_url for use later in &full_url() and elsewhere.  It's an
#   integral part of &full_url(), placed here for speed, similar to the
#   variables set in &fix_base_vars.
# $this_url is an absolute URL pointing to this proxy script, complete with
#   trailing slash.  It has to be absolute to serve those HTML attributes
#   that require absolute URLs, like <base href>.
# Included in $this_url is the initial flag segment of PATH_INFO.  Thus, a
#   complete $this_url might be "http://www.example.com/nph-proxy.cgi/01010/" .
# Another way to describe $this_url is that it's everything in a complete
#   proxified URL except the encoded target URL at the end.  It's almost
#   always used by appending it with &proxy_encode(something).
# $script_url is used later for Referer: support, and whenever a temporary
#   copy of $this_url has to be generated.

# Set $this_url to a random element of @PROXY_GROUP, if that is set.
if (@PROXY_GROUP) {
    srand( $$ ^ unpack('%32L*', $ENV{'PATH_INFO'}) ) ;  # seed with $$, URL
    $script_url= $PROXY_GROUP[ rand(scalar @PROXY_GROUP) ] ;
    $script_url=~ s#/$## ;    # guarantee no trailing slash

} else {
    # Build the local $this_url from environment variables.  Only include
    #   SERVER_PORT if it's not 80 (or 443 for SSL).
    $script_url= $RUNNING_ON_SSL_SERVER
	?  join('', 'https://', $THIS_HOST,
		($ENV{'SERVER_PORT'}==443  ? ''  : ':'.$ENV{'SERVER_PORT'}),
		$ENV{'SCRIPT_NAME'})
	:  join('', 'http://', $THIS_HOST,
		($ENV{'SERVER_PORT'}==80  ? ''  : ':'.$ENV{'SERVER_PORT'}),
		$ENV{'SCRIPT_NAME'}) ;
}

# Create $this_url and any needed variants: "$script_url/flags/"
$this_url= $script_url . '/' .
    &pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS, $HIDE_REFERER,
		$INSERT_ENTRY_FORM, $is_in_frame, '') . '/' ;
$this_url_inframe= $script_url . '/' .
    &pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS, $HIDE_REFERER,
		$INSERT_ENTRY_FORM, 1, '') . '/' ;
$this_url_noframe= $script_url . '/' .
    &pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS, $HIDE_REFERER,
		$INSERT_ENTRY_FORM, 0, '') . '/' ;


# If there's no $encoded_URL, then start a browsing session.
&show_start_form if $encoded_URL eq '' ;


# Decode the URL.
$URL= &proxy_decode($encoded_URL) ;


# Set the query string correctly, from either $ENV{QUERY_STRING} or what's
#   already in $URL.
# The query string may exist in either the encoded URL or in the containing
#   URL, as $ENV{QUERY_STRING}.  If the former, then the query string was
#   (definitely?) in a referenced URL, while the latter most likely implies a
#   GET form input.  Either query string is valid, but form input takes
#   precedence-- if $ENV{QUERY_STRING} exists, it should be used over any
#   query string in the encoded URL.
# Note that Netscape does not pass any query string data that is part of the
#   URL in the <form action> attribute, which is probably correct behaviour.
#   For this program to act exactly the same, it would need to strip the
#   query string when updating all <form action> URLs, way below.
# Question:  Is there ever a valid case when both QUERY_STRINGs exist??

$URL=~ s/(\?.*)?$/?$ENV{'QUERY_STRING'}/   if $ENV{'QUERY_STRING'} ne '' ;


# Parse the URL, using a regex modelled from the one in RFC 2396 (URI syntax),
#   appendix B.
# This assumes a hierarchical scheme; it won't work for e.g. mailto:
# "authority" is the combination of host, port, and possibly other info.
# Note that $path here will also contain any query component; it's more like
#   the request URI.
# Note that $URL is guaranteed to be an absolute URL with no "#" fragment,
#   though this does little error-checking.  Note also that the old ";"
#   parameters are now included in the path component.

($scheme, $authority, $path)= ($URL=~ m#^([\w+.-]+)://([^/?]*)(.*)$#i) ;
$scheme=~ tr/A-Z/a-z/ ;
$path= "/$path" if $path!~ m#^/# ;   # if path is '' or contains only query


# Magic here-- if $URL uses special scheme "x-proxy", immediately call the
#   general-purpose xproxy() routine.
&xproxy($URL) if $scheme eq 'x-proxy' ;


# Set $is_html if $path (minus query) ends in .htm or .html .
# MSIE has a bug (and privacy hole) whereby URLs with QUERY_STRING ending
#   in .htm or .html are mistakenly treated as HTML, and thus could have
#   untranslated links, <script> blocks, etc.  So for those cases, set
#   $is_html=true to make sure we later transform it as necessary.
if ($ENV{'HTTP_USER_AGENT'}=~ /MSIE/) {
    $is_html= 1  if $path=~ /\.html?(\?|$)/i ;
} else {
    $is_html= 1  if $path=~ /^[^?]*\.html?(\?|$)/i ;
}


# Alert the user to unsupported URL, with an intermediate page
&unsupported_warning($URL) unless ($scheme=~ /^(http|ftp)$/) ;

# Require a host to be present (for $base_url safety later)
# Testing for a valid hostname is more complex than just /\w/ , but this does
#   what we need here.
&HTMLdie('The target URL cannot contain an empty host name.')
    unless $authority=~ /^\w/ ;


# Parse $authority into $host, $port, and possibly others, depending on
#   which URL scheme is used.
# Since most URL schemes use the simple host:port, make that the default.
#   This may avoid oversight later when other URL schemes are added (though
#   be careful of username/password handling in that block below).
# Note that this does not set $port to a default.  In the interest of
#   encapsulation, the default $port should be set in the routine that
#   implements the protocol (i.e. http_get(), ftp_get(), etc.)

if ($scheme eq 'ftp') {
    # FTP authority can be username:password@host:port, with username,
    #   password, and port all optional.
    # Embedding your username/password in a URL is NOT RECOMMENDED!  Here,
    #   the second clause should almost always be used.
    if ($authority=~ /@/) {
	($username, $password, $host, $port)=
	    $authority=~ /([^:@]*):?([^@]*)@([^:]*):?(.*)/ ;
    } else {
	($username, $password)= ('anonymous', 'not@available.com') ;
	($host, $port)= $authority=~ /^([^:]*):?(.*)$/ ;
    }

# covers HTTP, etc.
} else {
    # Unlikely occurrence of username:password@host:port, but possible.
    #   Implies HTTP Basic authentication.  Not as much a security hole as
    #   doing the same in an FTP URL, above, but still not a great idea.
    if ($authority=~ /@/) {
	($username, $password, $host, $port)=
	    $authority=~ /([^:@]*):?([^@]*)@([^:]*):?(.*)/ ;
    } else {
	($host, $port)= $authority=~ /^([^:]*):?(.*)$/ ;
    }
}

$host=~ tr/A-Z/a-z/ ;   # hostnames are case-insensitive
$host=~ s/\.*$//g ;     # removes trailing dots to close a potential exploit


# If so configured, disallow browsing back through the script itself (looping).
# This assumes the script can only be called by an http:// or https:// URL.
# This could check SERVER_NAME in addition to $THIS_HOST, but that might
#   match falsely sometimes.  The way it is should still prevent deep loops.
if ($NO_BROWSE_THROUGH_SELF) {
    # Default $port's not set yet, so hack up an ad hoc version.
    local($port2)=  $port || ( $scheme eq 'https'  ? 443  : 80 ) ;
    &loop_disallowed_die($URL)
	if     ($scheme=~ /^https?/)
	    && ($host=~ /^$THIS_HOST$/i)
	    && ($port2 == $ENV{'SERVER_PORT'})
	    && ($path=~ /^$ENV{'SCRIPT_NAME'}\b/) ;
}


# Die if the target server is not allowed.
if (@ALLOWED_SERVERS) {
    local($server_is_allowed) ;
    foreach (@ALLOWED_SERVERS) {
	$server_is_allowed= 1, last   if $host=~ /$_/ ;
    }
    &banned_server_die unless $server_is_allowed ;
}
foreach (@BANNED_SERVERS) {
    &banned_server_die if $host=~ /$_/ ;
}


# If we're filtering ads, force $NO_COOKIE_WITH_IMAGE=1, and set
#   $images_are_banned_here appropriately.
if ($FILTER_ADS) {
    $NO_COOKIE_WITH_IMAGE= 1 ;
    foreach (@BANNED_IMAGE_URL_PATTERNS) {
	$images_are_banned_here= 1, last if $URL=~ /$_/ ;
    }
}


# Set $scripts_are_banned_here appropriately
$scripts_are_banned_here= $REMOVE_SCRIPTS ;
unless ($scripts_are_banned_here) {
    if (@ALLOWED_SCRIPT_SERVERS) {
	$scripts_are_banned_here= 1 ;
	foreach (@ALLOWED_SCRIPT_SERVERS) {
	    $scripts_are_banned_here= 0, last   if $host=~ /$_/ ;
	}
    }
    unless ($scripts_are_banned_here) {
	foreach (@BANNED_SCRIPT_SERVERS) {
	    $scripts_are_banned_here= 1, last   if $host=~ /$_/ ;
	}
    }
}


# Set $cookies_are_banned_here appropriately
$cookies_are_banned_here= $REMOVE_COOKIES ;
unless ($cookies_are_banned_here) {
    if (@ALLOWED_COOKIE_SERVERS) {
	$cookies_are_banned_here= 1 ;
	foreach (@ALLOWED_COOKIE_SERVERS) {
	    $cookies_are_banned_here= 0, last   if $host=~ /$_/ ;
	}
    }
    unless ($cookies_are_banned_here) {
	foreach (@BANNED_COOKIE_SERVERS) {
	    $cookies_are_banned_here= 1, last   if $host=~ /$_/ ;
	}
    }
}


# Regex that matches a script MIME type, and analogous hash.
if ($scripts_are_banned_here) {
    $SCRIPT_TYPE_REGEX= '(' . join("|", @SCRIPT_MIME_TYPES) . ')' ;
}


# Disallow the retrieval if the expected MIME type is banned, because some
#   browsers erroneously give the advisory content-type precedence over
#   everything else.
if ($scripts_are_banned_here && $expected_type ne '') {
    &script_content_die if $expected_type=~ /^$SCRIPT_TYPE_REGEX$/io ;
}

# Exclude non-text if it's not allowed.  Err on the side of allowing too much.
if ($TEXT_ONLY) {
    # First, forbid requests for filenames with non-text-type extensions
    &non_text_die if ($path=~ /\.($NON_TEXT_EXTENSIONS)(;|\?|$)/i) ;

    # Then, filter the "Accept:" header to accept only text
    $env_accept=~ s#\*/\*#text/*#g ;    # not strictly perfect
    $env_accept= join(', ', grep(m#^text/#i, split(/\s*,\s*/, $env_accept)) ) ;
    &non_text_die unless $env_accept ne '' ;
}


# For a potential banner ad, intercept request if it looks like an image is
#   requested, i.e. unless the Accept: header allows either text/... or */... .
if ($images_are_banned_here) {
    &skip_image unless grep(m#^(text|\*)/#i, split(/\s*,\s*/, $env_accept) ) ;
}


$*= 1 ;     # allow multi-line matching


# $base_url must be set correctly at any time &full_url() may be called.
#   &fix_base_vars() must be called as well, to set $base_scheme, $base_host,
#   and $base_path.
# Unfortunately, the base URL may change over the course of this program.  We
#   will keep it set based on whatever info we have so far, i.e. request URI,
#   then e.g. HTTP response headers, then e.g. <base> tag (which happens to
#   be in the reverse order of the ultimate precedence).
$base_url= $URL ;
&fix_base_vars ;   # must be called whenever $base_url is set


# The next two variables $default_style_type and $default_script_type must be
#   kept up-to-date throughout the run of this program, just like $base_url
#   and its related variables.  They should always be canonicalized to
#   lowercase (MIME types are case-insensitive).
# Note that if these aren't handled carefully, then there could be a privacy
#   hole-- for example, style sheets of a script type could cause execution of
#   script content.

# Any style content ("style" attributes, <style> elements, or external
#   style sheets) that does not have a type defined uses the default style
#   sheet language. That should be specified in a Content-Style-Type: header
#   or equivalent <meta> tag, but if not then the default is text/css.
# This *should* only be needed for style attributes, but if the other two
#   forms of style content erroneously don't specify a type then it could be
#   used for them.
$default_style_type= 'text/css' ;

# Any script content (intrinsic events attributes (i.e. those named "on___")
#   or <script> elements) that does not have a type defined uses the default
#   script language.  That should be specified in a Content-Script-Type: header
#   or equivalent <meta> tag, but if not then the default is
#   application/x-javascript.
# This *should* only be needed for intrinsic event attributes, but if <script>
#   elements erroneously don't specify a type then it could be used for them.
$default_script_type= 'application/x-javascript' ;


# Parse the cookie for real cookies and authentication information.
($cookie_to_server, %auth)=
    &parse_cookie($ENV{'HTTP_COOKIE'}, $path, $host, $scheme) ;


#--------------------------------------------------------------------------
#    Retrieve the resource into $body or @body using the correct scheme,
#      also setting $status, $headers, and $is_html (all globals).
#      $is_html indicates whether the original resource is HTML, not
#      if a generated response is in HTML (e.g. an error message).
#    $response_sent might be set, indicating the response was sent by the
#      subroutine.  This is appropriate for streaming media, for example.
#--------------------------------------------------------------------------

if ($scheme eq 'http') {
    &http_get ;
} elsif ($scheme eq 'ftp') {
    &ftp_get ;
}

#--------------------------------------------------------------------------
#    Modify entire response to point back through this script
#--------------------------------------------------------------------------

# NOTE: IT IS IMPORTANT TO DO THIS AS COMPLETELY AS POSSIBLE!  IF A
# USER UNKNOWINGLY GOES TO A PAGE DIRECTLY AND NOT THROUGH THIS PROXY,
# HE/SHE MAY REVEAL HIM/HERSELF IN AN UNINTENDED WAY.

#--------------------------------------------------------------------------
# These were notes to myself from testing the speed of different methods.
#   Names like "nph-proxy2" refer to different modifications.  This version,
#   the fastest, was called nph-proxy2b.
#
# If YOU figure out a faster method, please tell me about it!
#
# It would certainly be much faster if rewritten in C, because you could
#   very quickly read each character from the input and write it to the
#   output, maintaining state and altering the data stream as needed.
#
#--------------------------------------------------------------------------
#
# [This version is nph-proxy2b:  Break into @body array, do not use
#   %urlsin to test if tags should be updated, replacement strings
#   use (a|b|c) syntax.]
#
# This is by far the most time-consuming part of the script, the updating
#   of all URLs in an HTML file.
#
# Results of informal speed testing (not what I expected):
#   Breaking into @body array instead of one big string definitely saves
#       significant time-- compare nph-proxy2 to nph-proxy1.  One test
#       showed a time saving of 1/3 to 1/2.
#   Using a %urlsin array does NOT seem to save time, even when only used
#       as boolean test to see if tag might contain URL-- compare nph-proxy2
#       to nph-proxy3, nph-proxy4, and (most similar) nph-proxy5.  Oh well.
#   It seems that reading one tag at a time, converting, and sending it
#       through does NOT save elapsed time over reading all tags before
#       converting, like I thought it would.  Both the CPU and elapsed
#       time are longer for one-tag-at-a-time approach-- compare
#       nph-proxy2 and nph-piper2.
#
#   Results of nph-proxy2 (blocks for each tag) to nph-proxy2b
#   (single "(att1|att2|att3)" style regex) testing:
#       Mixed results, but overall, using a single regex (nph-proxy2b)
#       takes less CPU "user time", about the same "system time", and
#       slightly more elapsed time than nph-proxy2; I don't know why.
#       The elapsed time is slightly more both within the script and at
#       the HTTP client's end.  All differences are less than 10% on
#       average, and nph-proxy2b occasionally shows LESS elapsed time
#       than nph-proxy2.  CPU time is always less for nph-proxy2b than
#       for nph-proxy2.
#   Since the bottleneck of CPU time is tighter than for elapsed time,
#       and since the elapsed-time loss is less than the CPU time gain,
#       let's go with nph-proxy2b (not that it really makes much difference).
#       Besides, the code is easier to read that way.
#
# So basically, breaking into @body array helps, but not much else does.
#
#--------------------------------------------------------------------------
# 8-4-98 JSM: Found a bug in the regex, so changed it.  It more resembles
#   nph-proxy2.cgi now.
#--------------------------------------------------------------------------


# If the resource is HTML (and not empty), update all URLs in all tags that
#   refer to URLs.  Plus a bunch of other stuff.

if ( $is_html  && ($body[0] ne '') && !$response_sent ) {

    local($base_unframe, $base_target) ;

    # Set $base_url if there is an HTML <base> tag.  Note that $base_url is
    #   set to the original URL from the HTML, *not* the converted URL that
    #   goes through the script.
    # Also set $base_unframe to set frame flag correctly later.  It indicates
    #   whether the <base> target explicitly exits to _top or _blank.
    foreach (@body) {
	if (/<\s*base\b/i) {
	    # Only if there is a target attribute, set $base_unframe to 0 or 1.
	    $base_unframe= $base_target=~ /^(_top|_blank)$/i
		if ($base_target)= /<[^>]*[^\w.>\/?&-]target\s*=\s*["']?\s*([^\s"'>]*)\b/i ;
	    $base_url= &HTMLunescape($1), &fix_base_vars, last
		if m#<\s*base\b[^>]*[^\w.>\/?&-]href\s*=\s*["']?\s*([\w+.-]+://[^\s"'>]+)#i ;
	}
	last if m#<\s*(/head|body)\b#i ; # stop looking after </head> or <body>
    }


    # If so configured, remove script elements as much as possible.
    if ($scripts_are_banned_here) {
	local($s) ;  # sequence number in .. tests
	foreach (@body) {

	    # Strip everything between <script> and </script>.
	    # Note the use of the seldom-seen ".." scalar operator.
	    if ( $s=(/<\s*script\b/i .. /<\s*\/script\b/i) ) {
		# Don't remove the part before the initial "<script>".
		if ($s==1) { s/<\s*script\b(.|\n)*>//i }
		else       { $_= '' }
	    }

	    # Remove JavaScript "conditional comments", which begin with
	    #   "<!--&{".  They evaluate the initial expression, and depending
	    #   on that, include or exclude the rest of the comment.
	    # Now, comments don't have to end with "-->"; there can be stuff
	    #   in between the "--" and the final ">".  However, it seems that
	    #   most browsers only close comments on exactly "-->", so let's
	    #   err on the side of removing too much: remove the whole comment.
	    #   It shouldn't be a big deal either way, since embedded
	    #   <script>'s will still be stripped out, above.  We just don't
	    #   want the initial expression to be evaluated.
	    # Note the use of the seldom-seen ".." scalar operator.
	    if ( $s= (/<!--\s*&{/ .. /-->/) ) {
		# Don't remove the part before the initial "<!--&{".
		if ($s==1) { s/<!--(.|\n)*//i }
		else       { $_= '' }
	    }

	    # Remove intrinsic event attributes, aka "(Java-)script
	    #   attributes".  For now, assume all (and only) attributes that
	    #   begin with "on" are intrinsic event attributes; this is the
	    #   case with HTML 4.01 as of late 1999.
	    # Note that this regex suffers from the same imperfections as
	    #   the similar regex used below.  It may be more susceptible
	    #   to errors, since it catches every word beginning with "on"
	    #   followed by an "=", possibly with intervening blanks.
	    1 while s/(<\s*\w+[^>]*[^\w.>\/?&-])on\w+\s*=\s*("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)/$1/i ;

	    # Remove script macros, aka the Netscape-specific "JavaScript
	    #   entities", which allow JavaScript to be invoked within *any*
	    #   HTML attribute.  Such attributes contain "&{" (which
	    #   unfortunately means that attributes that innocently contain
	    #   that string will be removed too).  Do this for all browsers,
	    #   in case others emulate Netscape.  Remove the entire attribute.
	    1 while s/(<\s*\w+[^>]*[^\w.>\/?&-])\w+\s*=\s*("[^">]*&{[^">]*"?|'[^'>]*&{[^'>]*'?|[^\s>]*&{[^\s>]*)/$1/ ;

	    # Here was a block that removed tags that linked to a script
	    #   MIME type, but that code has now moved to the blocks of the
	    #   individual tags (<object>, <a>, <link>, and <param>, taken from
	    #   http://www.w3.org/TR/REC-html40/index/attributes.html , type
	    #   "%ContentType").


	}  # foreach (@body)


	# Remove the MSIE-specific "dynamic properties" in style attributes.
	#   I can't find a definitive reference for their syntax (tell me if
	#   you know of one), so err on the safe side and remove all style
	#   attributes that contain the string "expression(".  I think I've
	#   seen "function()" used too, so remove those.  But I don't know
	#   either JS or style sheets.
	if ($ENV{'HTTP_USER_AGENT'}=~ /MSIE/) {
	    foreach (@body) {
		1 while s/(<\s*\w+[^>]*[^\w.>\/?&-])style\s*=\s*("[^">]*expression\s*\([^">]*"?|'[^'>]*expression\s*\([^'>]*'?|[^\s>]*expression\s*\([^\s>]*)/$1/i ;
		1 while s/(<\s*\w+[^>]*[^\w.>\/?&-])style\s*=\s*("[^">]*function\s*\([^">]*"?|'[^'>]*function\s*\([^'>]*'?|[^\s>]*function\s*\([^\s>]*)/$1/i ;
	    }
	}


    }  # if ($scripts_are_banned_here)



    # Remove duplicate attributes-- they're not valid HTML and open a rare
    #   privacy hole in this script, because the conversion regex's below
    #   only transform the last of any attribute in a tag.
    # This can be done after scripts are removed, but has to be before scripts
    #   are proxified.
    unless ($FASTER_HTML_LESS_PRIVACY) {
	foreach (@body) {

	    # Skip this inside <script> blocks.
	    # <script> blocks are supposed to end on any "</", but in reality
	    #   browsers seem to end them on <script>.  See notes below in
	    #   $PROXIFY_SCRIPTS block; in short, we'll end on </script> for
	    #   now since that's what browsers do.  We can revisit this if
	    #   browsers start ending <script> blocks on "</".
	    # Alternately, one could argue that we should err on the side of
	    #   doing this too much rather than too little, and should thus
	    #   end <script> blocks here on "</" just in case any browsers
	    #   act correctly.  But that would mangle some script content and
	    #   have zero actual benefit, the way browsers are now.  *sigh*
	    # Frankly, if scripts are allowed by the user, then anonymity
	    #   can't be critical anyway.  Thus, ending on </script> is fine.
	    next if ( /<\s*script\b/i .. /<\s*\/script\b/i) ;

	    # This regex isn't perfect, but will catch all offending HTML.
	    # It may, however, catch some others, like script statements.
	    # Unfortunately, this one statement slows down the main block of
	    #   the program by about 10%-- yuck.  A few alternate methods
	    #   are below, but they all cost 9-11%.  They are: a) disable
	    #   the attribute by appending its name with "_bad"; b) remove
	    #   the whole tag (which messes up sites which mistakenly use
	    #   duplicate attributes, like Slashdot); and c) remove the part
	    #   of the tag after a duplicate attribute.
	    # If you find a faster way to do this, PLEASE TELL ME!
	    # 4-4-01:  Changed regex to only match when tag starts with \w .
	    #   This helps avoid altering some JavaScript, which often lives
	    #   inside <!-- --> comments.

	    1 while s/(<\s*\w[^>]*[^\w.>\/?&-]([\w.-]+)\s*=[^>]*[^\w.>\/?&-])\2\s*=\s*("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)/$1/i ;   # the "right" way
#	    1 while s/(<[^>]*[^\w.>\/?&-]([\w.-]+)\s*=[^>]*[^\w.>\/?&-])\2\s*=\s*("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)/$1/i ;   # the "right" way
#	    1 while s/(<[^>]*\b(\w+)\s*=[^>]*[\s"']\2)\s*=/${1}_bad=/i ;   # a)
#	    next if s/(<[^>]*\b(\w+)\s*=[^>]*[\s"']\2\s*=[^>]*>?)//i ;     # b)
#	    next if s/(<[^>]*\b(\w+)\s*=[^>]*)[\s"']\2\s*=[^>]*(>?)/$1$3/i;# c)

	}
    }



    # If the proxy owner is trying the experimental proxifying of scripts,
    #   then do this block.  The same cases are handled as immediately above,
    #   but the script content is altered rather than removed.  Those cases
    #   are:  <script> blocks, conditional comments, intrinsic event
    #   attributes, script macros, and MSIE-specific "dynamic properties".
    #   For comments regarding those, see immediately above.
    # Do this after the $FASTER_HTML_LESS_PRIVACY block above, because we
    #   want all duplicate attributes removed before trying to anonymize
    #   scripts.

    if (!$scripts_are_banned_here && $PROXIFY_SCRIPTS) {
	local($*)= 0 ;    # allow $ for EOS in this section

	# Handle <script>...</script> blocks.
	# For info about non-HTML blocks in an HTML document, see
	#   http://www.w3.org/TR/html40/appendix/notes.html#h-B.3.2
	# This block is copied from the <style>-handling block about 100
	#   lines down.  There are important comments there regarding how
	#   browsers vary in how they end <script> blocks, e.g. Konqueror
	#   allows "</script>" inside literal strings, and Netscape doesn't.
	#   In fact, the HTML spec indicates all script and style blocks
	#   should end with the first "</", but browsers don't do that.
	# Because it's a given that we can't anonymize scripts completely, but
	#   we do want to anonymize HTML completely, we'd rather accidentally
	#   treat script content as HTML than the other way around.  So err on
	#   ending the <script> block too soon for some browsers, i.e. end on
	#   the first "</script>" regardless of whether it's in a string.
	#   (We'd end on "</", but *no* browser seems to do that.)
	# The basic approach here is to extract a <script>...</script> block
	#   from @body, convert it, then insert it back into @body, being
	#    careful to keep $i set appropriately.
	# If a proxified block comes back '', then the element contents will be
	#   removed but the <script></script> tags will remain.  This will never
	#   be a problem as far as I know.
	local($i, $startblock, $script_type, $block, $newblock, @newblock) ;
	for ($i=0 ; $i<@body ; $i++) {
	    next unless $body[$i]=~ /<\s*script\b/i ;
	    $startblock= $i ;
	    $i++ until ($body[$i]=~ m#</script\s*>#i) || ($i>=@body) ;

	    ($script_type)= $body[$startblock]=~
		/<\s*script\b[^>]*[^\w.>\/?&-]type\s*=\s*["']?\s*([^\s"'>]*)/i ;
	    $script_type=~ tr/A-Z/a-z/ ;

	    # If there's no type, but there's a language attribute, then use
	    #   it instead to guess the expected type.
	    if (  ($script_type eq '') &&
		  ($body[$startblock]=~
			m#<[^>]*[^\w.>\/?&-]language\s*=\s*["']?\s*([/\w.+\$-]+)#i ) )
	    {
		local($language)= $1 ;
		$script_type=  $language=~ /javascript|ecmascript|livescript|jscript/i
						  ? 'application/x-javascript'
		      : $language=~ /css/i        ? 'text/css'
		      : $language=~ /vbscript/i   ? 'application/x-vbscript'
		      : $language=~ /perl/i       ? 'application/x-perlscript'
		      : $language=~ /tcl/i        ? 'text/tcl'
		      :                             ''
	    }

	    # else use $default_script_type, even though it's not legal HTML.
	    $script_type= $default_script_type if $script_type eq '' ;

	    $block= join('', @body[$startblock+1..$i]) ;
	    $block=~ s#</script\s*>$##i ;
	    $newblock= &proxify_block($block, $script_type) . '</script>' ;

	    # Split into list; each element ends in ">" or EOstring.
	    # This regex gives an ending '' element, so pop it.
	    @newblock= $newblock=~ /([^>]*>?)/g ;
	    pop(@newblock) if $newblock[$#newblock] eq '' ;

	    splice(@body, $startblock+1, $i-$startblock, @newblock) ;

	    # set list counter correctly; it will be incremented just after.
	    $i= $startblock + @newblock ;
	}


	# Handle conditional comments, which begin with "<!--&{" and end with
	#   "};" , and determine whether the rest of the comment block (until
	#   "-->" ) should be rendered on the page.  The only script content
	#   we handle here is the initial expression; any script content in
	#   the comment block is handled elsewhere.
	local($i, $startblock, $block, $prefix, $suffix, $newblock, @newblock) ;
	for ($i=0 ; $i<@body ; $i++) {
	    next unless $body[$i]=~ /<!--\s*&{/ ;
	    $startblock= $i ;
	    $i++ until ( $body[$i]=~ /}\s*;/ ) || ($i>=@body) ;

	    $block= join('', @body[$startblock..$i]) ;

	    # Split on the first "<!--&{" , and then on the first "};" after that.
	    ($prefix, $block)= split(/<!--\s*&{/, $block, 2) ;
	    ($block, $suffix)= split(/}\s*;/, $block, 2) ;
	    $newblock= $prefix . '<!--&{'
		     . &proxify_block($block, $default_script_type)
		     . '};' . $suffix ;

	    # Split into list; each element ends in ">" or EOstring.
	    # This regex gives an ending '' element, so pop it.
	    @newblock= $newblock=~ /([^>]*>?)/g ;
	    pop(@newblock) if $newblock[$#newblock] eq '' ;

	    splice(@body, $startblock, $i-$startblock+1, @newblock) ;

	    # set list counter correctly; it will be incremented just after.
	    $i= $startblock + @newblock - 1 ;
	}


	# For intrinsic event attributes and script macros, handle them
	#   together:  loop through each tag and replace each attribute as
	#   needed.
	# This requires fairly complete parsing of the tag.  To avoid doing
	#   this for every tag, test a regex to filter out almost all
	#   skippable tags.

	# Trouble with Netscape, script macros, and character entities--
	# Normally, we'd HTML-unescape the value and re-escape it when we're
	#   done altering it, but there's a problem with that-- Netscape
	#   handles &{}; blocks in a funny way as far as character entities go.
	#   Apparently it treats a block inside &{}; as if it can't contain
	#   entities, though quoted strings inside that block *can* (and
	#   sometimes *must*).  This seems to be in violation of the HTML and
	#   SGML specs, as far as I can tell; see:
	#     http://www.w3.org/TR/html40/types.html#h-6.14
	#     http://www.w3.org/TR/html40/appendix/notes.html#h-B.7
	#     http://www.w3.org/TR/html40/appendix/notes.html#h-B.3.2
	# Hopefully, the page author has already properly escaped the
	#   characters anyway.  So when processing script macros inside &{}; ,
	#   we do no HTML-unescaping or -escaping.
	# Netscape also allows quotes inside script macros that match the
	#   outer enclosing quotes, but for now this script doesn't allow
	#   those.  (I'm not sure Netscape should either.)
	# Intrinsic event attributes may have character entities in them,
	#   even though any contained script macros cannot (in Netscape's
	#   implementation).  In this situation, the "real" way to handle this
	#   is to HTMLunescape and HTMLescape only those parts of intrinsic
	#   event attributes that are not inside script macros.  But for now,
	#   we'll escape them fully if they have no script macros, and not
	#   escape them at all if they do.  This can only go wrong with
	#   intrinsic event attributes that contain script macros, which are
	#   rare anyway, and even most of those will be handled correctly.  If
	#   this ever turns out to be a problem, we can revisit this.

	local($newtag, $atts, $attname, $v, $q) ;
	foreach (@body) {
	    next unless /<\s*\w+[^>]*[^\w.>\/?&-]on\w+\s*=/i
		     || /<\s*\w+[^>]*&{/ ;

	    ($newtag, $atts)= /^([^<]*<\s*\w+)((.|\n)*)/ ;
	    while ($atts!~ /^\s*>?$/) {

		# Peel one attribute at a time from the front of $atts.
		($attname, $dummy, $v, $atts)=
		    $atts=~ /^[^\w.-]*([\w.-]*)\s*(=\s*("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*))?((.|\n)*)/ ;

		# Pass through "valueless" attributes, e.g. "selected".
		$newtag.= ' ' . $attname, next  if $v eq '' ;

		# Attributes with script content may have quotes in them, so
		#   preserve the existing quotes to prevent conflicts.
		if    ($v=~ /^"/)  { $q= '"' ; $v=~ s/^"|"$//g }
		elsif ($v=~ /^'/)  { $q= "'" ; $v=~ s/^'|'$//g }
		else               { $q= '' }


		# First, replace any script macros.
		# Script macros end with "};", which would be easy with
		#   Perl 5's non-greedy matching, but Perl 4 requires hackery.
		if ($v=~ /&{/) {
		    local($newv, $start, $macro) ;
		    local($v2)= $v ;
		    while ($v2=~ /&{/) {
			($start,$v2)= split(/&{/, $v2, 2) ;
			($macro,$v2)= split(/};/, $v2, 2) ;
			$newv.= $start . '&{'
			      . &proxify_block($macro, $default_script_type)
			      . '};' ;
		    }
		    $v= $newv . $v2 ;
		}

		# Lastly, all intrinsic event attributes (i.e. those starting
		#   with "on") contain script content, so must be modified.
		# As described above, only handle escaping if there is no
		#   script macro inside.
		if ($attname=~ /^on/i) {
		    $v= $v=~ /&{/
			?  &proxify_block($v, $default_script_type)
			:  &HTMLescape(&proxify_block(&HTMLunescape($v),
						      $default_script_type)) ;
		}


		# Build the new tag, attribute by attribute.
		$newtag.= join('', ' ', $attname, '=', $q, $v, $q) ;

	    }  # while ($atts!~ /^\s*>?$/) {

	    $_= $newtag . '>' ;

	}  # foreach (@body)


	# Handle MSIE's "dynamic properties" in style attributes:  treat as
	#   script any string inside "expression()" or "function()".
	# Note that the regex effectively skips tags without a style attribute.
	# As far as I can tell, the language used for these dynamic properties
	#   is text/jscript (Microsoft's JavaScript variant).
	if ($ENV{'HTTP_USER_AGENT'}=~ /MSIE/) {
	    local($q) ;
	    foreach (@body) {
		s#(<\s*\w[^>]*[^\w.>\/?&-]style\s*=\s*)("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)#
		    local($start,$s)= ($1,$2) ;
		    if    ($s=~ /^"/)  { $q= '"' ; $s=~ s/^"|"$//g }
		    elsif ($s=~ /^'/)  { $q= "'" ; $s=~ s/^'|'$//g }
		    else               { $q= '' }

		    $s=~ s!expression\(([^)]*)\)!
			   'expression('
			 . &HTMLescape(&proxify_block(&HTMLunescape($1), 'text/jscript'))
			 . ')' !gie ;
		    $s=~ s!function\(([^)]*)\)!
			   'function('
			 . &HTMLescape(&proxify_block(&HTMLunescape($1), 'text/jscript'))
			 . ')' !gie ;

		    $start . $q . $s . $q ;
		#ie ;
	    }
	}


    }  # if (!$scripts_are_banned_here && $PROXIFY_SCRIPTS)



    # Handle <style>...</style> blocks.
    # For info about non-HTML blocks in an HTML document, see
    #   http://www.w3.org/TR/html40/appendix/notes.html#h-B.3.2
    # Note that script and style blocks are supposed to end with the first
    #   "</" string, but in fact browsers seem to end those blocks at the
    #   actual </script> or </style> tags.  This is most likely what the
    #   HTML author expects anyway, though it violates the HTML spec.  In
    #   this script, we should over-proxify rather than under-proxify, so
    #   we'll end those blocks on those end tags as browsers (erroneously) do.
    # Worse, Konqueror allows the string "</script>" inside JS literal strings,
    #   i.e. doesn't end the script block on them.  Netscape does end the block
    #   there, and both browsers end style blocks on embedded </style> strings.
    # The basic approach here is to extract a <style>...</style> block from
    #   @body, convert it, then insert it back into @body, being careful to
    #   keep $i set appropriately.
    # If a proxified block comes back '', then the element contents will be
    #   removed but the <style></style> tags will remain.  This will never
    #   be a problem as far as I know.
    local($i, $startblock, $style_type, $block, $newblock, @newblock) ;
    for ($i=0 ; $i<@body ; $i++) {
	next unless $body[$i]=~ /<\s*style\b/i ;
	$startblock= $i ;
	$i++ until ($body[$i]=~ m#</style\s*>#i) || ($i>=@body) ;

	($style_type)= $body[$startblock]=~
	    /<\s*style\b[^>]*[^\w.>\/?&-]type\s*=\s*["']?\s*([^\s"'>]*)/i ;
	$style_type=~ tr/A-Z/a-z/ ;
	$style_type= $default_style_type if $style_type eq '' ;

	$block= join('', @body[$startblock+1..$i]) ;
	$block=~ s#</style\s*>$##i ;
	$newblock= &proxify_block($block, $style_type) . '</style>' ;

	# Split into list; each element ends in ">" or EOstring.
	# This regex gives an ending '' element, so pop it.
	@newblock= $newblock=~ /([^>]*>?)/g ;
	pop(@newblock) if $newblock[$#newblock] eq '' ;

	splice(@body, $startblock+1, $i-$startblock, @newblock) ;

	# set list counter correctly; it will be incremented just after.
	$i= $startblock + @newblock ;
    }


    # Remove title if so configured.
    # For efficiency, only remove the first title, and stop looking after a
    #   </head> or <body> tag.  This works for all correct pages, and does not
    #   open any kind of privacy hole.
    if ($REMOVE_TITLES) {
	foreach (@body) {
	    s/^[^<>]*//, last  if /<\s*\/\s*title\b/i ;
	    last if m#<\s*(/head|body)\b#i ;
	}
    }


    # If we're doing an insertion, then find the position in @body where
    #   it should be inserted.  In order of precedence, that's:  after the
    #   <body> tag, or after the <html> tag, or at the start of the HTML.
    # While we're at it, try to get the title, so if it's a frame document
    #   we can display it in our own frame document.
    # Don't look inside <script> and <style> blocks.
    # If the page is a frame document, as indicated by the <frameset> tag,
    #   then reframe it in our own frame document if an insertion is required.
    # The variables $insert_pos and $is_frameset are set for later use.
    # Of course, this whole operation is skipped if we think this HTML
    #   resource is destined for a frame, as indicated by $is_in_frame.

    local($insert_pos, $is_frameset) ;
    if ($DOING_INSERT && !$is_in_frame) {
	local($body_pos, $html_pos, $title, $i) ;
	for $i (0..$#body) {
	    next if ( $body[$i]=~ /<\s*script\b/i .. $body[$i]=~ /<\s*\/script\b/i) ;
	    next if ( $body[$i]=~ /<\s*style\b/i .. $body[$i]=~ /<\s*\/style\b/i) ;
	    $body_pos= $i+1,  last  if $body[$i]=~ /<\s*body\b/i ;
	    $is_frameset= 1,  last  if $body[$i]=~ /<\s*frameset\b/i ;
	    $html_pos= $i+1         if $body[$i]=~ /<\s*html\b/i ;
	    $title= $1              if $body[$i]=~ /([^<>]*)<\s*\/\s*title\b/i ;
	}
	$insert_pos= $body_pos || $html_pos ;  # results in 0 if neither is set

	# If it's a frame document, then return another frame document that puts
	#   the insertion on top, and the frame doc in the lower subframe.
	# Note that if we're not doing an insertion, this never happens.
	# MUST be careful not to do this when $is_in_frame!  Else will recurse!
	&return_frame_doc(&proxy_encode($URL), $title) if $is_frameset ;

	# This block adds the insertion when it SHOULD BE anonymized, which
	#   is why it comes before the URL conversion below.  Note that the
	#   insertion has to be correctly split into tags for the URL
	#   conversion to work.
	# $INSERT_HTML takes precedence over $INSERT_FILE.  If $INSERT_HTML
	#   is empty, we set it to the file contents of $INSERT_FILE and
	#   use $INSERT_HTML hereafter.  This means the inserted file is only
	#   opened and read once, even under mod_perl.  :)
	if ( $ANONYMIZE_INSERTION && !$is_frameset &&
	     (($INSERT_HTML ne '') || ($INSERT_FILE ne '')) ) {
	    local(@insertion) ;
	    $INSERT_HTML= &readfile($INSERT_FILE) if $INSERT_HTML eq '' ;
	    @insertion= $INSERT_HTML=~ /([^>]*>?)/g ;  # split on ">" or EOS
	    pop(@insertion) if $insertion[$#insertion] eq '' ;  # pop final ''
	    splice(@body, $insert_pos, 0, @insertion) ;
	    $insert_pos+= @insertion  if $FORM_AFTER_INSERTION ;
	}
    }



    # This is a complete list of HTML tags/attributes that may include a URL,
    #   compiled from the list at Sandia (read on 9-17-96, but last updated on
    #   12-7-95), and the latest HTML DTD as of 9-17-96.  The Sandia list
    #   includes non-standard tags found to be used by Netscape or Microsoft.
    # List updated 7-31-98 by JSM to include all URL-type attributes defined
    #   in HTML 4.0.
    # If anyone knows of a well-maintained list of standard and non-standard
    #   tags/attributes with URLs in them, please let me know!!
    # Note that most of these are very rarely used, if ever.  They're included
    #   for safety, since we don't want an anonymous user accidentally
    #   revealing themselves because of a non-anonymized URL.
    # This regex favors speed over readability.
    # This could probably be more efficient, but the solution eludes me.
    # Note that this is not a perfect regex.  The substitutions can be fooled
    #   by, e.g., <input value="<a href=bad"> or <input value="src=bad">.
    #   It could be fixed with a slower regex.
    # The regex terminates URLs at spaces; URLs with spaces are disallowed,
    #   but fragments sometimes contain them anyway.
    # Denoting which of these are for images/binaries might be helpful, if
    #   we need more elaborate text-only support.
    # Err on the side of matching too much-- we'd rather a little mangled HTML
    #   than an anonymity hole.
    # 8-4-98 JSM:  Changed the regex to fix a bug: would only change the
    #   first URL in a tag.  New regex is probably slower.
    # 11-18-00 JSM:  Changed to handle attributes with leading spaces.
    # 1-13-01 JSM: Changed regex to match attribute boundaries better, because
    #   a simple \b is matching too much.  SGML attributes match \w[\w.-]*
    #   (except no underscores), so a boundary excludes those.  Add ">" too.
    #   We don't want to match words inside of URLs, so add "/" because of
    #   paths, and "?" and "&" because of query strings.  Thus the boundary
    #   match is [^\w.>\/?&-]... yeesh.  Regex now makes line longer than 80
    #   chars, bummer.  Time to use wider editor.

    foreach (@body) {

	# Extract tag name first-- saves ~40% time over old way, for this loop.
	# Also extract first non-blank after tag name-- if it's ">", then we
	#   can skip this too.
	($tag_name, $att_init)= /<\s*(\w+)\s*(.)/ ; # assumes tag names are \w+

	next if $tag_name eq '' ;    # e.g. for comments and end tags
	next if $att_init eq '>' ;   # when a tag has no attributes

	$tag_name=~ tr/A-Z/a-z/ ;


	# Change any style= attribute before testing for any tag, because it
	#   could exist in any tag.
	# Note that quotes must be preserved, to prevent quote conflicts.
	s#(<\s*\w[^>]*[^\w.>\/?&-]style\s*=\s*)("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)#
	    local($start,$s)= ($1,$2) ;
	    local($q) ;
	    if    ($s=~ /^"/)  { $q= '"' ; $s=~ s/^"|"$//g }
	    elsif ($s=~ /^'/)  { $q= "'" ; $s=~ s/^'|'$//g }

	    $start . $q . &proxify_block($s, $default_style_type) . $q ;
	#ie ;


	# The block that used to be here has been dispersed to the blocks of
	#   the individual tags it handled, regarding frame entering and
	#   leaving.  The tags are <a>, <area>, <link>, <form>, <frame>,
	#   and <iframe>.  The old comments immediately below are still true,
	#   even though the block has been spread to several places.
	# This block used to be the "if ($DOING_INSERT)" block.

	# Handle the specific cases where full_url_by_frame() should
	#   be called instead of full_url.  These include "going into
	#   frame mode" for <frame src="..."> tags, and "leaving frame
	#   mode" for <a target="_top"> or <a target="_blank">.
	# Everything else leaves the "frame mode" unchanged.

	# Note that the <base> tag may have a target attribute, which is
	#   then applied to any <a>, <link>, <area>, or <form> without a
	#   target attribute.  To support the <base> tag with frames and
	#   targets, we have to see if the <base> tag will cause untargeted
	#   links to exit frames (i.e. for our purposes, if the target is
	#   "_top" or "_blank").  If so, then we have to apply that target
	#   when updating links for those four tags.  We do so by setting
	#   $base_unframe above to indicate whether the <base> target causes
	#   an explicit exit from frames, and then checking it when updating
	#   those four tags, so that the frame flag in the URL is correct.
	# Not all frame exits will be caught.  :(  This is because when
	#   reading a target attribute, we don't know whether it leads to
	#   a new open window, or another frame in the existing window.  It
	#   could hypothetically be fixed *somewhat* by maintaining some list
	#   of which target names identify frames, based on the immediate
	#   browsing history (i.e. record the frame name when a <frame>
	#   tag is processed).  This would be rather elaborate, I think.
	# This only matters in that if a link causes the user to leave frames
	#   in a way we don't catch it, then we won't go back into frame mode
	#   if this isn't fixed.  This does NOT affect whether the user is
	#   still surfing through the proxy; it ONLY affects the display of
	#   the inserted HTML.
	# Apparently, Netscape only checks a matching prefix for _top and
	#   _blank, i.e. "_topxx" and "_blankxx" act like _top and _blank.
	#   IE works correctly.
	# This whole thing is a messy, messy fix for the rare case of when a
	#   <base> tag causes a frame to exit.  :P


	# Put the most common cases first


	# Handle <a> tag.  This only entails updating the href attribute, but
	#   that must include embedding the correct type code in the URL, and
	#   handling any deframing that might be happening.
	# BUT:  Browsers are inconsistent in whether a tag's "type" attribute
	#   takes precedence over actual Content-Type: header, or vice versa.
	#   It appears that for <link> tags, the type attribute (erroneously)
	#   always takes precedence, while for the <a> tag the type attribute
	#   is apparently ignored.  So to be consistent with browsers, we need
	#   to IGNORE the expected type code for the <a> tag.  In fact, we
	#   actually remove the type attribute altogether to remove a privacy
	#   hole in any browsers that *do* use it.  *sigh*  This wouldn't be a
	#   problem if <link>'s type attribute was handled correctly by
	#   browsers, i.e. of lower precedence than Content-Type:, but it's not.
	#   So the last part of http_get() gets hacked a little, which leads to
	#   this hack.  (Another solution would be to add yet another flag into
	#   the URL, a "linked-from-which-tag" flag.)
	# Since <a href> attribute sometimes has illegal quotes in it, such as
	#   when using JavaScript, we should check with the complete ''/""/\S*
	#   quoted string regex.  It's slower, so we use the ["']? shortcut for
	#   most other tags below.
	if ($tag_name eq 'a') {
	    # local($type)= m#<[^>]*[^\w.>\/?&-]type\s*=\s*["']?\s*([/\w.+\$-]+)#i ;
	    # $type=~ tr/A-Z/a-z/ ;

	    # Remove type attribute altogether.
	    s/(<\s*\w+[^>]*[^\w.>\/?&-])type\s*=\s*("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)/$1/i ;


	    # Unlike other tags with type attribute, don't remove <a> tag
	    #   if it links to banned content.  It will only be activated
	    #   by user action, not automatically like the others.

	    # If needed, detect if frame state might change.
	    # Deframe if (target unframes) or (no target and base target unframes)
	    # Testing $DOING_INSERT here saves time, but doesn't exit frame
	    #   state when nothing is being inserted (so presumably it doesn't
	    #   matter).  If accurate frame state is ever needed even when
	    #   $DOING_INSERT is false, then remove it from this boolean test.
	    # The same goes for the <area>, <form>, and <link> blocks below.
	    local($link_unframe) ;
	    if ($is_in_frame && $DOING_INSERT) {
		$link_unframe=
		       ( $base_unframe  &&  ! /<[^>]*[^\w.>\/?&-]target\s*=/i )
		    || /<[^>]*[^\w.>\/?&-]target\s*=\s*["']?\s*(_top|_blank)\b/i ;
	    }

	    # This temporary copy lets us call full_url() normally below.
	    # Only generate new value if the type flag or is_in_frame flag
	    #   has changed.
	    # ... except we don't test $type anymore.
	    local($this_url)= $this_url ;
#	    if ( ($type ne '') || $link_unframe ) {
	    if ($link_unframe) {
		$this_url= $script_url . '/' .
		    &pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS,
				$HIDE_REFERER, $INSERT_ENTRY_FORM,
				$link_unframe  ? 0  : $is_in_frame,
				'')     # would be "$type" here instead  :P
			. '/' ;
	    }

	    # Preserve quotes, to minimize conflicts with script constructs.
	    # Match even if final quote missing, because browsers accept that.
	    # In truth, all tag attributes should be HTML-unescaped before
	    #   conversion and re-escaped after.  That would be strictly
	    #   accurate.  But in the interest of running faster, that's
	    #   not done.  However, a few sites *do* (correctly) use
	    #   character entities inside attributes (e.g. "&" -> "&amp;"),
	    #   and may not work right without correct handling, especially
	    #   if proxy_encode()'ing is used.  So we use HTMLunescape() and
	    #   &HTMLescape() for the <a> tag, because that's where most of
	    #   the affected URLs will lie, e.g. links to queries with "&" in
	    #   them.
	    s#(<\s*\w[^>]*[^\w.>\/?&-]href\s*=\s*)("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)#
		local($start,$s)= ($1,$2) ;
		local($q) ;
		if    ($s=~ /^"/)  { $q= '"' ; $s=~ s/^"|"$//g }
		elsif ($s=~ /^'/)  { $q= "'" ; $s=~ s/^'|'$//g }

		$start . $q . &HTMLescape(&full_url(&HTMLunescape($s))) . $q ;
	    #ie ;

	    next ;
	}



	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	s/(<[^>]*[^\w.>\/?&-]lowsrc\s*=\s*["']?)\s*([^\s"'>]*)/   $1 . &full_url($2) /ie,
	s/(<[^>]*[^\w.>\/?&-]longdesc\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie,
	s/(<[^>]*[^\w.>\/?&-]usemap\s*=\s*["']?)\s*([^\s"'>]*)/   $1 . &full_url($2) /ie,
	s/(<[^>]*[^\w.>\/?&-]dynsrc\s*=\s*["']?)\s*([^\s"'>]*)/   $1 . &full_url($2) /ie,
	    next if $tag_name eq 'img' ;

	s/(<[^>]*[^\w.>\/?&-]background\s*=\s*["']?)\s*([^\s"'>]*)/ $1.&full_url($2) /ie,
	    next if $tag_name eq 'body' ;


	# if it's an issue later, we can set $base_unframe here too.
	s/(<[^>]*[^\w.>\/?&-]href\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie,
	    next if $tag_name eq 'base' ;     # has special significance


	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url_by_frame($2,1) /ie,
	s/(<[^>]*[^\w.>\/?&-]longdesc\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie,
	    next if $tag_name eq 'frame' ;


	# <form> tag needs special attention, here and elsewhere.
	# is <form script='...'> attribute ever used, or even recognized by
	#    any browser?  It's not defined in any W3C DTD.
	if ($tag_name eq 'form') {
	    # Deframe if (target unframes) or (no target and base target unframes)
	    if ( $DOING_INSERT
		    && (    ( $base_unframe  &&  ! /<[^>]*[^\w.>\/?&-]target\s*=/i )
			 || /<[^>]*[^\w.>\/?&-]target\s*=\s*["']?\s*(_top|_blank)\b/i  )  )  {
		s/(<[^>]*[^\w.>\/?&-]action\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url_by_frame($2,0) /ie ;
	    } else {
		s/(<[^>]*[^\w.>\/?&-]action\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie ;
	    }

	    if ($scripts_are_banned_here) {
		# Use strict quote checking to avoid <form script="decoy evil">
		s/(<[^>]*[^\w.>\/?&-])script\s*=\s*("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)/$1/i ;
	    } else {
		s/(<[^>]*[^\w.>\/?&-]script\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie ;
	    }
	    next ;
	}


	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	s/(<[^>]*[^\w.>\/?&-]usemap\s*=\s*["']?)\s*([^\s"'>]*)/   $1 . &full_url($2) /ie,
	    next if $tag_name eq 'input' ;


	# The only special handling for <area> is to handle any deframing that
	#   might be happening.
	if ($tag_name eq 'area') {
	    # Deframe if (target unframes) or (no target and base target unframes)
	    if ( $DOING_INSERT
		    && (    ( $base_unframe  &&  ! /<[^>]*[^\w.>\/?&-]target\s*=/i )
			 || /<[^>]*[^\w.>\/?&-]target\s*=\s*["']?\s*(_top|_blank)\b/i  )  )  {
		s/(<[^>]*[^\w.>\/?&-]href\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url_by_frame($2,0) /ie ;
	    } else {
		s/(<[^>]*[^\w.>\/?&-]href\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie ;
	    }
	    next ;
	}



	# <applet> tags are handled much like <object> tags; see the comments
	#   in that block to explain this block.  Also see
	#     http://www.w3.org/TR/html40/struct/objects.html#edef-APPLET
	# <applet> tags must convert the codebase, code, object, and archive
	#   attributes.
	# archive here is COMMA-separated list of URI's.
	# archive, code, and object are all relative to codebase, which
	#   may be relative to base URL.  Its default is the base URL.
	#   Note that values of codebase are not supposed to depart from
	#   dirs and subdirs of the base URL, because of security reasons,
	#   but some do anyway.  See the above URL for details.
	# This is untested with real applets.  It *has* been tested to ensure
	#   that test HTML tags are converted as intended, so any applet code
	#   that conforms to standards should work.

	if ($tag_name eq 'applet') {
	    # First, extract $base_url from codebase.
	    local($codebase_url)= &HTMLunescape($1)
		if m#<[^>]*[^\w.>\/?&-]codebase\s*=\s*["']?\s*([^\s"'>]*)#i ;

	    # Here is where we would guard against codebase leaving the dir.
	    #s/<[^>]*>?//  if     ( $codebase_url=~ /\.\./       )
	    #                  || ( $codebase_url=~ /^\//        )
	    #                  || ( $codebase_url=~ /^[\w+.-]*:/ )  ;

	    # if $codebase_url is relative, then make it absolute based on
	    #   current $base_ vars.  This is the quick method from full_url().
	    # Only do this if $codebase_url is not empty.
	    $codebase_url=
		  $codebase_url eq ''            ? $codebase_url
		: $codebase_url=~ m#^[\w+.-]*:#i ? $codebase_url
		: $codebase_url=~ m#^//#         ? $base_scheme . $codebase_url
		: $codebase_url=~ m#^/#          ? $base_host . $codebase_url
		:                                  $base_path . $codebase_url ;

	    # codebase must be converted with normal $base_ vars first,
	    #   although its URL needs to be extracted before that.
	    s/(<[^>]*[^\w.>\/?&-]codebase\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie ;

	    # Note that the right side is evaluated before the local() takes
	    #   effect, so we can use the old variables as defaults.
	    local($base_url, $base_scheme, $base_host, $base_path)=
		($base_url, $base_scheme, $base_host, $base_path) ;

	    # Now, set $base_url etc. accordingly.
	    $base_url= $codebase_url, &fix_base_vars  if $codebase_url ne '' ;

	    # These two can now be converted normally, with new $base_ vars.
	    s/(<[^>]*[^\w.>\/?&-]code\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie ;
	    s/(<[^>]*[^\w.>\/?&-]object\s*=\s*["']?)\s*([^\s"'>]*)/   $1 . &full_url($2) /ie ;

	    # archive is a comma-separated list of URIs: split, convert, join.
	    # When extracting attribute, note that it may contain spaces.
	    # This is done without map{}, for Perl 4.  Was:
	    # $1 . '"' . join(',', map {&full_url($_)} split(/\s*,\s*/, $+) ) . '"' #ie ;
	    s#(<[^>]*[^\w.>\/?&-]archive\s*=\s*)("([^">]*)"?|'([^'>]*)'?|([^'"][^\s>]*))#
		local($start)= $1 ;
		local(@arc, $arc) ;
		foreach $arc (split(/\s*,\s*/, $+)) { push(@arc, &full_url($arc)) }
		$start . '"' . join(',', @arc) . '"'
	    #ie ;

	    next ;
	}

#	s/(<[^>]*[^\w.>\/?&-]codebase\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie,
#	s/(<[^>]*[^\w.>\/?&-]code\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie,
#	s/(<[^>]*[^\w.>\/?&-]object\s*=\s*["']?)\s*([^\s"'>]*)/   $1 . &full_url($2) /ie,
#	s/(<[^>]*[^\w.>\/?&-]archive\s*=\s*["']?)\s*([^\s"'>]*)/  $1 . &full_url($2) /ie,
#	    next if $tag_name eq 'applet' ;



	# These are seldom-used tags, or tags that seldom have URLs in them

	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	    next if $tag_name eq 'bgsound' ;  # Microsoft only

	s/(<[^>]*[^\w.>\/?&-]cite\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie,
	    next if $tag_name eq 'blockquote' ;

	s/(<[^>]*[^\w.>\/?&-]cite\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie,
	    next if $tag_name eq 'del' ;

	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	    next if $tag_name eq 'embed' ;    # Netscape only

	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	s/(<[^>]*[^\w.>\/?&-]imagemap\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie,
	    next if $tag_name eq 'fig' ;      # HTML 3.0

	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	    next if $tag_name=~ /^h[1-6]$/ ;  # HTML 3.0


	# <head>'s profile attribute can be a space-separated list of URIs.
	if ($tag_name eq 'head') {
	    # This is done without map{}, for Perl 4.
	    s#(<[^>]*[^\w.>\/?&-]profile\s*=\s*)("([^">]*)"?|'([^'>]*)'?|([^'"][^\s>]*))#
		local($start)= $1 ;
		local(@prof, $prof) ;
		foreach $prof (split(" ", $+)) { push(@prof, &full_url($prof)) }
		$start . '"' . join(' ', @prof) . '"'
	    #ie ;
	    next ;
	}


	s/(<[^>]*[^\w.>\/?&-]profile\s*=\s*["']?)\s*([^\s"'>]*)/  $1 . &full_url($2) /ie,
	    next if $tag_name eq 'head' ;

	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	    next if $tag_name eq 'hr' ;       # HTML 3.0

	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url_by_frame($2,1) /ie,
	s/(<[^>]*[^\w.>\/?&-]longdesc\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie,
	    next if $tag_name eq 'iframe' ;

	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	    next if $tag_name eq 'ilayer' ;

	s/(<[^>]*[^\w.>\/?&-]cite\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie,
	    next if $tag_name eq 'ins' ;

	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	    next if $tag_name eq 'layer' ;


	# Handle <link> tag.  The extra requirement here is to check the "type"
	#   attribute and include that in a temporary $this_url when the
	#   other attributes are converted.  This allows e.g. external style
	#   sheets to be handled correctly when they're downloaded later.
	#   Also, the <link> tag is removed altogether if scripts are banned
	#   and "type" matches a script type.  (Note that the type attribute
	#   indicates an *advisory* *expected* MIME type, not a required type,
	#   though some browsers seem to treat it erroneously as the ultimate
	#   authority.)
	# An additional task is to handle the frame state correctly, i.e.
	#   make sure set the frame flag to 0 when the tag causes an exit
	#   from frame state.  (See the "if ($DOING_INSERT)" block above.)
	# In Netscape, the Content-Style-Type: header has no effect in the
	#   interpretation of external style sheets.  This is probably correct.
	#   Thus, $default_style_type is not used here.
	# See http://www.w3.org/TR/html40/struct/links.html#edef-LINK  and
	#     http://www.w3.org/TR/html40/types.html#type-links
	if ($tag_name eq 'link') {
	    local($type)=
		m#<[^>]*[^\w.>\/?&-]type\s*=\s*["']?\s*([/\w.+\$-]+)#i ;
	    $type=~ tr/A-Z/a-z/ ;

	    # When a type attribute is not given, some browsers erroneously
	    #   use a default type of "text/css" for any <link> tag indicating
	    #   a stylesheet, even to the point of overriding a subsequent
	    #   Content-Type: header.  So set that default type here if it's
	    #   a stylesheet, as indicated by the rel attribute.  Note that the
	    #   rel attribute may contain spaces.
	    if ($type eq '') {
		local($rel) ;
		$rel= $+
		    if /<[^>]*[^\w.>\/?&-]rel\s*=\s*("([^">]*)"?|'([^'>]*)'?|([^'"][^\s>]*))/i ;
		$type= 'text/css' if $rel=~ /\bstylesheet\b/i ;
	    }

	    # Remove tag if it links to a script type and scripts are banned.
	    if ($scripts_are_banned_here && $type ne '') {
		s/<[^>]*>?// , next  if $type=~ /^$SCRIPT_TYPE_REGEX$/io ;
	    }

	    # If needed, detect if frame state might change.
	    # Deframe if (target unframes) or (no target and base target unframes)
	    local($link_unframe) ;
	    if ($is_in_frame && $DOING_INSERT) {
		$link_unframe=
		       ( $base_unframe  &&  ! /<[^>]*[^\w.>\/?&-]target\s*=/i )
		    || /<[^>]*[^\w.>\/?&-]target\s*=\s*["']?\s*(_top|_blank)\b/i ;
	    }

	    # This temporary copy lets us call full_url() normally below.
	    # Only generate new value if the type flag or is_in_frame flag
	    #   has changed.
	    local($this_url)= $this_url ;
	    if ( ($type ne '') || $link_unframe ) {
		$this_url= $script_url . '/' .
		    &pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS,
				$HIDE_REFERER, $INSERT_ENTRY_FORM,
				$link_unframe  ? 0  : $is_in_frame,
				$type)
			. '/' ;
	    }

	    s/(<[^>]*[^\w.>\/?&-]href\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie ;
	    s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/  $1 . &full_url($2) /ie ;   # Netscape?
	    s/(<[^>]*[^\w.>\/?&-]urn\s*=\s*["']?)\s*([^\s"'>]*)/  $1 . &full_url($2) /ie ;
	    next ;
	}


	# <meta http-equiv> is a special case-- these must be handled like real
	#   HTTP headers.
	# Remove http-equiv attribute if content is empty, else may generate
	#   empty cookie.
	# The Netscape-only "url" attribute can be handled normally.
	# Bug in Perl 4: using $new_value in "content" regex below makes it
	#   become defined.  Thus, we need to use its definedness before that.
	# Need ";" covered when checking for "url=", because it could be in
	#   e.g. <meta http-equiv="refresh" content="5;URL=...">.
	# Therefore, ugh, we need to handle if it's instead 'content=5; URL=',
	#   i.e. has a space before "URL".  Hack here is to remove those
	#   spaces.
	# Since nonstandard headers like Link: and URI: may contain "<>"  :P ,
	#   we need to HTMLunescape() and HTMLescape() here.
	if ($tag_name eq 'meta') {
	    s/(<[^>]*;)\s*url\s*=/ $1 . 'url=' /ie ;    # hack here
	    s/(<[^>]*[^\w.>\/?&;-]url\s*=\s*["']?)\s*([^\s"'>]*)/  $1 . &full_url($2) /ie ;
	    local($name, $value, $new_value) ;
	    next unless ( ($name)=  /<[^>]*[^\w.>\/?&-]http-equiv\s*=\s*["']?\s*([\w.-]*)/i)
		     && ( ($value)= /<[^>]*[^\w.>\/?&-]content\s*=\s*("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)/i ) ;
	    $value=~ s/^["']|["']$//g ;   # strip quotes

	    $new_value=
		&HTMLescape(&new_header_value($name, &HTMLunescape($value))) ;
	    s/(<[^>]*[^\w.>\/?&-]http-equiv\s*=\s*["']?)\s*([^\s"'>]*)/ $1 /ie
		unless defined($new_value) ;
	    s/(<[^>]*[^\w.>\/?&-]content\s*=)\s*("[^">]*"?|'[^'>]*'?|[^'"][^\s>]*)/
	      qq($1"$new_value") /ie ;

	    next ;
	}


	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	    next if $tag_name eq 'note' ;     # HTML 3.0



	# <object> tags need special treatment, particularly regarding which
	#   attributes use which base URIs to resolve them when relative.
	#   For details, see
	#     http://www.w3.org/TR/html40/struct/objects.html#edef-OBJECT
	# The <object> tag has five attributes that may contain URLs to be
	#   converted:  usemap, codebase, classid, data, archive.
	# codebase is used as the base URL for classid, data, and archive;
	#   it defaults to normal base URI.
	# In any case, we must convert those last three to absolute URLs,
	#   because if they're relative there's no way a browser could resolve
	#   them through the proxy.  But they must be resolved using the
	#   codebase URL.  Since full_url() uses the three $base_ variables,
	#   we use the trick here of using local() to temporarily mask those
	#   three, plus $base_url, if needed.
	# Note that data and classid depend on MIME types in type and codetype,
	#   so the types must be included in updated URLs.  We do this with
	#   a local copy of $this_url.
	# We indicate (via full_url_by_frame() or flag in the local $this_url)
	#   that the object is in a frame, so that any embedded objects won't
	#   have the URL entry form inserted (it's legal to embed HTML pages
	#   using the <object> tag).
	# There are a few sites with embedded objects or applets that may not
	#   work if the routine proxy_encode() is not the simple default.  See
	#   the comments above that routine for details.  It has to do with
	#   the objects trying to resolve relative URLs.

	if ($tag_name eq 'object') {
	    local($type)=
		m#<[^>]*[^\w.>\/?&-]type\s*=\s*["']?\s*([/\w.+\$-]+)#i ;
	    local($codetype)=
		m#<[^>]*[^\w.>\/?&-]codetype\s*=\s*["']?\s*([/\w.+\$-]+)#i ;
	    $type=~ tr/A-Z/a-z/ ;
	    $codetype=~ tr/A-Z/a-z/ ;

	    # Remove tag if it links to a script type and scripts are banned.
	    # Check both type and codetype.
	    if ($scripts_are_banned_here) {
		s/<[^>]*>?// , next
		    if     $type=~     /^$SCRIPT_TYPE_REGEX$/io
			|| $codetype=~ /^$SCRIPT_TYPE_REGEX$/io ;
	    }

	    # usemap is the only attribute converted normally.
	    s/(<[^>]*[^\w.>\/?&-]usemap\s*=\s*["']?)\s*([^\s"'>]*)/   $1 . &full_url($2) /ie ;

	    # First, extract $base_url from codebase.
	    local($codebase_url)= &HTMLunescape($1)
		if m#<[^>]*[^\w.>\/?&-]codebase\s*=\s*["']?\s*([^\s"'>]*)#i ;

	    # if $codebase_url is relative, then make it absolute based on
	    #   current $base_ vars.  This is the quick method from full_url().
	    # Only do this if $codebase_url is not empty.
	    $codebase_url=
		  $codebase_url eq ''            ? $codebase_url
		: $codebase_url=~ m#^[\w+.-]*:#i ? $codebase_url
		: $codebase_url=~ m#^//#         ? $base_scheme . $codebase_url
		: $codebase_url=~ m#^/#          ? $base_host . $codebase_url
		:                                  $base_path . $codebase_url ;

	    # codebase must be converted with normal $base_ vars first,
	    #   although its URL needs to be extracted before that.
	    s/(<[^>]*[^\w.>\/?&-]codebase\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url_by_frame($2,1) /ie ;

	    # Note that the right side is evaluated before the local() takes
	    #   effect, so we can use the old variables as defaults.
	    local($base_url, $base_scheme, $base_host, $base_path)=
		($base_url, $base_scheme, $base_host, $base_path) ;

	    # Now, set $base_url etc. accordingly.
	    $base_url= $codebase_url, &fix_base_vars  if $codebase_url ne '' ;


	    # archive is a space-separated list of URIs: split, convert, join.
	    # When extracting attribute, note that it may contain spaces.
	    # Do this before changing $this_url for data and classid handling.
	    # This is done without map{}, for Perl 4.  Was:
	    # $1 . '"' . join(' ', map {&full_url_by_frame($_,1)} split(" ", $+) ) . '"' /ie ;
	    s#(<[^>]*[^\w.>\/?&-]archive\s*=\s*)("([^">]*)"?|'([^'>]*)'?|([^'"][^\s>]*))#
		local($start)= $1 ;
		local(@arc, $arc) ;
		foreach $arc (split(" ", $+)) { push(@arc, &full_url_by_frame($arc,1)) }
		$start . '"' . join(' ', @arc) . '"'
	    #ie ;


	    # data and classid attributes can now be converted normally, with
	    #   new $base_ vars and possibly updated $this_url's.
	    # data uses a $this_url based on type; classid uses a $this_url
	    #   based on codetype if it exists, else type.

	    # Use a local copy of $this_url to call full_url() normally.
	    # Note that $is_in_frame is set to 1 anyway, so go ahead and
	    #   generate a new $this_url regardless.
	    local($this_url)= $script_url . '/' .
		&pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS,
			    $HIDE_REFERER, $INSERT_ENTRY_FORM, 1, $type)
		. '/' ;
	    s/(<[^>]*[^\w.>\/?&-]data\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie ;

	    # Generate another $this_url for classid if necessary.
	    if (($codetype ne '') && ($codetype ne $type)) {
		$this_url= $script_url . '/' .
		    &pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS,
				$HIDE_REFERER, $INSERT_ENTRY_FORM, 1, $codetype)
			. '/' ;
	    }

	    # Special case:  Don't convert classid if it begins with "clsid:".
	    #   "clsid:" is a non-standard URL scheme that indicates the "UUID"
	    #   of an object, and is sometimes used with embedded objects like
	    #   Flash, etc.  It is described in a 1996 draft at
	    #   http://www.w3.org/Addressing/clsid-scheme .
	    s{(<[^>]*[^\w.>\/?&-]classid\s*=\s*["']?)\s*([^\s"'>]*)}{
		local($start,$att)= ($1,$2) ;  # avoid trouble with next regex
		$att=~ /^clsid:/i  ? ( $start . $att )
				   : ( $start . &full_url($att) ) ;
	    }ie ;

	    next ;
	}

#	s/(<[^>]*[^\w.>\/?&-]usemap\s*=\s*["']?)\s*([^\s"'>]*)/   $1 . &full_url($2) /ie,
#	s/(<[^>]*[^\w.>\/?&-]codebase\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie,
#	s/(<[^>]*[^\w.>\/?&-]data\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie,
#	s/(<[^>]*[^\w.>\/?&-]archive\s*=\s*["']?)\s*([^\s"'>]*)/  $1 . &full_url($2) /ie,
#	s/(<[^>]*[^\w.>\/?&-]classid\s*=\s*["']?)\s*([^\s"'>]*)/  $1 . &full_url($2) /ie,
#	s/(<[^>]*[^\w.>\/?&-]name\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie,
#	    next if $tag_name eq 'object' ;


	# The <param> tag is special-- if its valuetype attribute is "ref",
	#   then the value attribute is a URI.  Also, in this case it has a
	#   type attribute which indicates an expected MIME type.
	# In http://www.w3.org/TR/html40/struct/objects.html#edef-PARAM ,
	#   we're told not to resolve the value URI; however, not doing so
	#   could open a privacy hole, normally only when it's an absolute
	#   URI.  So based on our priorities, we update the value URI here
	#   if it's an absolute URI.
	if ($tag_name eq 'param') {
	    local($valuetype)=
		m#<[^>]*[^\w.>\/?&-]valuetype\s*=\s*["']?\s*(\w+)#i ;
	    next unless $valuetype=~ /^ref$/i ;

	    local($type)=
		m#<[^>]*[^\w.>\/?&-]type\s*=\s*["']?\s*([/\w.+\$-]+)#i ;
	    $type=~ tr/A-Z/a-z/ ;

	    # Remove tag if it links to a script type and scripts are banned.
	    if ($scripts_are_banned_here && $type ne '') {
		s/<[^>]*>?// , next  if $type=~ /^$SCRIPT_TYPE_REGEX$/io ;
	    }

	    # Use a local copy of $this_url to call full_url() normally.
	    # Only generate new $this_url if the type flag has changed.
	    local($this_url)= $this_url ;
	    if ($type ne '') {
		$this_url= $script_url . '/' .
		    &pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS,
				$HIDE_REFERER, $INSERT_ENTRY_FORM,
				$is_in_frame, $type)
			. '/' ;
	    }
	    s{(<[^>]*[^\w.>\/?&-]value\s*=\s*["']?)\s*([^\s"'>]*)}{
		local($start,$att)= ($1,$2) ;  # avoid trouble with next regex
		$att=~ /^[\w.+-]+:/i  ? ( $start . &full_url($att) )
				      : ( $start . $att )  ;
	    }ie ;


	    next ;
	}



	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	s/(<[^>]*[^\w.>\/?&-]imagemap\s*=\s*["']?)\s*([^\s"'>]*)/ $1 . &full_url($2) /ie,
	    next if $tag_name eq 'overlay' ;  # HTML 3.0

	s/(<[^>]*[^\w.>\/?&-]cite\s*=\s*["']?)\s*([^\s"'>]*)/     $1 . &full_url($2) /ie,
	    next if $tag_name eq 'q' ;



	# Netscape apparently trusts expected type here, including the default
	#   to JavaScript, and completely ignores the Content-Type: header.
	#   The expected type first comes from "type" attribute, else "language"
	#   attribute, else default.
	# Konqueror, on the other hand, treats all external scripts as
	#   JavaScript, regardless of either expected type or Content-Type: .
	# Set the expected type in the URL flags, and the resource will be
	#   interpreted appropriately when downloaded in e.g. http_get().
	if ($tag_name eq 'script') {
	    # Can "for" even exist in <script>, or was that an old mistake?
	    s/(<[^>]*[^\w.>\/?&-]for\s*=\s*["']?)\s*([^\s"'>]*)/  $1 . &full_url($2) /ie ;

	    local($type)=
		m#<[^>]*[^\w.>\/?&-]type\s*=\s*["']?\s*([/\w.+\$-]+)#i ;
	    $type=~ tr/A-Z/a-z/ ;

	    # If there's no type, but there's a language attribute, then use
	    #   it instead to guess the expected type.
	    if (  ($type eq '') &&
		  (m#<[^>]*[^\w.>\/?&-]language\s*=\s*["']?\s*([/\w.+\$-]+)#i ) )  {
		local($language)= $1 ;
		$type=  $language=~ /javascript|ecmascript|livescript|jscript/i
						  ? 'application/x-javascript'
		      : $language=~ /css/i        ? 'text/css'
		      : $language=~ /vbscript/i   ? 'application/x-vbscript'
		      : $language=~ /perl/i       ? 'application/x-perlscript'
		      : $language=~ /tcl/i        ? 'text/tcl'
		      :                             ''
	    }

	    # This would already have been removed if $scripts_are_banned_here,
	    #   so no need to do it here.

	    # Use a local copy of $this_url to call full_url() normally.
	    # Only generate new $this_url if the type flag has changed.
	    local($this_url)= $this_url ;
	    if ($type ne '') {
		$this_url= $script_url . '/' .
		    &pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS,
				$HIDE_REFERER, $INSERT_ENTRY_FORM,
				$is_in_frame, $type)
			. '/' ;
	    }

	    s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/  $1 . &full_url($2) /ie ;
	    next ;
	}



	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	    next if $tag_name eq 'select' ;   # HTML 3.0

	s/(<[^>]*[^\w.>\/?&-]src\s*=\s*["']?)\s*([^\s"'>]*)/      $1 . &full_url($2) /ie,
	    next if $tag_name eq 'ul' ;       # HTML 3.0

	s/(<[^>]*[^\w.>\/?&-]background\s*=\s*["']?)\s*([^\s"'>]*)/ $1.&full_url($2) /ie,
	    next if $tag_name eq 'td' ;       # Netscape extension?

	s/(<[^>]*[^\w.>\/?&-]background\s*=\s*["']?)\s*([^\s"'>]*)/ $1.&full_url($2) /ie,
	    next if $tag_name eq 'th' ;       # Netscape extension?

	s/(<[^>]*[^\w.>\/?&-]background\s*=\s*["']?)\s*([^\s"'>]*)/ $1.&full_url($2) /ie,
	    next if $tag_name eq 'tr' ;       # Netscape extension?

	s/(<[^>]*[^\w.>\/?&-]background\s*=\s*["']?)\s*([^\s"'>]*)/ $1.&full_url($2) /ie,
	    next if $tag_name eq 'table' ;    # Netscape extension?

    }   # foreach (@body)



    # Insert anything needed-- entry form and/or custom insertion.
    # This relies on the variables $insert_pos and $is_frameset, which were
    #   set in the block above the foreach(@body) block directly above
    #   here.  The comments in that block apply here too.
    if (!$is_in_frame && !$is_frameset) {

	# Insert the custom insertion if needed.
	# This block adds it when it SHOULD NOT BE anonymized, which is why
	#   it comes after the URL conversion above.
	if ( !$ANONYMIZE_INSERTION &&
	     (($INSERT_HTML ne '') || ($INSERT_FILE ne '')) ) {
	    $INSERT_HTML= &readfile($INSERT_FILE) if $INSERT_HTML eq '' ;
	    splice(@body, $insert_pos, 0, $INSERT_HTML) ;
	    $insert_pos++  if $FORM_AFTER_INSERTION ;
	}

	# Insert the entry form if needed-- never anonymized.
	splice(@body, $insert_pos, 0, &mini_start_form($URL,0))
	    if $INSERT_ENTRY_FORM ;
    }



    # Does this ever cause a problem, putting a comment before the entire body?
    #   It may affect the initial <!DOCTYPE ...> tag, etc.
    unshift(@body, "<!-- resource has been modified by proxy -->\n" ) ;

    # Change Content-Length header, since we changed the content
    $headers=~ s/^Content-Length:.*\012/
		 'Content-Length: ' . length(join('',@body)) . "\015\012"/ie ;

}


#--------------------------------------------------------------------------
#    Send response back to user
#--------------------------------------------------------------------------

# If the response has not already been sent, print the status line, headers,
#   and the entire (possibly modified) resource.
# The $response_sent flag was added purely to support streaming media and
#   large files.
if (!$response_sent) {
    if ($ENV{'REQUEST_METHOD'} eq 'HEAD') {
	print $status, $headers ;
    } elsif ($is_html) {
	print $status, $headers, @body ;
	# print $debug ;   # handy for sprinkling checks throughout the code
    } else {
	print $status, $headers, $body ;
    }
}


# Put this back in to run speed trials
#if ($is_html) {
#    # OK, let's time this thing
#    ($eutime,$estime)= (times)[0,1] ;
#    open(LOG,">>proxy.log") ;
#    print LOG "full times: ", $eutime-$sutime, " ", $estime-$sstime,
#        " ", time-$starttime, "  URL: $URL\n" ;
#    close(LOG) ;
#}



# mod_perl scripts must not exit.
EXIT:
exit unless $ENV{'MOD_PERL'} ;

#--------------------------------------------------------------------------
#   DONE!!
#--------------------------------------------------------------------------


# Returns the full absolute URL to query our script for the given URI
#   reference.  PATH_INFO will include the encoded absolute URL of the target,
#   but the fragment will be appended unencoded so browsers will resolve it
#   correctly.
# This is a major bottleneck for the whole program, so speed is important here.
# Note that the calculations of $this_url, $base_scheme, $base_host, and
#   $base_path throughout the program are an integral part of this routine,
#   placed elsewhere for speed.
# This was rewritten 2-18-99 to be much more flexible, e.g. to allow encoding
#   and avoid illegal PATH_INFO sequences.  Unfortunately, it is a little
#   slower.
# For HTTP, The URL to be encoded should include everything that is sent in
#   the request, including any query, but not any fragment.
# This only returns absolute URLs, though relative URLs would usually suffice.
#   If it matters, we could have a fullrelurl() and fullabsurl(), the latter
#   used for those HTML attributes that require an absolute URL (like <base>).
#
# The ?:?:?: statement resolves relative URLs to absolute URLs, given the
#   $base_{url,scheme,host,path} variables figured earlier.  It does it
#   simply and efficiently, and accurately enough; the full procedure is
#   described in RFC 2396 (URI syntax), section 5.2.
# RFC 2396, section 5 states that there are three types of relative URIs:
#   net_path (beginning with //, rarely used), abs_path (beginning with /),
#   and rel_path, any of which may be followed by a "?query"; the query must
#   be included in the result.  Thus, we only need to examine the start of
#   the relative URL.
# This ?:?:?: statement passes all test cases in RFC 2396 appendix C, except
#   for the following:  It does not reduce . and .. path segments (to do
#   so would take a lot more time), and it assumes $uri_ref has something
#   other than an empty fragment in it, i.e. that the URI is non-empty.
# This only works for hierarchical schemes, like HTTP or FTP.  Conceivably,
#   there's a problem if the base URL uses a non-hierarchical scheme, and
#   the document contains relative URLs.  Absolute URLs will be OK.
# Any HTML-escaping/unescaping should be done outside of this routine, since
#   it is used for any relative->absolute URL conversion, not just HTML.

sub full_url {
    local($uri_ref)= @_ ;
    local($1,$2) ;   # Required for Perl 4-- else, caller's $1,$2 get clobbered

    $uri_ref=~ s/^\s+// ;  # remove leading spaces

    # For now, prevent redirecting into x-proxy URLs.
    # This slows down the main tag-converting loop by 0-1%.
    return undef if $uri_ref=~ m#^x-proxy://#i ;

    # Handle "javascript:" URLs separately.
    if ($uri_ref=~ /^javascript:/i) {
	return undef if $scripts_are_banned_here ;
	return $uri_ref unless $PROXIFY_SCRIPTS ;
	local($script)= $uri_ref=~ /^javascript:((.|\n)*)$/i ;
	return 'javascript:' . &proxify_block($script, 'application/x-javascript') ;
    }	

    # Separate fragment from URI
    local($uri,$frag)= $uri_ref=~ /^([^#]*)(#.*)?/ ;
    return $uri_ref if $uri eq '' ;  # allow bare fragments to pass unchanged

    # calculate absolute URL based on four possible cases
    local($absurl)=
	    $uri=~ m#^[\w+.-]*:#i   ?  $uri                 # absolute URL
	  : $uri=~ m#^//#           ?  $base_scheme . $uri  # net_path (rare)
	  : $uri=~ m#^/#            ?  $base_host . $uri    # abs_path, rel URL
	  :                            $base_path . $uri ;  # relative path

    return join('', $this_url, &proxy_encode($absurl), $frag ) ;
}


# Identical to full_url(), except second parameter explicitly determines
#   whether we use $this_url_inframe or $this_url_noframe.
# This could be wrapped into the full_url() routine, but I'm guessing it
#   is more efficient to do it this way.  This won't be called often and
#   full_url() is called a lot.
# This uses a little trick with local() that lets us use full_url(), which
#   keeps the routines synchronized and reduces code size.  We set a local
#   version of $this_url, which is used by full_url() because it remains
#   in scope there, but when we exit this routine the scope closes and
#   the old $this_url is restored.
sub full_url_by_frame {
    local($uri_ref, $is_frame)= @_ ;
    local($this_url)= $is_frame   ? $this_url_inframe  : $this_url_noframe ;
    return &full_url($uri_ref) ;
}


# Set globals $base_url, $base_scheme, $base_host, and $base_path, based on
#   value of $base_url.  This must be called whenever $base_url is set, which
#   unfortunately may vary over the course of the program.
# These are an integral part of &full_url(), placed outside of that for speed.
# To specify:
#   $base_scheme is the scheme of the base URL, ending in ":", like "http:".
#   $base_host is the scheme/host/port of the base URL, with no final slash.
#   $base_path is the scheme/host/port/path, through final slash.
# These are only relevant (and accurate) for hierarchical "/"-using schemes,
#   like HTTP or FTP.
# Any HTML-escaping/unescaping should be done outside of this routine.
sub fix_base_vars {
    # Guarantee that $base_url has at least a path of '/', inserting before
    #   ?query if needed.
    $base_url=~ s#^([\w+.-]+://[^/?]+)/?#$1/# ;

    ($base_scheme)= $base_url=~ m#^([\w+.-]+:)//# ;
    ($base_host)=   $base_url=~ m#^([\w+.-]+://[^/?]+)# ; # no ending slash
    ($base_path)=   $base_url=~ m#^([^?]*/)# ;            # use greedy matching
}



# Given a block of code, convert it to be "proxy-safe", depending on
#   the given content type (language).  Usually that conversion just means
#   updating any URLs in it.
# This is used for style sheets, (potentially) scripts, etc.
# Preserve correct quotes.
sub proxify_block {
    local($s, $type)= @_ ;

    if ($scripts_are_banned_here) {
	return undef if $type=~ /^$SCRIPT_TYPE_REGEX$/io ;
    }

    if ($type eq 'text/css') {
	# The only URIs in CSS2 are invoked with "url(...)".
	# Ugly regex, but gets virtually all real matches and is privacy-safe.
	# Hard part is handling "\"-escaping.  See
	#   http://www.w3.org/TR/REC-CSS2/syndata.html#uri
	# Hopefully we'll use a whole different approach in the new rewrite.

	$s=~ s/url\s*\(\s*(([^)]*\\\))*[^)]*)(\)|$)/
	      'url(' . &css_full_url($1) . ')'    /gie ;

	return $s ;


    # JavaScript can be identified by any of these MIME types.  :P  The
    #   "ecma" ones are the standard, the "javascript" and "livescript" ones
    #   refer to Netscape's implementations, and the "jscript" one refers to
    #   Microsoft's implementation.  Until we need to differentiate, let's
    #   treat them all the same here.
    # If you implement your own JavaScript handling, please see the notes above
    #   where $PROXIFY_SCRIPTS is set.  In short: you may get a lot of it to
    #   work correctly through the proxy, but you will not likely be able to
    #   anonymize it all reliably.
    } elsif ($type=~ m#^(application/x-javascript|application/x-ecmascript|application/javascript|application/ecmascript|text/javascript|text/ecmascript|text/livescript|text/jscript)$#i) {

	# Here's where to make any desired changes to JavaScript code.
	# For example, the next line (imperfectly) removes most window.open()
	#   commands, thus removing most popups.  A better version would
	#   account for parentheses inside quoted parameters, etc.

	#$s=~ s/window\.open\s*\([^)]*\)//g ;

	# You could instead "proxify" some window.open commands by updating
	#   any literal URLs inside them with full_url(), but that only works
	#   if the URLs are literal strings.  To get the others, you'd need to
	#   write full_url() and proxy_encode() routines in JavaScript itself,
	#   and then include those with the returned page.  And then what about
	#   eval() statements?  Or all the document.write() statements, which
	#   can write any HTML... you see the challenge.
	# But you can play with this to handle your specific needs, and cover
	#   virtually all cases you actually deal with.
	# If you decide to go for anonymity, consider removing all eval()
	#   statements and living with the minor lost functionality.


	return $s ;


    } else {
	# If we don't understand the type, return the block unchanged.
	# This would be a privacy hole, if we didn't check for script types
	#   when $scripts_are_banned_here above.  If later we want the option
	#   of returning undef for an unknown type, we can add a parameter to
	#   specify that.

	return $s ;

    }

}



# For CSS only:  takes entire contents between parentheses in "url(...)",
#   extracts the URL therein (accounting for quotes, "\"-escaped chars, etc.),
#   and returns the full_url() of that, suitable for placing back inside
#   "url(...)", including all "\"-escaping, quotes, etc.  :P
# Preserve correct quotes, because this may be embedded in a larger quoted
#   context.
# In external style sheets, relative URLs are resolved relative to the style
#   sheet, not the source HTML document.  This makes it easy for us-- no
#   special $base_url handling.
sub css_full_url {
    local($url)= @_ ;
    local($q) ;

    $url=~ s/\s*$// ;       # leading spaces already stripped above
    if    ($url=~ /^"/)  { $q= '"' ; $url=~ s/^"|"$//g }  # strip quotes
    elsif ($url=~ /^'/)  { $q= "'" ; $url=~ s/^'|'$//g }
    $url=~ s/\\(.)/$1/g ;   # "\"-unescape
    $url=~ s/^\s*|\s*$//g ; # finally, strip spaces once more

    $url= &full_url($url) ;

    $url=~ s/([(),\s'"\\])/\\$1/g ;    # put "\"-escaping back in

    return $q . $url . $q ;
}


#--------------------------------------------------------------------------
#    Scheme-specific routines
#--------------------------------------------------------------------------

#
# <scheme>_get: get resource at URL and set globals $status, $headers,
#   $is_html, and either $body or @body depending on $is_html.  Optionally,
#   set $response_sent to signal that the response has already been sent.
#   These are all globals for speed, to prevent unneeded copying of huge
#   strings.
#

# http_get: actually supports both GET and POST

sub http_get {
    local($portst, $realhost, $realport, $request_uri, $content_type,
	  $realm, $tried_realm, $auth,
	  $proxy_auth_header, $referer_header, $connection_header,
	  $lefttoget, $postblock, @postbody, $body_too_big, $rin,
	  $first_block) ;
    local($/)= "\012" ;

    $port= 80 if $port eq '' ;

    # Some servers don't like default port in a Host: header, so use $portst.
    $portst= ($port==80)  ? ''  : ":$port" ;

    $realhost= $host ;
    $realport= $port ;
    $request_uri= $path ;

    # there must be a smoother way to handle proxies....
    if ($ENV{'http_proxy'}) {
	local($dont_proxy) ;
	foreach (split(/\s*,\s*/, $ENV{'no_proxy'})) {
	    $dont_proxy= 1, last if $host=~ /$_$/ ;
	}
	unless ($dont_proxy) {
	    # could be slightly more efficient in Perl 5
	    ($dummy,$realhost,$realport)=
		$ENV{'http_proxy'}=~ m#^(http://)?([^/?:]*):?([^/?]*)#i ;
	    $realport= 80 if $realport eq '' ;
	    $request_uri= $URL ;
	    $proxy_auth_header= "Proxy-Authorization: Basic $PROXY_AUTH\015\012"
	       if $PROXY_AUTH ne '' ;
	}
    }


    # Create Referer: header if so configured.
    if (!$HIDE_REFERER) {
	local($referer)= $ENV{'HTTP_REFERER'} ;
	# Only include Referer: if we successfully remove $script_url + flags
	#   from start of referring URL.
	# Note that flags may not always be there.
	if ($referer=~ s#^$script_url(/[^/]*/?)?##) {
	    $referer_header= 'Referer: ' . &proxy_decode($referer) . "\015\012"
		if $referer ne '' ;
	}
    }

    # Add "Connection: close" header if we're using HTTP 1.1.
    $connection_header= "Connection: close\015\012" if $HTTP_VERSION eq '1.1' ;


    #------ Connect socket to host; send request; wait with select() ------

    # To be able to retry on a 401 Unauthorized response, put the whole thing
    #   in a labeled block.  Note that vars have to be reinitialized.
    HTTP_GET: {

	# Open a socket and send the request.
	# Accept: and User-Agent: headers are here because they affect results.
	# We're not fully supporting HTTP 1.1, but at least send Host: header.
	# We're anonymously browsing, so don't include From: or Referer:.
	#   The User-Agent: header is a very teensy privacy risk, but some
	#   pages load differently with different browsers.

	&newsocketto(*S, $realhost, $realport) ;
	binmode S ;   # see note with "binmode STDOUT", above

	print S $ENV{'REQUEST_METHOD'}, ' ', $request_uri, " HTTP/$HTTP_VERSION\015\012",
		'Host: ', $host, $portst, "\015\012",    # needed for multi-homed servers
		'Accept: ', $env_accept, "\015\012",     # possibly modified
		'User-Agent: ', $USER_AGENT, "\015\012",
		$proxy_auth_header,                      # empty if not needed
		$referer_header,                         # empty if not needed
		$connection_header ;                     # empty if not needed

	# Add the cookie if it exists and cookies aren't banned here.
	print S 'Cookie: ', $cookie_to_server, "\015\012"
	    if !$cookies_are_banned_here && ($cookie_to_server ne '') ;


	# Add Authorization: header if we've had a challenge.
	if ($realm ne '') {
	    # If we get here, we know $realm has a defined $auth and has not
	    #   been tried.
	    print S "Authorization: Basic $auth{$realm}\015\012" ;
	    $tried_realm= $realm ;

	} else {
	    # If we have auth information for this server, what the hey, let's
	    #   try one, it may save us a request/response cycle.
	    # First case is for rare case when auth info is in URL.  Related
	    #   block 100 lines down needs no changes.
	    if ($username ne '') {
		print S "Authorization: Basic ",
			&base64($username . ':' . $password),
			"\015\012" ;
	    } elsif ( ($tried_realm,$auth)= each %auth ) {
		print S "Authorization: Basic $auth\015\012" ;
	    }
	}


	# A little problem with authorization and POST requests: If auth
	#   is required, we won't know which realm until after we make the
	#   request and get part of the response.  But to make the request,
	#   we have to send the entire POST body, because some servers
	#   mistakenly require that before returning even an error response.
	#   So this means we have to send the entire POST body, and be
	#   prepared to send it a second time, thus we have to store it
	#   locally.  Either that, or fail to send the POST body a second
	#   time.  Here, we let the owner of this proxy set $MAX_REQUEST_SIZE:
	#   store and post a second time if a request is smaller, or else
	#   die with 413 the second time through.

	# If request method is POST, copy content headers and body to request.
	# The first time through here, save body to @postbody, if the body's
	#   not too big.
	if ($ENV{'REQUEST_METHOD'} eq 'POST') {

	    if ($body_too_big) {
		# Quick 'n' dirty response for an unlikely occurrence.
		# 413 is not actually an HTTP/1.0 response...
		&HTMLdie("Sorry, this proxy can't handle a request larger "
		       . "than $MAX_REQUEST_SIZE bytes at a password-protected"
		       . " URL.  Try reducing your submission size, or submit "
		       . "it to an unprotected URL.", 'Submission too large',
			 '413 Request Entity Too Large') ;
	    }

	    # Otherwise...
	    $lefttoget= $ENV{'CONTENT_LENGTH'} ;
	    print S 'Content-type: ', $ENV{'CONTENT_TYPE'}, "\015\012",
		    'Content-length: ', $lefttoget, "\015\012\015\012" ;

	    if (@postbody) {
		print S @postbody ;
	    } else {
		$body_too_big= ($lefttoget > $MAX_REQUEST_SIZE) ;
		# Loop to guarantee all is read from STDIN.
		do {
		    $lefttoget-= read(STDIN, $postblock, $lefttoget) ;
		    print S $postblock ;
		    # efficient-- only doing test when input is slow anyway.
		    push(@postbody, $postblock) unless $body_too_big ;
		} while $lefttoget && ($postblock ne '') ;
	    }

	# For GET requests, just add extra blank line
	} else {
	    print S "\015\012" ;
	}


	# Wait a minute for the response to start
	vec($rin= '', fileno(S), 1)= 1 ;
	select($rin, undef, undef, 60)
	    || &HTMLdie("No response from $realhost:$realport") ;


	#------ Read full response into $status, $headers, and $body/@body ----

	# Support both HTTP 1.x and HTTP 0.9
	$status= <S> ;  # first line, which is the status line in HTTP 1.x


	# HTTP 0.9
	# Ignore possibility of HEAD, since it's not defined in HTTP 0.9.
	# Do any HTTP 0.9 servers really exist anymore?
	unless ($status=~ m#^HTTP/#) {
	    $is_html= 1 ;   # HTTP 0.9 by definition implies an HTML response
	    $content_type= 'text/html' ;
	    undef $/ ;
	    $body= $status . <S> ;
	    $status= '' ;

	    # split $body into @body; each element ends in ">" or EOstring
	    @body= $body=~ /([^>]*>?)/g ;

	    close(S) ;
	    return ;
	}


	# After here, we know we're using HTTP 1.x

	# Be sure to handle case when server doesn't send blank line!  It's
	#   rare and erroneous, but a couple servers out there do that when
	#   responding with a redirection.  This can cause some processes to
	#   linger and soak up resources, particularly under mod_perl.
	# To handle this, merely check for eof(S) in until clause below.

	$headers= '' ;   # could have been set by first attempt
	do {
	    $headers.= $_= <S> ;    # $headers includes last blank line
	} until (/^(\015\012|\012)$/) || eof(S) ; # lines end with LF or CRLF

	# Unfold long header lines, a la RFC 822 section 3.1.1
	$headers=~ s/(\015\012|\012)[ \t]+/ /g ;

	# Check for 401 Unauthorized response
	if ($status=~ m#^HTTP/\d+\.\d+\s+401\b#) {
	    ($realm)=
		$headers=~ /^WWW-Authenticate:\s*Basic\s+realm="([^"\n]*)/i ;
	    &HTMLdie("Error by target server: no WWW-Authenticate header.")
		unless $realm ne '' ;

	    if ($auth{$realm} eq '') {
		&get_auth_from_user($host, $realm, $URL) ;
	    } elsif ($realm eq $tried_realm) {
		&get_auth_from_user($host, $realm, $URL, 1) ;
	    }

	    # so now $realm exists, has defined $auth, and has not been tried
	    close(S) ;
	    redo HTTP_GET ;
	}


	# Extract $content_type, used in several places
	($content_type)= $headers=~ m#^Content-type:\s*([\w/.+\$-]*)#i ;
	$content_type=~ tr/A-Z/a-z/ ;

	# If we're text only, then cut off non-text responses (but allow
	#   unspecified types).
	if ($TEXT_ONLY) {
	    if ( ($content_type ne '') && ($content_type!~ m#^text/#i) ) {
		close(S) ;
		&non_text_die ;
	    }
	}

	# If we're removing scripts, then disallow script MIME types.
	if ($scripts_are_banned_here) {
	    close(S), &script_content_die
		if $content_type=~ /^$SCRIPT_TYPE_REGEX$/io ;

	    # Note that the non-standard Link: header, which may link to a
	    #   style sheet, is handled in http_fix().
	}


	# If URL matches one of @BANNED_IMAGE_URL_PATTERNS, then skip the
	#   resource unless it's clearly a text type.
	if ($images_are_banned_here) {
	    close(S), &skip_image  unless $content_type=~ m#^text/#i ;
	}

	# Keeping $base_url, $base_host, and $base_path up-to-date is an
	#   ongoing job.  Here, we look in appropriate headers.  Note that if
	#   Content-Base: doesn't exist, Content-Location: is an absolute URL.
	if        ($headers=~ m#^Content-Base:\s*([\w+.-]+://\S+)#i) {
	    $base_url= $1, &fix_base_vars ;
	} elsif   ($headers=~ m#^Content-Location:\s*([\w+.-]+://\S+)#i) {
	    $base_url= $1, &fix_base_vars ;
	} elsif   ($headers=~ m#^Location:\s*([\w+.-]+://\S+)#i) {
	    $base_url= $1, &fix_base_vars ;
	}

	# Now, fix the headers with &http_fix().  It uses &full_url(), and
	#   may modify the headers we just extracted the base URL from.
	# This also includes cookie support.
	&http_fix ;


	# Set $is_html if headers indicate HTML response.
	# Question: are there any other HTML-like MIME types, including x-... ?
	$is_html= 1  if $content_type eq 'text/html' ;


	# Some servers return HTML content without the Content-Type: header.
	#   These MUST be caught, because Netscape displays them as HTML, and
	#   a user could lose their anonymity on these pages.
	# According to the HTTP 1.1 spec, section. 7.2.1, browsers can choose
	#   how to deal with HTTP bodies with no Content-Type: header.  See
	#       http://www.ietf.org/rfc/rfc2616.txt
	# In such a case, Netscape seems to always assume "text/html".
	#   Konqueror seems to guess the MIME type by using the Unix "file"
	#   utility on the first 1024 bytes, and possibly other clues (e.g.
	#   resource starts with "<h1>").
	# In any case, we must interpret as HTML anything that *may* be
	#   interpreted as HTML by the browser.  So if there is no
	#   Content-Type: header, set $is_html=1 .  The worst that would
	#   happen would be the occasional content mangled by modified URLs,
	#   which is better than a privacy hole.

	$is_html= 1  if ($content_type eq '') ;


	# To support non-NPH hack, replace first part of $status with
	#   "Status:" if needed.
	$status=~ s#^\S+#Status:#  if $NOT_RUNNING_AS_NPH ;


	# To support streaming media and large files, read the data from
	#   the server and send it immediately to the client.  The exception
	#   is HTML content, which still must be read fully to be converted
	#   in the main block.  HTML content is not normally streaming or
	#   very large.
	# This requires $status and $headers to be returned now, which is
	#   OK since headers have been completely cleaned up by now.  This
	#   also means that changes after this point to $body/@body won't
	#   have any effect, which in fact is fine in the case of non-HTML
	#   resources.  Set $response_sent to prevent the main block from
	#   sending a response.
	# Also, handle any non-HTML types here which must be proxified.
	# Unfortunately, Perl 4 seems to have buffer problems when using
	#   read() and misses the first block-- conflict with <S>, maybe?
	#   Anyway, it has to be handled differently.  :P
	# Be sure to prepend $first_block to socket input, if it's set above.
	# This is a bit sloppy now, just a quick hack to get rudimentary
	#   handling of multiple types working and released.  It will be
	#   rewritten more cleanly at some point, when the whole proxifying
	#   of different types is modularized better.

	# Only read body if the request method is not HEAD
	if ($ENV{'REQUEST_METHOD'} ne 'HEAD') {

	    # Because of the erroneous way some browsers use the expected
	    #   MIME type instead of the actual Content-Type: header, check
	    #   $expected_type first.
	    # Since style sheets tend to be automatically loaded, whereas other
	    #   types (like scripts) are more user-selected, plus the fact that
	    #   CSS can be safely proxified and scripts cannot, we treat a
	    #   resource as CSS if it *may* be treated as CSS by the browser.
	    #   This is relevant when $expected_type and Content-Type: differ.

	    if (    ($expected_type=~ /^$TYPES_TO_HANDLE_REGEX$/io)
		 || ($content_type=~  /^$TYPES_TO_HANDLE_REGEX$/io)   ) {

		local($buf, $type) ;
		if ( ($expected_type eq 'text/css') || ($content_type eq 'text/css') ) {
		    $type= 'text/css' ;
		} elsif ($expected_type=~ /^$TYPES_TO_HANDLE_REGEX$/io) {
		    $type= $expected_type ;
		} else {
		    $type= $content_type ;
		}

		$body= $first_block ;

		if ($]>=5) { 1 while read(S, $body, 16384, length($body)) }  # Perl>=5
		else       { $body.= $buf while $buf= <S> }                  # Perl 4

		$body= &proxify_block($body, $type) ;
		$headers=~ s/^Content-Length:.*/
			     'Content-Length: ' . length($body) /ie ;

		print $status, $headers, $body ;
		$response_sent= 1 ;


	    } elsif ($is_html) {
		$/= '>' ;
		@body= () ;
		if ($first_block ne '') {
		    $first_block.= <S> ;  # read up through the next ">"

		    # split $first_block into @body; each item ends in > or EOS
		    @body= $first_block=~ /([^>]*>?)/g ;
		}
		push(@body, <S>) ;


	    # This is for when the resource is passed straight through without
	    #   modification.
	    } else {
		local($buf) ;
		print $status, $headers, $first_block ;

		if ($]>=5) { print $buf while read(S, $buf, 16384) }  # Perl>=5
		else       { print $buf while $buf= <S> }             # Perl 4

		$response_sent= 1 ;

	    }


	} else {
	    $body= ''; @body= () ;
	}

	close(S) ;

    }  # HTTP_GET:

}  # sub http_get()



# ftp_get:

sub ftp_get {
    local($is_dir, $rcode, @r, @cwdmsg, $dataport, $remote_addr,
	  $ext, $content_type, %content_type, $content_length,
	  $enc_URL) ;
    local($/)= "\012" ;

    $port= 21 if $port eq '' ;

    # List of file extensions and associated MIME types, or at least the ones
    #   a typical browser distinguishes from a nondescript file.
    # I'm open to suggestions for improving this.  One option is to read the
    #   file mime.types if it's available.
    %content_type=
	  ('txt',  'text/plain',
	   'text', 'text/plain',
	   'htm',  'text/html',
	   'html', 'text/html',
	   'png',  'image/png',
	   'jpg',  'image/jpeg',
	   'jpeg', 'image/jpeg',
	   'jpe',  'image/jpeg',
	   'gif',  'image/gif',
	   'xbm',  'image/x-bitmap',
	   'mpg',  'video/mpeg',
	   'mpeg', 'video/mpeg',
	   'mpe',  'video/mpeg',
	   'qt',   'video/quicktime',
	   'mov',  'video/quicktime',
	   'aiff', 'audio/aiff',
	   'aif',  'audio/aiff',
	   'au',   'audio/basic',
	   'snd',  'audio/basic',
	   'wav',  'audio/x-wav',
	   'mp2',  'audio/x-mpeg',
	   'mp3',  'audio/mpeg',
	   'ram',  'audio/x-pn-realaudio',
	   'rm',   'audio/x-pn-realaudio',
	   'ra',   'audio/x-pn-realaudio',
	   'zip',  'application/x-gzip-compressed',
	   ) ;


    $is_dir= $path=~ m#/$# ;
    $is_html= 0 if $is_dir ;   # for our purposes, do not treat dirs as HTML

    # Set $content_type based on file extension.
    # Hmm, still unsure how best to handle unknown file types.  This labels
    #   them as text/plain, so that README's, etc. will display right.
    ($ext)= $path=~ /\.(\w+)$/ ;  # works for FTP, not for URLs with query etc.
    $ext=~ tr/A-Z/a-z/ ;
    $content_type= ($is_html || $is_dir)  ? 'text/html'
					  : $content_type{$ext}
					    || 'text/plain' ;


    # If we're removing scripts, then disallow script MIME types.
    if ($scripts_are_banned_here) {
	&script_content_die if $content_type=~ /^$SCRIPT_TYPE_REGEX$/io ;
    }


    # Create $status and $headers, and leave $body/@body and $is_html as is.
    # Directories use an HTML response, though $is_html is false when $is_dir.
    $status= "$HTTP_1_X 200 OK\015\012" ;
    $headers= ($content_type  ? "Content-type: $content_type\015\012"  : '')
	      . "Date: " . &rfc1123date(time,0) . "\015\012\015\012" ;


    # Open the control connection to the FTP server
    &newsocketto(*S, $host, $port) ;
    binmode S ;   # see note with "binmode STDOUT", above

    # Luckily, RFC 959 (FTP) has a really good list of all possible response
    #   codes to all possible commands, on pages 50-53.

    # Connection establishment
    ($rcode)= &ftp_command('', '120|220') ;
    &ftp_command('', '220') if $rcode==120 ;

    # Login
    ($rcode, @welcome)= &ftp_command("USER $username\015\012", '230|331') ;
    ($rcode, @welcome)= &ftp_command("PASS $password\015\012", '230|202')
	if $rcode==331 ;

    # Set transfer parameters
    &ftp_command("TYPE I\015\012", '200') ;

    # Create and listen on data socket
    socket(DATA_LISTEN, $AF_INET, $SOCK_STREAM, (getprotobyname('tcp'))[2])
	|| &HTMLdie("Couldn't create FTP data socket: $!") ;
    bind(DATA_LISTEN, pack('S n a4 x8', $AF_INET, 0, "\0\0\0\0") )
	|| &HTMLdie("Couldn't bind FTP data socket: $!") ;
    $dataport= (unpack('S n a4 x8', getsockname(DATA_LISTEN)))[1] ;
    listen(DATA_LISTEN,1)
	|| &HTMLdie("Couldn't listen on FTP data socket: $!") ;
    select((select(DATA_LISTEN), $|=1)[0]) ;    # unbuffer the socket
    select((select(DATA_XFER), $|=1)[0]) ;      # unbuffer the socket

    # Tell FTP server which port to connect to
    &ftp_command( sprintf("PORT %d,%d,%d,%d,%d,%d\015\012",
			  unpack('C4', substr(getsockname(S),4,4)),
			  $dataport>>8, $dataport & 255),
		  '200') ;


    # Hack to help handle spaces in pathnames.  :P
    # $path should be delivered to us here with spaces encoded as "%20".
    #   But that's not what the FTP server wants (or what we should display),
    #   so translate them back to spaces in a temporary copy of $path.
    #   Hopefully the FTP server will allow spaces in the FTP commands below,
    #   like "CWD path with spaces".
    local($path)= $path ;
    $path=~ s/%20/ /g ;


    # Do LIST for directories, RETR for files
    if ($is_dir) {
	# If we don't CWD first, then symbolic links won't be followed.
	($rcode, @cwdmsg)= &ftp_command("CWD $path\015\012", '250') ;
	($rcode, @r)= &ftp_command("LIST\015\012", '125|150') ;
# was:  ($rcode, @r)= &ftp_command("LIST $path\015\012", '125|150') ;

    } else {
	($rcode, @r)= &ftp_command("RETR $path\015\012", '125|150|550') ;

	# If 550 response, it may be a symlink to a directory.
	# Try to CWD to it; if successful, do a redirect, else die with the
	#   original error response.  Note that CWD is required by RFC 1123
	#   (section 4.1.2.13), which updates RFC 959.
	if ($rcode==550) {
	    ($rcode)= &ftp_command("CWD $path\015\012", '') ;
	    &ftp_error(550,@r) unless $rcode==250 ;

	    ($enc_URL= $URL)=~ s/ /%20/g ;  # URL-encode any spaces

	    # Redirect the browser to the same URL with a trailing slash
	    print "$HTTP_1_X 301 Moved Permanently\015\012",
		  "Date: ", &rfc1123date(time,0), "\015\012",
		  "Location: ", $this_url, &proxy_encode($enc_URL . '/'),
		  "\015\012\015\012" ;
	    goto EXIT ;
	}
    }


    # Streaming support added in 1.3.  For notes about streaming, look near
    #   the end of the http_get() routine.  Basically, as long as a resource
    #   isn't HTML (or a directory listing, in the case of FTP), we can pass
    #   the data immediately to the client, since it won't be modified.  Be
    #   sure to set $response_sent here.

    # Accept the connection and read the data into @body or $body
    ($remote_addr= accept(DATA_XFER, DATA_LISTEN))
	|| &HTMLdie("Error accepting FTP data socket: $!") ;
#    &HTMLdie("Intruder Alert!  Someone other than the server is trying to "
#	   . "send you data.")
#	unless (substr($remote_addr,4,4) eq substr(getpeername(S),4,4)) ;


    # This first block is for the rare case when an FTP resource is a special
    #   type that needs to be converted, e.g. a style sheet.  The block is
    #   copied in from http_get() and modified.  It will be cleaner and
    #   handled differently in a future version.

    if ( !$is_dir && !$is_html &&
	 (    ($expected_type=~ /^$TYPES_TO_HANDLE_REGEX$/io)
	   || ($content_type=~  /^$TYPES_TO_HANDLE_REGEX$/io)   ) ) {

	local($type) ;
	if ( ($expected_type eq 'text/css') || ($content_type eq 'text/css') ) {
	    $type= 'text/css' ;
	} elsif ($expected_type=~ /^$TYPES_TO_HANDLE_REGEX$/io) {
	    $type= $expected_type ;
	} else {
	    $type= $content_type ;
	}

	1 while read(DATA_XFER, $body, 16384, length($body)) ;

	$body= &proxify_block($body, $type) ;

	print $status, $headers, $body ;
	$response_sent= 1 ;


    } elsif ($is_html) {
	$/= '>' ;
	@body= <DATA_XFER> ;

    } elsif ($is_dir) {
	undef $/ ;            # This was used for all non-HTML before streaming
	$body= <DATA_XFER> ;  #   was supported.

    } else {

	# Stick a Content-Length: header into the headers if appropriate (often
	#   there's a "(xxx bytes)" string in a 125 or 150 response line).
	# Be careful about respecting previous value of $headers, which may
	#   already end in a blank line.
	foreach (grep(/^(125|150)/, @r)) {
	    if ( ($content_length)= /\((\d+)[ \t]+bytes\)/ ) {
		$headers= "Content-Length: $content_length\015\012" .$headers ;
		last ;
	    }
	}

	# This is the primary change to support streaming media.
	local($buf) ;
	print $status, $headers ;
	print $buf while read(DATA_XFER, $buf, 16384) ;
	$response_sent= 1 ;

    }


    close(DATA_XFER) ;
    close(DATA_LISTEN) ;

    # Get the final completion response
    &ftp_command('', '226|250') ;

    &ftp_command("QUIT\015\012") ;   # don't care how they answer

    close(S) ;

    # Make a user-friendly directory listing.  Add Content-Length: header.
    if ($is_dir) {
	&ftp_dirfix ;
	$headers= "Content-Length: " . length($body) . "\015\012" . $headers ;
    }

}


# Send $cmd and return response code followed by full lines of  FTP response.
# Die if response doesn't match the regex $ok_response.
sub ftp_command {
    local($cmd,$ok_response)= @_ ;
    local(@r,$rcode) ;
    local($/)= "\012" ;

    print S $cmd ;

    $_= $r[0]= <S> ;
    $rcode= substr($r[0],0,3) ;
    until (/^$rcode /) {      # this catches single- and multi-line responses
	push(@r, $_=<S>) ;
    }

    &ftp_error($rcode,@r) if $ok_response ne '' && $rcode!~ /$ok_response/ ;
    return $rcode, @r ;
}


# Convert a directory listing to user-friendly HTML.
# A couple of tangles here to handle spaces in filenames.  We should probably
#   handle spaces in other protocols too, but URLs normally prohibit spaces--
#   it's only relative paths within a scheme (like FTP) that would have them.
sub ftp_dirfix {
    local($newbody, $parent_link, $max_namelen,
	  @f, $is_dir, $is_link, $link, $name, $size, $size_type, $file_type,
	  $welcome, $cwdmsg, $enc_path) ;

    # Set minimum name column width; longer names will widen the column
    $max_namelen= 16 ;

    # each file should have name/, size, date
    @body= split(/\015?\012/, $body) ;
    foreach (@body) {
	# Hack to handle leading spaces in filenames-- only allow a single
	#   space after the 8th field before filename starts.
	# ftp_get() should probably be rewritten to use NLST instead of LIST
	#   to get FTP directory listings.
#	@f= split(" ", $_, 9) ;   # Note special use of " " pattern.
#	next unless $#f>=8 ;
	@f= split(" ", $_, 8) ;   # Note special use of " " pattern.
	next unless $#f>=7 ;
	@f[7,8]= $f[7]=~ /^(\S*) (.*)/ ;  # handle leading spaces in filenames

	next if $f[8]=~ /^\.\.?$/ ;
	$file_type= '' ;
	$is_dir=  $f[0]=~ /^d/i ;
	$is_link= $f[0]=~ /^l/i ;
	$file_type= $is_dir     ? 'Directory'
		  : $is_link    ? 'Symbolic link'
		  :               '' ;
	$name= $f[8] ;
	$name=~ s/^(.*) ->.*$/$1/ if $is_link ;   # remove symlink's " -> xxx"
	$name.= '/' if $is_dir ;
	$max_namelen= length($name) if length($name)>$max_namelen ;
	if ($is_dir || $is_link) {
	    ($size, $size_type)= () ;
	} else {
	    ($size, $size_type)= ($f[4], 'bytes') ;
	    ($size, $size_type)= ($size>>10, 'Kb') if $size > 10240 ;
	}

	# Easy absolute URL calculation, because we know it's a relative path.
	($enc_path= $base_path . $name)=~ s/ /%20/g ;  # URL-encode any spaces
	$link=  &HTMLescape( $this_url . &proxy_encode($enc_path) ) ;

	$newbody.=
	    sprintf("  <a href=\"%s\">%s</a>%s %5s %-5s %3s %2s %5s  %s\012",
			   $link, $name, "\0".length($name),
			   $size, $size_type,
			   @f[5..7],
			   $file_type) ;
    }

    # A little hack to get filenames to line up right-- replace embedded
    #  "\0"-plus-length with correct number of spaces.
    $newbody=~ s/\0(\d+)/ ' ' x ($max_namelen-$1) /ge ;

    if ($path eq '/') {
	$parent_link= '' ;
    } else {
	($enc_path= $base_path)=~ s#[^/]*/$## ;
	$enc_path=~ s/ /%20/g ;  # URL-encode any spaces
	$link=  &HTMLescape( $this_url . &proxy_encode($enc_path) ) ;
	$parent_link= "<a href=\"$link\">Up to higher level directory</a>" ;
    }

    if ($SHOW_FTP_WELCOME) {
	$welcome= join('', grep(s/^230-//, @welcome)) ;
	# Make links of any URLs in $welcome.  Imperfect regex, but does OK.
	# "\"abc" doesn't work in Perl 4 substitutions; must use '"' . "abc"
	$welcome=~ s#\b([\w+.-]+://[^\s"']+[\w/])(\W)#
	    '<a href="' . &full_url($1) . '"' . ">$1</a>$2" #ge ;
	$welcome.= "<hr>" if $welcome ne '' ;
    } else {
	$welcome= '' ;
    }

    # If CWD returned a message about this directory, display it.  Make links
    #   a la $welcome, above.
    $cwdmsg= join('', grep(s/^250-//, @cwdmsg)) ;
    $cwdmsg=~ s#\b([\w+.-]+://[^\s"']+[\w/])(\W)#
	'<a href="' . &full_url($1) . '"' . ">$1</a>$2" #ge ;
    $cwdmsg.= "<hr>" if $cwdmsg ne '' ;


    $body= <<EOS ;
<html>
<title>FTP directory of $URL</title>
<body>
<h1>FTP server at $host</h1>
<h2>Current directory is $path</h2>
<hr>
<pre>
$welcome$cwdmsg
$parent_link
$newbody
</pre>
<hr>
</body>
</html>
EOS

}


# Return a generalized FTP error page.
# For now, respond with 200.  In the future, give more appropriate codes.
sub ftp_error {
    local($rcode,@r)= @_ ;

    close(S) ; close(DATA_LISTEN) ; close(DATA_XFER) ;

    local($date_header)= &rfc1123date(time, 0) ;

    print <<EOH ;
$HTTP_1_X 200 OK
Date: $date_header
Content-type: text/html

<html>
<head><title>FTP Error</title></head>
<body>
<h1>FTP Error</h1>
<h3>The FTP server at $host returned the following error response:</h3>
<pre>
EOH
    print @r, "</pre>\n" ;

    &footer ;
    goto EXIT ;
}


#--------------------------------------------------------------------------

#
# <scheme>_fix: modify response as appropriate for given protocol (scheme).
#

# http_fix: modify headers as needed, including cookie support.
# Note that headers have already been unfolded, when they were read in.
# Some HTTP headers are defined as comma-separated lists of values, and they
#   should be split before being processed.  According to the HTTP spec in
#   RFC 2616, such headers are:
#     Accept|Accept-Charset|Accept-Encoding|Accept-Language|Accept-Ranges|
#     Allow|Cache-Control|Connection|Content-Encoding|Content-Language|
#     If-Match|If-None-Match|Pragma|Public|Transfer-Encoding|Upgrade|Vary|
#     Via|Warning|WWW-Authenticate
#   As it turns out, none of these need to undergo any transformation in
#   this program; more directly, none are handled in new_header_value().
#   Thus, we don't need to split any standard headers before processing.
#   See section 4.2 of RFC 2616, plus the header definitions, for more info.
# Conceivably, Via: and Warning: could be exceptions to this, since they
#   do contain hostnames.  But a) these are primarily for diagnostic info and
#   not used to connect to those hosts, and b) we couldn't distinguish the
#   hostnames from pseudonyms anyway.
# Unfortunately, the non-standard Link: and URI: headers may be lists, and
#   we *do* have to process them.  Because of their unusual format and rarity,
#   these are handled as lists directly in new_header_value().
sub http_fix {
    local($name, $value, $new_value) ;
    local(@headers)= $headers=~ /^([^\012]*\012?)/g ;  # split into lines

    foreach (@headers) {
	next unless ($name, $value)= /^([\w.-]+):\s*([^\015\012]*)/ ;
	$new_value= &new_header_value($name, $value) ;
	$_= defined($new_value)
	    ? "$name: $new_value\015\012"
	    : '' ;
    }

    $headers= join('', @headers) ;
}


# Returns the value of an updated header, e.g. with URLs transformed to point
#   back through this proxy.  Returns undef if the header should be removed.
# This is used to translate both real headers and <meta http-equiv> headers.
# Special case for URI: and Link: -- these headers can be lists of values
#   (see the HTTP spec, and comments above in http_fix()).  Thus, we must
#   process these headers as lists, i.e. transform each URL in the header.
sub new_header_value {
    local($name, $value)= @_ ;
    $name=~ tr/A-Z/a-z/ ;

    # sanity check
    return undef if $name eq '' ;

    # First, modify any appropriate HTTP headers to point back through the
    #   proxy.  Location: is most important.
    # Note that all these are absolute URIs, except possibly Content-Location:
    #   or URI:, which may be relative to Content-Base or the request URI--
    #   notably, NOT relative to anything in the content, like a <base> tag.
    return &full_url($value)
	if    $name eq 'content-base'
	   || $name eq 'content-location'
	   || $name eq 'location' ;
#	   || $name eq 'uri' ;   # I think this was a bug in previous versions


    # And the non-standard Refresh: header... any others?
    $value=~ s/(;\s*URL=)(\S*)/ $1 . &full_url($2) /ie,   return $value
	if $name eq 'refresh' ;

    # The deprecated URI: header may contain several URI's, inside <> brackets.
    $value=~ s/<(\s*[^>\015\012]*)>/ '<'.&full_url($1).'>' /gie, return $value
	if $name eq 'uri' ;


    # The non-standard Link: header is a little problematic.  It's described
    #   in the HTTP 1.1 spec, section 19.6.2.4, but it is not standard.  It
    #   can be used to link to style sheets, but the mechanism for indicating
    #   the style sheet type (=language, which could be a script MIME type)
    #   is not defined.
    # The HTML 4.0 spec (section 14.6) gives a little more detail regarding
    #   its use of the Link: header, but is still ambiguous-- e.g. their
    #   examples don't specify the type, though elsewhere it's implied that's
    #   required.
    # Generally speaking, we handle this like a <link> tag.  For notes about
    #   this block, see the block above that handles <link> tags.  For a
    #   description of the unusual format of this header, see the HTTP spec.
    # Note that this may be a list of values, and all URIs in it must be
    #   handled.  This gets a little messy, because we split on commas, but
    #   don't split on commas that are inside <> brackets, because that's
    #   the URL.
    if ($name eq 'link') {
	local($v, @new_values) ;

	local(@values)= $value=~ /(<[^>]*>[^,]*)/g ;
	foreach $v (@values) {
	    local($type)=
		$v=~ m#[^\w.\/?&-]type\s*=\s*["']?\s*([/\w.+\$-]+)#i ;
	    $type=~ tr/A-Z/a-z/ ;

	    if ($type eq '') {
		local($rel) ;
		$rel= $+
		    if $v=~ /[^\w.\/?&-]rel\s*=\s*("([^"]*)"|'([^']*)'|([^'"][^\s]*))/i ;
		$type= 'text/css' if $rel=~ /\bstylesheet\b/i ;
	    }

	    return undef
		if $scripts_are_banned_here && $type=~ /^$SCRIPT_TYPE_REGEX$/io ;

	    local($this_url)= $this_url ;
	    if ($type ne '') {
		$this_url= $script_url . '/' .
		    &pack_flags($REMOVE_COOKIES, $REMOVE_SCRIPTS, $FILTER_ADS,
				$HIDE_REFERER, $INSERT_ENTRY_FORM,
				$is_in_frame, $type)
			. '/' ;
	    }

	    $v=~ s/<(\s*[^>\015\012]*)>/ '<' . &full_url($1) . '>' /gie ;
	    push(@new_values, $v) ;
	}

	return join(', ', @new_values) ;
    }


    # Modify cookies to point back through the script, or they won't work.
    # If they're banned from this server, or they're not allowed with the
    #   current non-text resource, then filter them all out.  This actually
    #   covers all non-text resources, not just images.
    # Apparently, some sites (e.g. zdnet.com) return erroneous responses that
    #   end up bypassing the $NO_COOKIE_WITH_IMAGE setting:  a redirection
    #   response to an image contains cookies along with a meaningless
    #   "Content-Type: text/plain" header.  They only do this on images that
    #   look like Web bugs.  (Hmm, what are the odds of THAT happening by
    #   accident, eh?)  So, we protect this further here:  Not only must the
    #   response indicate a text content-type, but the original request must
    #   have an Accept: header that indicates that text responses are
    #   acceptable.  (I also considered filtering by file extension of the
    #   Location: header, but that's unreliable and blocks some legitimate
    #   cookies.)
    if ($name eq 'set-cookie') {
	return undef
	    if $cookies_are_banned_here ||
#	      ($NO_COOKIE_WITH_IMAGE && $headers!~ m#^Content-Type:\s*text/#i);
	      ($NO_COOKIE_WITH_IMAGE &&
		   ! ( $headers=~ m#^Content-Type:\s*text/#i
		    && grep(m#^(text|\*)/#i, split(/\s*,\s*/, $env_accept)) )
	      );

	return &cookie_to_client($value, $path, $host) ;
    }


    # Extract $default_style_type as needed.
    # Strictly speaking, a MIME type is "token/token", where token is
    #    ([^\x00-\x20\x7f-\xff()<>@,;:\\"/[\]?=]+)   (RFCs 1521 and 822),
    #   but this below covers all existing and likely future MIME types.
    if ($name eq 'content-style-type') {
	$default_style_type= $1, $default_style_type=~ tr/A-Z/a-z/
	    if $value=~ m#^\s*([/\w.+\$-]+)# ;
	return $value ;
    }


    # Extract $default_script_type as needed.
    # Same deal about "token/token" as above.
    if ($name eq 'content-script-type') {
	$default_script_type= $1, $default_script_type=~ tr/A-Z/a-z/
	    if $value=~ m#^\s*([/\w.+\$-]+)# ;
	return $value ;
    }



    # For all non-special headers, return $value
    return $value ;

}


#--------------------------------------------------------------------------
#    Special admin routines, when called via the scheme type "x-proxy://"
#--------------------------------------------------------------------------

#--------------------------------------------------------------------------
#
#   I took the liberty of creating a general mechanism to let this proxy do
#   whatever tricks it needs to do, via the magic URL scheme "x-proxy://".
#   It was required to support HTTP Basic Authentication, and it's useful
#   for other things too.  The mechanism uses a heirarchical URL space: a
#   function family is in the normal "hostname" location, then the functions
#   and subfunctions are where the path segments would be.  A query string
#   is allowed on the end.
#
#   Don't add functions to this that may compromise security, since anyone
#   can request a URL beginning with x-proxy://.  For that matter, malicious
#   Web pages can automatically invoke these URLs, which could be annoying
#   if e.g. they clear your cookies without warning or other acts.
#
#   Which URLs map to which functions should really be documented here.  So,
#
#     //auth/make_auth_cookie
#         receives the authorization form data, sends a formatted auth
#         cookie to the user, and redirects the user to the desired URL.
#
#     //start
#         initiates a browsing session.
#
#     //cookies/clear
#         clears all of a user's cookies.
#
#     //cookies/manage
#         present the user with a page to manage her/his cookies
#
#     //cookies/update
#         process whatever actions are requested from the //cookies/manage
#         page (currently only deletion of cookies).
#
#     //frames/topframe
#         returns the special top frame with the entry form and/or the
#         other insertion.
#
#     //frames/framethis
#         given a URL, returns a page that frames that URL in the lower
#         frame with the top frame above (not currently used).
#
#--------------------------------------------------------------------------

# A general-purpose routine to handle all x-proxy requests.
# This is expected to exit when completed, so make sure any called routines
#   exit if needed.  (By "exit", I mean "goto EXIT".)
sub xproxy {
    local($URL)= @_ ;
    $URL=~ s/^x-proxy://i ;

    # $qs will contain the query string in $URL, whether it was encoded with
    #   the URL or came from QUERY_STRING.
    local($family, $function, $qs)=  $URL=~ m#^//(\w+)(/?[^?]*)\??(.*)#i ;

    if ($family eq 'auth') {

	# For //auth/make_auth_cookie, return an auth cookie and redirect user
	#   to the desired URL.  The URL is already
	if ($function eq '/make_auth_cookie') {
	    local(%in)= &getformvars() ; # must use () or will pass current @_!
	    local($location)= $this_url . $in{'l'} ;  # was already encoded
	    local($cookie)= &auth_cookie(@in{'u', 'p', 'r', 's'}) ;

	    print "$HTTP_1_X 302 Moved\015\012",
		  "Date: ", &rfc1123date(time,0), "\015\012",
		  "Set-Cookie: $cookie\015\012",
		  "Location: $location\015\012\015\012" ;

	    goto EXIT ;
	}


    } elsif ($family eq 'start') {
	&startproxy ;


    } elsif ($family eq 'cookies') {

	# If pages could link to x-proxy:// URLs directly, this would be a
	#   security hole in that malicious pages could clear or update one's
	#   cookies.  But full_url() prevents that.  If that changes, then we
	#   should consider requiring POST in /cookie/clear and /cookie/update
	#   to minimize this risk.
	if ($function eq '/clear') {
	    local($location)=
		$this_url . &proxy_encode('x-proxy://cookies/manage') ;
	    $location.= '?' . $qs    if $qs ne '' ;
	    print "$HTTP_1_X 302 Moved\015\012",
		  "Date: ", &rfc1123date(time,0), "\015\012",
		  &cookie_clearer( $ENV{'HTTP_COOKIE'} ) ,
		  "Location: $location\015\012\015\012" ;

	    goto EXIT ;


	} elsif ($function eq '/manage') {
	    &manage_cookies($qs) ;


	# For //cookies/update, clear selected cookies and go to manage screen.
	} elsif ($function eq '/update') {
	    local(%in)= &getformvars() ; # must use () or will pass current @_!
	    local($location)=
		$this_url . &proxy_encode('x-proxy://cookies/manage') ;

	    # Add encoded "from" parameter to URL if available.
	    if ($in{'from'} ne '') {
		local($from_param)= $in{'from'} ;
		$from_param=~ s/([^\w.-])/ '%' . sprintf('%02x',ord($1)) /ge ;
		$location.=  '?from=' . $from_param ;
	    }

	    # "delete=" input fields are in form &base64(&cookie_encode($name)).
	    local(@cookies_to_delete) ;
	    foreach ( split(/\0/, $in{'delete'}) ) {
		push(@cookies_to_delete, &unbase64($_)) ; # use map{} in Perl 5
	    }

	    print "$HTTP_1_X 302 Moved\015\012",
		  "Date: ", &rfc1123date(time,0), "\015\012",
		  &cookie_clearer( @cookies_to_delete ) ,
		  "Location: $location\015\012\015\012" ;

	    goto EXIT ;
	}


    } elsif ($family eq 'frames') {
	local(%in)= &getformvars($qs) ;

	# Send the top proxy frame when a framed page is reframed.
	if ($function eq '/topframe') {
	    &return_top_frame($in{'URL'}) ;

	# Not currently used
	} elsif ($function eq '/framethis') {
	    &return_frame_doc($in{'URL'}, &HTMLescape(&proxy_decode($in{'URL'}))) ;

	}

    }


    &HTMLdie("Sorry, no such function as //". &HTMLescape("$family$function."),
	     '', '404 Not Found') ;

}


#--------------------------------------------------------------------------
#    Support routines for x-proxy
#--------------------------------------------------------------------------

# Initiate a browsing session. Formerly in the separate program startproxy.cgi.
sub startproxy {
    local(%in)= &getformvars() ;  # must use () or will pass current @_!
    $in{'URL'}=~ s/^\s+|\s+$//g ;    # strip leading or trailing spaces

    # scheme defaults to FTP if host begins with "ftp.", otherwise HTTP.
    if ($in{'URL'}!~ m#^[\w+.-]+://#) {
	$in{'URL'}=  ($in{'URL'}=~ /^ftp\./i)
	    ? "ftp://$in{'URL'}"
	    : "http://$in{'URL'}" ;
    }

    # scheme is case-insensitive, path must start with "/".
    local($scheme, $authority, $path)=
	($in{'URL'}=~ m#^([\w+.-]+)://([^/?]*)(.*)$#i) ;
    $scheme=~ tr/A-Z/a-z/ ;
    $path= "/$path" if $path!~ m#^/# ;


    &HTMLdie("Sorry, only HTTP and FTP are currently supported.")
	unless $scheme=~ /^(http|ftp|x-proxy)$/ ;


    # Allow various shorthand forms of the authority (which is usually
    #   host+port), such as adding "www"+"com" or "ftp"+"com" if the host
    #   doesn't exist on the LAN.

    if ($scheme eq 'http') {
	local($host,$portst)= $authority=~ /^([^:]*)(:?.*)$/ ;
	$host=~ tr/A-Z/a-z/ ;

	# Convert integer hostnames like 3467251275 to a.b.c.d format.
	# This is for big-endian; reverse the list for little-endian.
	$host= join('.', $host>>24 & 255, $host>>16 & 255, $host>>8 & 255,
			 $host & 255)
	    if $host=~ /^\d+$/ ;

	$host= "www.$host.com"  unless ($host=~ /\./) || gethostbyname($host) ;
	$authority= "$host$portst" ;

    } elsif ($scheme eq 'ftp') {
	# Don't mess with it if there's username/password embedded (which you
	#   REALLY shouldn't do).
	unless ($authority=~ /@/) {
	    local($host, $portst)= $authority=~ /^([^:]*)(:?.*)$/ ;
	    $host=~ tr/A-Z/a-z/ ;
	    $host= "ftp.$host.com"
		unless ($host=~ /\./) || gethostbyname($host) ;
	    $authority= "$host$portst" ;
	}

    }


    # Prepend flag segment of PATH_INFO
    # This "erroneously" sets flags to "000000" when user config is not
    #   allowed, but it doesn't really affect anything.
    $this_url=~ s#[^/]*/$## ;   # remove old flag segment from $this_url
    $this_url.= &pack_flags(@in{'rc', 'rs', 'fa', 'br', 'if'}, $is_in_frame, '') . '/' ;


    print "$HTTP_1_X 302 Moved\015\012",
	  "Date: ", &rfc1123date(time,0), "\015\012",
	  "Location: ", $this_url, &proxy_encode("$scheme://$authority$path"),
	  "\015\012\015\012" ;

    goto EXIT ;
}



# Create the flag segment of PATH_INFO from the given flags, not including
#   slashes.  Result should be a valid path segment (i.e. alphanumeric and
#   certain punctuation OK, but no slashes or white space).
# This routine defines the structure of the flag segment.
# Note that an $expected_type of '' explicitly means that no type in particular
#   is expected, which will be the case for almost all resources.
# Note that any unrecognized MIME type (i.e. no element in %MIME_TYPE_ID)
#   is treated the same as '', i.e. element #0 -> "A" .
sub pack_flags {
    local($remove_cookies, $remove_scripts, $filter_ads, $hide_referer,
	  $insert_entry_form, $is_in_frame, $expected_type)= @_ ;
    local($flags) ;

    # Force all values to boolean for this format.
    $flags=   $remove_cookies     ? 1  : 0  ;
    $flags.=  $remove_scripts     ? 1  : 0  ;
    $flags.=  $filter_ads         ? 1  : 0  ;
    $flags.=  $hide_referer       ? 1  : 0  ;
    $flags.=  $insert_entry_form  ? 1  : 0  ;
    $flags.=  $is_in_frame        ? 1  : 0  ;

    # Add MIME type flag, packed into one character.
    $expected_type=~ tr/A-Z/a-z/ ;
    $expected_type= pack('C', $MIME_TYPE_ID{$expected_type}) ;
    $expected_type=~ tr#\x00-\x3f#A-Za-z0-9+-# ; # almost same as base64 chars

    $flags.=  $expected_type ;

    return $flags ;
}


# The reverse of pack_flags()-- given a flag segment from PATH_INFO, break
#   out all flag info.  The return list should match the input list for
#   pack_flags().
sub unpack_flags {
    local($flags)= @_ ;
    local($remove_cookies, $remove_scripts, $filter_ads, $hide_referer,
	  $insert_entry_form, $is_in_frame, $expected_type) ;

    ($remove_cookies, $remove_scripts, $filter_ads, $hide_referer,
     $insert_entry_form, $is_in_frame, $expected_type)=
	split(//, $flags) ;

    # Force all flags to valid values (currently all are 1 or 0).
    $remove_cookies=    $remove_cookies     ? 1  : 0  ;
    $remove_scripts=    $remove_scripts     ? 1  : 0  ;
    $filter_ads=        $filter_ads         ? 1  : 0  ;
    $hide_referer=      $hide_referer       ? 1  : 0  ;
    $insert_entry_form= $insert_entry_form  ? 1  : 0  ;
    $is_in_frame=       $is_in_frame        ? 1  : 0  ;

    # Extract expected MIME type from final one-character flag
    $expected_type=~ tr#A-Za-z0-9+-#\x00-\x3f# ;
    $expected_type= $ALL_TYPES[unpack('C', $expected_type)] ;

    return ($remove_cookies, $remove_scripts, $filter_ads, $hide_referer,
	    $insert_entry_form, $is_in_frame, $expected_type) ;
}


#--------------------------------------------------------------------------
#    Cookie routines
#--------------------------------------------------------------------------

# As of version 1.3, cookies are now a general mechanism for sending various
#   data to the proxy.  So far that's only authentication info and actual
#   cookies, but more functions could be added.  The new scheme essentially
#   divides up the cookie name space to accommodate many categories.
# Explanation: Normally, a cookie is uniquely identified ("keyed") by the
#   domain, path, and name, but for us the domain and path will always be
#   that of the proxy script, so we need to embed all "key" information into
#   the cookie's name.  Here, the general format for a cookie's name is
#   several fields, joined by ";".  The first field is always a cookie type
#   identifier, like "AUTH" or "COOKIE", and the remaining fields vary
#   according to cookie type.  This compound string is then URL-encoded as
#   necessary (cookie names and values can't contain semicolons, commas, or
#   white space).  The cookie's value contains whatever you need to store,
#   also URL-encoded as necessary.

# A general bug in cookie routines-- ports are not considered, which may
#   matter for both AUTH and COOKIE cookies.  It only matters when two ports
#   on the same server are being used.


# Returns all info we need from cookies.  Right now, that means one composite
#   cookie with all cookies that match the domain and path (and no others!),
#   and an %auth hash to look up auth info by server and realm.  Essentially,
#   this undoes the transformation done by the cookie creation routines.
# @auth is used instead of %auth for slight speedup.
# See notes where the various cookies are created for descriptions of their
#   format; currently, that's in cookie_to_client() and auth_cookie().
sub parse_cookie {
    local($cookie, $target_path, $target_server, $target_scheme)= @_ ;
    local($name, $value, $type, @n,
	  $cname, $path, $domain, $cvalue, $secure, @matches, %pathlen,
	  $realm, $server, @auth) ;

    foreach ( split(/\s*;\s*/, $cookie) ) {
	($name, $value)= split(/=/, $_, 2) ;     # $value may contain "="
	$name= &cookie_decode($name) ;
	$value= &cookie_decode($value) ;
	($type, @n)= split(/;/, $name) ;
	if ($type eq 'COOKIE') {
	    ($cname, $path, $domain)= @n ;
	    ($cvalue, $secure)= split(/;/, $value) ;
	    next if $secure && ($target_scheme ne 'https') ;
	    if ($target_server=~ /$domain$/i  && $target_path=~ /^$path/) {
		push(@matches, $cname.'='.$cvalue) ;
		$pathlen{$matches[$#matches]}= length($path) ;
	    }
	} elsif ($type eq 'AUTH') {
	    # format of auth cookie's name is AUTH;$enc_realm;$enc_server
	    ($realm, $server)= @n ;
	    $realm=~  s/%([\da-fA-F]{2})/ pack('C', hex($1)) /ge ;
	    $server=~ s/%([\da-fA-F]{2})/ pack('C', hex($1)) /ge ;
	    push(@auth, $realm, $value)  if  $server eq $target_server ;
	}
    }

    # More specific path mappings (i.e. longer paths) should be sent first.
    # If %pathlen is declared with my(), this generates a warning in Perl 5.
    sub reversepathlength { $pathlen{$b} <=> $pathlen{$a} }
    $cookie= join('; ', sort reversepathlength @matches) ;

    return $cookie, @auth ;
}


# Old notes:
#
# Cookie support:  The trick is how to send a cookie back to the client that
#   it will return for appropriate hosts.  Given that the target URL may be
#   encoded, and the client can't always tell where the target URL is, the
#   only way to do that is to get *all* the cookies from the client and send
#   along the matching ones.  If the client has a lot of cookies through the
#   proxy, this could conceivably be a problem.  Oh well, it works for the
#   limited amount I've tested.
# Here, we transform the cookie from the server into something the client
#   will always send back to us, and embed the real server/path info in the
#   name of the name-value pair, since the cookie is uniquely identified by
#   the domain, path, and name.  Upon return from the client, we split the
#   name back into its original fields.
# One way to get around *some* of the all-cookies-all-the-time problem,
#   *sometimes*, may be possible to program with the following approach:
#   First, the target URL must be "encoded" (in proxy_encode()) in a way
#   that it resembles a path.  For example, the default "://" --> "/"
#   encoding does this.  Then, let the cookies go back to the client with
#   the target paths still intact.  This would only work when the cookie
#   domain is the default, i.e. the source host.  Check other possibilities
#   carefully, too, but I think you could get it to work somehow.
# Question-- is the port supposed to be used in the domain field?
#   Everything here assumes not, which is conceivably a security risk.

# Transform one cookie into something the client will send back through
#   the script, but still has all the needed info.  Returns a transformed
#   cookie, or undef if the cookie is invalid (e.g. comes from
#   the wrong host).
# A cookie is uniquely identified by the domain, path, and name, so this
#   transformation embeds the path and domain info into the "name".
# This doesn't handle multiple comma-separated cookies-- possible, but
#   which seems a slight contradiction between the HTTP spec (section 4.2
#   of both HTTP 1.0 and 1.1 specs) and the cookie spec at
#   http://www.netscape.com/newsref/std/cookie_spec.html.
sub cookie_to_client {
    local($cookie, $source_path, $source_server)= @_ ;
    local($name, $value, $expires_clause, $path, $domain, $secure_clause) ;
    local($new_name, $new_value, $new_cookie) ;

    # Start last four regexes with ";" to avoid extracting from name=value.
    ($name, $value)=   $cookie=~ /^\s*([^=;,\s]*)=?([^;,\s]*)/ ;
    ($expires_clause)= $cookie=~ /;\s*(expires\s*=[^;]*)/i ;
    ($path)=     $cookie=~ /;\s*path\s*=\s*([^;,\s]*)/i ;  # clash w/ ;-params?
    ($domain)=         $cookie=~ /;\s*domain\s*=\s*([^;,\s]*)/i ;
    ($secure_clause)=  $cookie=~ /;\s*(secure\b)/i ;

    # Path defaults to path of URL that sent the cookie
    $path= $source_path if $path eq '' ;

    # Domain must be checked for validity: defaults to the server that sent
    #   the cookie; otherwise, must match end of that server name, and must
    #   contain at least two dots if in one of these seven top-level domains,
    #   three dots otherwise.
    # As it turns out, hostnames ending in extraneous dots, like
    #   "slashdot.org.." resolve to the hostname without the dots.  So we
    #   need to guard against malicious cookie servers getting around the
    #   two/three-dot requirement this way.
    if ($domain eq '') {
	$domain= $source_server ;
    } else {
	$domain=~ s/\.*$//g ;  # removes trailing dots!
	$domain=~ tr/././s ;   # ... and double dots for good measure.
	return(undef) if $source_server!~ /$domain$/ ;
	return(undef) unless
	    ( ( ($domain=~ tr/././) >= 3 ) ||
	      ( ($domain=~ tr/././) >= 2 &&
		$domain=~ /\.(com|edu|net|org|gov|mil|int)$/i )
	    ) ;
    }


    # This is hereby the transformed format: name is COOKIE;$name;$path;$domain
    #   (the three values won't already have semicolons in them); value is
    #   $value;$secure_clause .  Both name and value are then cookie_encode()'d.
    #   The name contains everything that identifies the cookie, and the value
    #   contains all info we might care about later.
    # "secure" is rare and not even relevant to most installations, given its
    #   nature and the fact that SSL isn't normally supported here.  But it
    #   could *conceivably* be used in future versions or if a user modifies
    #   this script or if a server uses cookies in a strange way, and to not
    #   support it could *conceivably* be a security hole.  So, it is
    #   supported.
    $new_name= &cookie_encode(join(';', 'COOKIE', $name, $path, $domain)) ;

    # New value is "$value;$secure_clause", then cookie_encode()'d.
    $new_value= &cookie_encode(join(';', $value, $secure_clause)) ;


    # Create the new cookie from its components, removing the empty ones.
    # The new domain is this proxy server, which is the default if it is not
    #   specified.
    $new_cookie= join('; ', grep(length,
				 $new_name . '=' . $new_value,
				 $expires_clause,
				 'path=' . $ENV{'SCRIPT_NAME'} . '/',
				 $secure_clause
		     )) ;
    return $new_cookie ;

}



# Returns a cookie that contains authentication information for a particular
#   realm and server.  The format of the cookie is:  The name is
#   AUTH;$URL_encoded_realm;$URL_encoded_server, and the value is the
#   base64-encoded "$username:$password" needed for the Authorization: header.
#   On top of that, both name and value are cookie_encode()'d.
# Leave the "expires" clause out, which means the cookie lasts as long as
#   the session, which is what we want.
sub auth_cookie {
    local($username, $password, $realm, $server)= @_ ;

    $realm=~ s/(\W)/ '%' . sprintf('%02x',ord($1)) /ge ;
    $server=~ s/(\W)/ '%' . sprintf('%02x',ord($1)) /ge ;

    return join('', &cookie_encode(join(';', 'AUTH', $realm, $server)), '=',
		    &cookie_encode(&base64($username.':'.$password)),
		    '; path=' . $ENV{'SCRIPT_NAME'} . '/' ) ;

}



# Generates a set of cookies that will delete the cookies contained in the
#   given cookie strings (e.g. from HTTP_COOKIE).  This is done by giving
#   each cookie an expiration time in the past, and setting their values
#   to "" for good measure.
# The input @cookies can each be a list of cookies separated by ";" .  The
#   cookies themselves can be either "name=value" or just "name".
# The return value is one long string of multiple "Set-Cookie:" headers.
# Slight quirk in Netscape and other browsers-- if cookie expiration is
#   set to the epoch time of "01-Jan-1970 00:00:00 GMT" (meaning second #0),
#   the cookie is treated as a session cookie instead of a deleted cookie.
#   Using second #1, i.e. "01-Jan-1970 00:00:01 GMT", causes the cookies to
#   be correctly deleted.

sub cookie_clearer {
    local(@cookies)= @_ ;   # may be one or more lists of cookies
    local($ret, $cname) ;

    foreach (@cookies) {
	foreach $cname ( split(/\s*;\s*/) ) {
	    $cname=~ s/=.*// ;      # change "name=value" to "name"
	    $ret.= "Set-Cookie: $cname=; expires=Thu, 01-Jan-1970 00:00:01 GMT; "
		 . "path=$ENV{'SCRIPT_NAME'}/\015\012" ;
	}
    }
    return $ret ;
}


#--------------------------------------------------------------------------
#    Utility routines
#--------------------------------------------------------------------------

# The following subroutine looks messy, but can be used to open any
#   TCP/IP socket in any Perl program.  Except for the &HTMLdie() part.
# Note that $AF_INET and $SOCK_STREAM are set near the start of the program,
#   near the "use Socket" statement.
sub newsocketto {
    local(*S, $host, $port)= @_ ;
    local($hostaddr, $remotehost) ;

    # If $host is long integer like 3467251275, break it into a.b.c.d format.
    # This is for big-endian; reverse the list for little-endian.
    $host= join('.', $host>>24 & 255, $host>>16 & 255, $host>>8 & 255,
		     $host & 255)
	if $host=~ /^\d+$/ ;

    # Create the remote host data structure, from host name or IP address.
    # Note that gethostbyname() handles both alpha names and IP addresses.
    $hostaddr= (gethostbyname($host))[4]
	|| &HTMLdie("Couldn't find address for $host: $!") ;
    $remotehost= pack('S n a4 x8', $AF_INET, $port, $hostaddr) ;

    # If the target IP address is a banned host or network, die appropriately.
    # This assumes that IP address structs have the most significant byte first.
    # This is a quick addition that will be fleshed out in a later version.
    for (@BANNED_NETWORK_ADDRS) {
	&banned_server_die if $hostaddr=~ /^$_/ ;
    }

    # Create the socket and connect to the remote host
    socket(S, $AF_INET, $SOCK_STREAM, (getprotobyname('tcp'))[2])
	|| &HTMLdie("Couldn't create socket: $!") ;
    connect(S, $remotehost)
	|| &HTMLdie("Couldn't connect to $host:$port: $!") ;
    select((select(S), $|=1)[0]) ;      # unbuffer the socket
}


# Read a specific number of bytes from a socket, looping if necessary.
# Returns all bytes read (possibly less than $length), or undef on error.
# Typeglobbing *STDIN into *S doesn't seem to work with mod_perl 1.21,
#   so pass socket handle as a string instead (e.g. 'STDIN').
# Using *S, the read() below immediately fails under mod_perl.
sub read_socket {
#    local(*S, $length)= @_ ;
    local($S, $length)= @_ ;
    local($ret, $numread, $thisread) ;

    #$numread= 0 ;
    while (    ($numread<$length)
#	    && ($thisread= read(S, $ret, $length-$numread, $numread) ) )
	    && ($thisread= read($S, $ret, $length-$numread, $numread) ) )
    {
	$numread+= $thisread ;
    }

    return undef unless defined($thisread) ;
    return $ret ;
}



# This is a minimal routine that reads URL-encoded variables from a string,
#   presumably from something like QUERY_STRING.  If no string is passed,
#   it will read from either QUERY_STRING or STDIN, depending on
#   REQUEST_METHOD.  STDIN can't be read more than once for POST requests.
# It returns a hash.  In the event of multiple variables with the same name,
#   it concatenates the values into one hash element, delimiting with "\0".
# Returns undef on error.
sub getformvars {
    local($in)= @_ ;
    local(%in, $name, $value) ;

    # If no string is passed, read it from the usual channels.
    unless (defined($in)) {
	if ( ($ENV{'REQUEST_METHOD'} eq 'GET') ||
	     ($ENV{'REQUEST_METHOD'} eq 'HEAD') ) {
	    $in= $ENV{'QUERY_STRING'} ;
	} elsif ($ENV{'REQUEST_METHOD'} eq 'POST') {
	    return undef unless
	      ($ENV{'CONTENT_TYPE'}=~ m#^application/x-www-form-urlencoded$#i);
	    return undef unless defined($ENV{'CONTENT_LENGTH'}) ;
	    $in= &read_socket('STDIN', $ENV{'CONTENT_LENGTH'}) ;
	    # should we return undef if not all bytes were read?
	} else {
	    return undef ;   # unsupported REQUEST_METHOD
	}
    }

    foreach (split('&', $in)) {
	s/\+/ /g ;
	($name, $value)= split('=', $_, 2) ;
	$name=~ s/%([\da-fA-F]{2})/ pack('C', hex($1)) /ge ;
	$value=~ s/%([\da-fA-F]{2})/ pack('C', hex($1)) /ge ;
	$in{$name}.= "\0" if defined($in{$name}) ;  # concatenate multiple vars
	$in{$name}.= $value ;
    }
    return %in ;
}



# For a given timestamp, returns a date in one of the following two forms,
#   depending on the setting of $use_dash::
#     "Wdy, DD Mon YYYY HH:MM:SS GMT"
#     "Wdy, DD-Mon-YYYY HH:MM:SS GMT"
# The first form is used in HTTP dates, and the second in Netscape's cookie
#   spec (although cookies sometimes use the first form, which seems to be
#   handled OK by most recipients).
# The first form is basically the date format in RFC 822 as updated in RFC
#   1123, except GMT is always used here.
# jsm-- should make this more efficient-- store arrays, etc.
sub rfc1123date {
    local($time, $use_dash)= @_ ;
    local($s) =  $use_dash  ? '-'  : ' ' ;
    local(@t)= gmtime($time) ;
    local($wday)= ('Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')[$t[6]] ;
    local($mon)=  ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')[$t[4]] ;

    return sprintf("%s, %02d$s%s$s%04d %02d:%02d:%02d GMT",
		   $wday, $t[3], $mon, $t[5]+1900, $t[2], $t[1], $t[0] ) ;
}



# Escape any &"<> chars to &xxx; and return resulting string.
sub HTMLescape {
    local($s)= @_ ;
    $s=~ s/&/&amp;/g ;      # must be before all others
    $s=~ s/"/&quot;/g ;
    $s=~ s/</&lt;/g ;
    $s=~ s/>/&gt;/g ;
    return $s ;
}


# Unescape any &xxx; codes back to &"<> and return resulting string.
# Simplified version here; only includes &"<>.
# Some people accidentally leave off final ";", and some browsers support that
#   if the word ends there, so make the final ";" optional.
sub HTMLunescape {
    local($s)= @_ ;
    $s=~ s/&quot\b;?/"/g ;
    $s=~ s/&lt\b;?/</g ;
    $s=~ s/&gt\b;?/>/g ;
    $s=~ s/&amp\b;?/&/g ;      # must be after all others
    return $s ;
}



# Base64-encode a string, except not inserting line breaks.
sub base64 {
    local($s)= @_ ;
    local($ret, $p, @c, $t) ;

    # Base64 padding is done with "=", but that's in the first 64 characters.
    #   So, use "@" as a placeholder for it until the tr/// statement.

    # For each 3 bytes, build a 24-bit integer and split it into 6-bit chunks.
    # Insert one or two padding chars if final substring is less than 3 bytes.
    while ($p<length($s)) {
	@c= unpack('C3', substr($s,$p,3)) ;
	$p+= 3 ;
	$t= ($c[0]<<16) + ($c[1]<<8) + $c[2] ;     # total 24-bit integer
	$ret.= pack('C4',     $t>>18,
			     ($t>>12)%64,
		    (@c>1) ? ($t>>6) %64 : 64,
		    (@c>2) ?  $t     %64 : 64 ) ;  # "@" is chr(64)
    }

    # Translate from bottom 64 chars into base64 chars, plus @ to = conversion.
    $ret=~ tr#\x00-\x3f@#A-Za-z0-9+/=# ;

    return $ret ;
}


# Opposite of base64() .
sub unbase64 {
    local($s)= @_ ;
    local($ret, $p, @c, $t, $pad) ;

    $pad++ if $s=~ /=$/ ;
    $pad++ if $s=~ /==$/ ;

    $s=~ tr#A-Za-z0-9+/##cd ;          # remove non-allowed characters
    $s=~ tr#A-Za-z0-9+/#\x00-\x3f# ;   # for speed, translate to \x00-\x3f

    # For each 4 chars, build a 24-bit integer and split it into 8-bit bytes.
    # Remove one or two chars from result if input had padding chars.
    while ($p<length($s)) {
	@c= unpack('C4', substr($s,$p,4)) ;
	$p+= 4 ;
	$t= ($c[0]<<18) + ($c[1]<<12) + ($c[2]<<6) + $c[3] ;
	$ret.= pack('C3', $t>>16, ($t>>8) % 256, $t % 256 ) ;
    }
    chop($ret) if $pad>=1 ;
    chop($ret) if $pad>=2 ;

    return $ret ;
}



# Read an entire file into a string and return it; return undef on error.
# Does NOT check for any security holes in $fname!
sub readfile {
    local($fname)= @_ ;
    local(*F, $/, $ret) ;

    open(F, "<$fname") || return undef ;
    undef $/ ;
    $ret= <F> ;
    close(F) ;

    return $ret ;
}

#--------------------------------------------------------------------------
#    Output routines
#--------------------------------------------------------------------------


# Print the footer common to most error responses
sub footer {
    local($rightlink)= $NO_LINK_TO_START
	? ''
	: qq(<a href="$script_url"><i>Restart</i></a>) ;

    print <<EOF ;
<p>
<hr>
<table width="100%"><tr>
<td align=left>
<a href="http://www.jmarshall.com/tools/cgiproxy/"><i>CGIProxy 1.5.1</i></a>
</td>
<td align=right>
$rightlink
</td>
</tr></table>
<p>
</body>
</html>
EOF
}



# Return the contents of the top frame, i.e. the one with whatever insertion
#   we have-- the entry form, and/or the inserted HTML or file.
# NOTE that $ANONYMIZE_INSERTION is not supported when using this top frame!
#   Thus, the inserted HTML or file is simply not included, rather than
#   inserting it without anonymizing it.  This will likely not be a problem
#   in a future release.
sub return_top_frame {
    local($enc_URL)= @_ ;
    local($body) ;
    local($date_header)= &rfc1123date(time, 0) ;

    $body= &mini_start_form(&proxy_decode($enc_URL),1) if $INSERT_ENTRY_FORM ;

    if ( !$ANONYMIZE_INSERTION &&
	 (($INSERT_HTML ne '') || ($INSERT_FILE ne '')) ) {
	$INSERT_HTML= &readfile($INSERT_FILE) if $INSERT_HTML eq '' ;
	if ($FORM_AFTER_INSERTION) {
	    $body= $INSERT_HTML . $body ;
	} else {
	    $body.= $INSERT_HTML ;
	}
    }

    print <<EOF ;
$HTTP_1_X 200 OK
Date: $date_header
Content-Type: text/html

<html><body>
$body
</body></html>
EOF

    goto EXIT ;
}


# Return a frame document that puts the insertion in the top frame and the
#   actual page in the lower frame.  Both of these will have the is_in_frame
#   flag set.
# MUST be careful to set $is_in_frame flag!  Else will recurse!
sub return_frame_doc {
    local($enc_URL, $title)= @_ ;
    local($qs_URL, $top_URL, $page_URL) ;
    local($date_header)= &rfc1123date(time, 0) ;

    ($qs_URL= $enc_URL) =~ s/([^\w.-])/ '%' . sprintf('%02x',ord($1)) /ge ;
    $top_URL= &HTMLescape($this_url_inframe
			. &proxy_encode('x-proxy://frames/topframe?URL=' . $qs_URL) ) ;
    $page_URL= &HTMLescape($this_url_inframe . $enc_URL) ;

    print <<EOF ;
$HTTP_1_X 200 OK
Date: $date_header
Content-Type: text/html

<html>
<head>
<title>$title</title>
</head>
<frameset rows="$INSERTION_FRAME_HEIGHT,*">
    <frame src="$top_URL">
    <frame src="$page_URL">
</frameset>
</html>
EOF

    goto EXIT ;
}



# Display cookies to the user and let user selectively delete them.
# No expiration date is displayed because to make that available would
#   require embedding it in every cookie.
sub manage_cookies {
    local($qs)= @_ ;
    local($return_url, $action, $clear_cookies_url, $cookie_rows, $auth_rows,
	  $from_tag) ;
    local($name, $value, $type, @n, $delete_cb,
	  $cname, $path, $domain, $cvalue, $secure,
	  $realm, $server, $username) ;

    local($date_header)= &rfc1123date(time, 0) ;

    local(%in)= &getformvars($qs) ;

    # $in{'from'} is already proxy_encoded
    $return_url= &HTMLescape( $this_url . $in{'from'} ) ;
    $action=     &HTMLescape( $this_url . &proxy_encode('x-proxy://cookies/update') ) ;

    # Create "clear cookies" link, preserving any query string.
    $clear_cookies_url= $this_url . &proxy_encode('x-proxy://cookies/clear') ;
    $clear_cookies_url.= '?' . $qs    if $qs ne '' ;
    $clear_cookies_url= &HTMLescape($clear_cookies_url) ;   # probably never necessary

    # Include from-URL in form if it's available.
    $from_tag= '<input type=hidden name="from" value="' . &HTMLescape($in{'from'}) . '">'
	if $in{'from'} ne '';

    # First, create $cookie_rows and $auth_rows from $ENV{'HTTP_COOKIE'}.
    # Note that the "delete" checkboxes use the encoded name as their identifier.
    foreach ( split(/\s*;\s*/, $ENV{'HTTP_COOKIE'}) ) {
	($name, $value)= split(/=/, $_, 2) ;  # $value may contain "="
	$delete_cb= '<input type=checkbox name="delete" value="'
		  . &base64($name) . '">' ;
	$name= &cookie_decode($name) ;
	$value= &cookie_decode($value) ;
	($type, @n)= split(/;/, $name) ;
	if ($type eq 'COOKIE') {
	    ($cname, $path, $domain)= @n ;
	    ($cvalue, $secure)= split(/;/, $value) ;

	    $cookie_rows.= sprintf("<tr align=center><td>%s</td>\n<td>%s</td>\n<td>%s</td>\n<td>%s</td>\n<td>%s</td>\n<td>%s</td></tr>\n",
				   $delete_cb,
				   &HTMLescape($domain),
				   &HTMLescape($path),
				   &HTMLescape($cname),
				   &HTMLescape($cvalue),
				   $secure ? 'Yes' : 'No',
				  ) ;

	} elsif ($type eq 'AUTH') {
	    # format of auth cookie's name is AUTH;$enc_realm;$enc_server
	    ($realm, $server)= @n ;
	    $realm=~  s/%([\da-fA-F]{2})/ pack('C', hex($1)) /ge ;
	    $server=~ s/%([\da-fA-F]{2})/ pack('C', hex($1)) /ge ;
	    ($username)= split(/:/, &unbase64($value)) ;

	    $auth_rows.= sprintf("<tr align=center><td>%s</td>\n<td>%s</td>\n<td>%s</td>\n<td>%s</td></tr>\n",
				 $delete_cb,
				 &HTMLescape($server),
				 &HTMLescape($username),
				 &HTMLescape($realm),
				) ;

	}
    }


    # If either $cookie_rows or $auth_rows is empty, set appropriate messages.
    $cookie_rows= "<tr><td colspan=6 align=center>&nbsp;<br><b><font face=Verdana size=2>You are not currently sending any cookies through this proxy.</font></b><br>&nbsp;</td></tr>\n"
	if $cookie_rows eq '' ;

    $auth_rows= "<tr><td colspan=4 align=center>&nbsp;<br><b><font face=Verdana size=2>You are not currently authenticated to any sites through this proxy.</font></b><br>&nbsp;</td></tr>\n"
	if $auth_rows eq '' ;


    print <<EOF ;
$HTTP_1_X 200 OK
Date: $date_header
Content-type: text/html

<html>
<head>
<title>CGIProxy Cookie Management</title>
</head>
<body>
<h3><a href="$return_url">Return to browsing</a></h3>
<h1>Here are the cookies you're using through CGIProxy:</h1>

<form action="$action" method=post>
$from_tag

<table bgcolor="#ccffff" border=1>
<tr><th>Delete this cookie?</th>
    <th>For server names ending in:</th>
    <th>... and a path starting with:</th>
    <th>Cookie name</th>
    <th>Value</th>
    <th>Secure?</th>
</tr>
$cookie_rows
</table>

<h3>Authentication cookies:</h3>
<table bgcolor="#ccffcc" border=1>
<tr><th>Delete this cookie?</th>
    <th>Server</th>
    <th>User</th>
    <th>Realm</th>
</tr>
$auth_rows
</table>

<p><font color=red>
<input type=submit value="Delete selected cookies">
</font>
</form>

<h3><a href="$clear_cookies_url">Delete all cookies</a></h3>
EOF

    &footer ;
    goto EXIT ;
}



# Present the initial entry form
sub show_start_form {
    local($method, $action, $flags, $cookies_url) ;
    local($date_header)= &rfc1123date(time, 0) ;

    $method= $USE_POST_ON_START   ? 'post'   : 'get' ;

    $action=      &HTMLescape( $this_url . &proxy_encode('x-proxy://start') ) ;
    $cookies_url= &HTMLescape( $this_url . &proxy_encode('x-proxy://cookies/manage') ) ;

    # Include checkboxes if user config is allowed.
    if ($ALLOW_USER_CONFIG) {
	local($rc_on)= $REMOVE_COOKIES     ? ' checked'  : '' ;
	local($rs_on)= $REMOVE_SCRIPTS     ? ' checked'  : '' ;
	local($fa_on)= $FILTER_ADS         ? ' checked'  : '' ;
	local($br_on)= $HIDE_REFERER       ? ' checked'  : '' ;
	local($if_on)= $INSERT_ENTRY_FORM  ? ' checked'  : '' ;
	$flags= <<EOF ;
<br><input type=checkbox name="rc"$rc_on> Remove all cookies (except certain proxy cookies)
<br><input type=checkbox name="rs"$rs_on> Remove all scripts (recommended for anonymity)
<br><input type=checkbox name="fa"$fa_on> Remove ads
<br><input type=checkbox name="br"$br_on> Hide referrer information
<br><input type=checkbox name="if"$if_on> Show URL entry form
EOF
    }

    print <<EOF ;
$HTTP_1_X 200 OK
Date: $date_header
Content-type: text/html

<html>
<head>
<title>Start Using CGIProxy</title>
</head>
<body>
<h1>CGIProxy</h1>
<p>Start browsing through this CGI-based proxy by entering a URL below.
Only HTTP and FTP URLs are supported.  Not all functions will work
(e.g. some JavaScript), but most pages will be fine.

<form action="$action" method=$method>
<input name="URL" size=50>
$flags
<p><input type=submit value="   Begin browsing   ">
</form>

<h3><a href="$cookies_url">Manage cookies</a></h3>
EOF

    &footer ;
    goto EXIT ;
}



# Returns a mini version of the start form, as a string.  It requires
#   $this_url and $URL to be already set.
# To support this correctly in a frame, point it to target="_top" and use
#   $this_url_noframe in the action.
# Put the cookie management in the full window, and when the user "returns to
#   browsing" the frame flag will cause the frames to reload correctly.
sub mini_start_form {
    local($URL, $in_top_frame)= @_ ;
    local($method, $action, $flags, $hr, $cookies_url, $from_param, $safe_URL) ;

    $method= $USE_POST_ON_START   ? 'post'   : 'get' ;
    $action= &HTMLescape( $this_url_noframe . &proxy_encode('x-proxy://start') ) ;
    $safe_URL= &HTMLescape($URL) ;

    # In "manage cookies" link, provide a way to return to page user came from.
    # Exclude certain characters from URL-encoding, to make URL more readable
    #   in the event it's not obscured.  Unfortunately, ":" and "/" are
    #   reserved in query component (RFC 2396), so we can't exclude them.
    # Don't confusing "URL-encoding" with the "encoding of the URL"!  The
    #   latter uses proxy_encode().  Unfortunate language.
    $from_param= &proxy_encode($URL) ;   # don't send unencoded URL
    $from_param=~ s/([^\w.-])/ '%' . sprintf('%02x',ord($1)) /ge ;
    $cookies_url= $this_url_noframe . &proxy_encode('x-proxy://cookies/manage')
		. '?from=' . $from_param ;
    $cookies_url= &HTMLescape($cookies_url) ;

    # Include checkboxes if user config is allowed.
    if ($ALLOW_USER_CONFIG) {
	local($rc_on)= $REMOVE_COOKIES     ? ' checked'  : '' ;
	local($rs_on)= $REMOVE_SCRIPTS     ? ' checked'  : '' ;
	local($fa_on)= $FILTER_ADS         ? ' checked'  : '' ;
	local($br_on)= $HIDE_REFERER       ? ' checked'  : '' ;
	local($if_on)= $INSERT_ENTRY_FORM  ? ' checked'  : '' ;
	$flags= <<EOF ;
<br><a href="$cookies_url" target="_top">[Manage cookies]</a>&nbsp;&nbsp;
<font size="-1"><input type=checkbox name="rc"$rc_on> No cookies
&nbsp;&nbsp;<input type=checkbox name="rs"$rs_on> No scripts
&nbsp;&nbsp;<input type=checkbox name="fa"$fa_on> No ads
&nbsp;&nbsp;<input type=checkbox name="br"$br_on> No referrer
&nbsp;&nbsp;<input type=checkbox name="if"$if_on> Show this form
</font>
EOF
    }


    # Alter various HTML depending on whether we're in the top frame or not.
    $hr=  $in_top_frame  ? ''  : '<hr>' ;


    return <<EOF ;
<form action="$action" method=$method target="_top">
<center>
Location&nbsp;via&nbsp;proxy:<input name="URL" size=66 value="$safe_URL"><input type=submit value="Go">
$flags
$hr
</center>
</form>
EOF
}



# Present the user with a special form that lets them enter authentication.
# The target URL is proxy_encoded in the form, for obscurity.
# Uses POST, because a GET request would show auth info in a logged URL.
sub get_auth_from_user {
    local($server, $realm, $URL, $tried)= @_ ;
    local($action, $msg) ;
    local($date_header)= &rfc1123date(time, 0) ;

    $server= &HTMLescape($server) ;
    $realm=  &HTMLescape($realm) ;
    $URL=    &HTMLescape(&proxy_encode($URL)) ;

    $action= &HTMLescape( $this_url . &proxy_encode('x-proxy://auth/make_auth_cookie') ) ;

    $msg= "<h3><font color=red>Authorization failed.  Try again.</font></h3>"
	if $tried ;

    print <<EOF ;
$HTTP_1_X 200 OK
Date: $date_header
Cache-Control: no-cache
Pragma: no-cache
Content-type: text/html

<html>
<head><title>Enter username and password for $realm at $server</title></head>
<body>
<h1>Authorization Required</h1>
$msg

<form action="$action" method=post>
<input type=hidden name="s" value="$server">
<input type=hidden name="r" value="$realm">
<input type=hidden name="l" value="$URL">

<table border=1 cellpadding=5>
<tr><th bgcolor="#ff6666">
    Enter username and password for <nobr>$realm</nobr> at $server:</th></tr>
<tr><td bgcolor="#b0b0b0">
    <table cellpadding=0 cellspacing=0>
    <tr><td>Username:</td><td><input name="u" size=20></td>
	<td>&nbsp;&nbsp;&nbsp;<input type=submit value="OK"></tr>
    <tr><td>Password:</td><td><input type=password name="p" size=20></td></tr>
    </table>
</table>
</form>
<p>This requires cookie support turned on in your browser.
<p><i><b>Note:</b> Anytime you use a proxy, you're trusting the owner of that
proxy with all information you enter, including your name and password here.
This is true for <b>any</b> proxy, not just this one.
EOF

    &footer ;
    goto EXIT ;
}



# Return "403 Forbidden" message if the target server is forbidden.
sub banned_server_die {
    local($date_header)= &rfc1123date(time, 0) ;

    print <<EOF ;
$HTTP_1_X 403 Forbidden
Date: $date_header
Content-type: text/html

<html>
<head><title>The proxy can't access that server, sorry.</title></head>
<body>
<h1>The proxy can't access that server, sorry.</h1>
<p>The owner of this proxy has restricted which servers it can access,
presumably for security or bandwidth reasons.  The server you just tried
to access is not on the list of allowed servers.
EOF

    &footer ;
    goto EXIT ;
}



# If so configured, disallow browsing back through this same script.
sub loop_disallowed_die {
    local($URL)= @_ ;
    local($date_header)= &rfc1123date(time, 0) ;

    print <<EOF ;
$HTTP_1_X 403 Forbidden
Date: $date_header
Content-type: text/html

<html>
<head><title>Proxy cannot loop back through itself</title></head>
<body>
<h1>Proxy cannot loop back through itself</h1>
<p>The URL you tried to access would cause this proxy to access itself,
which is redundant and probably a waste of resources.  The owner of this
proxy has configured it to disallow such looping.
<p>Rather than telling the proxy to access the proxy to access the desired
resource, try telling the proxy to access the resource directly.  The link
below <i>may</i> do this.
<blockquote><tt><a href="$URL">$URL</a></tt></blockquote>
EOF

    &footer ;
    goto EXIT ;
}



# Alert the user to an unsupported URL, with this intermediate page.
sub unsupported_warning {
    local($URL)= @_ ;
    local($date_header)= &rfc1123date(time, 0) ;

    print <<EOF ;
$HTTP_1_X 200 OK
Date: $date_header
Content-type: text/html

<html>
<head><title>WARNING: Entering non-anonymous area!</title></head>
<body>
<h1>WARNING: Entering non-anonymous area!</h1>
<h3>This proxy only supports HTTP and FTP.  Any browsing to another URL will
be directly from your browser, and no longer anonymous.</h3>
<h3>Follow the link below to continue to the URL, non-anonymously.</h3>
<blockquote><tt><a href="$URL">$URL</a></tt></blockquote>
EOF

    &footer ;
    goto EXIT ;
}



# Return "406 Not Acceptable" response for script content-type.
sub script_content_die {
    local($date_header)= &rfc1123date(time, 0) ;

    print <<EOF ;
$HTTP_1_X 406 Not Acceptable
Date: $date_header
Content-type: text/html

<html>
<head><title>Script content blocked</title></head>
<body>
<h1>Script content blocked</h1>
<p>The resource you requested (or were redirected to without your knowledge)
is apparently an executable script.  Such resources have been blocked by this
proxy, presumably for your own protection.
<p>Even if you're sure you want the script, you can't get it through this
proxy the way it's configured.  If permitted, try browsing through this proxy
without removing scripts.  Otherwise, you'll need to reconfigure the proxy or
find another way to get the resource.
EOF

    &footer ;
    goto EXIT ;
}



# Return "403 Forbidden" message if images are forbidden.
sub non_text_die {
    local($date_header)= &rfc1123date(time, 0) ;

    print <<EOF ;
$HTTP_1_X 403 Forbidden
Date: $date_header
Content-type: text/html

<html>
<head><title>Proxy cannot forward non-text files</title></head>
<body>
<h1>Proxy cannot forward non-text files</h1>
<p>Due to bandwidth limitations, the owner of this particular proxy is
forwarding only text files.  For best results, turn off automatic image
loading if your browser lets you.
<p>If you need access to images or other binary data, route your browser
through another proxy (or install one yourself--
<a href="http://www.jmarshall.com/tools/cgiproxy/">it's easy</a>).
EOF

    &footer ;
    goto EXIT ;
}



# When an image should be blanked, returns either a transparent 1x1 GIF or
#   a 406 result ("Not Acceptable").  Yes, that's an inlined 43-byte GIF.
sub skip_image {
    local($date_header)= &rfc1123date(time, 0) ;

    if ($RETURN_EMPTY_GIF) {
	print <<EOF ;
$HTTP_1_X 200 OK
Date: $date_header
Content-Type: image/gif
Content-Length: 43

GIF89a\x01\0\x01\0\x80\0\0\0\0\0\xff\xff\xff\x21\xf9\x04\x01\0\0\0\0\x2c\0\0\0\0\x01\0\x01\0\x40\x02\x02\x44\x01\0\x3b
EOF
    } else {
	print "$HTTP_1_X 406 Not Acceptable\015\012Date: $date_header\015\012\015\012" ;
    }

    goto EXIT ;
}



# Die, outputting HTML error page, with optional response code and title.
sub HTMLdie {
    local($msg, $title, $status)= @_ ;
    $title= 'CGIProxy Error' if $title eq '' ;
    $status= '200 OK' if $status eq '' ;
    local($date_header)= &rfc1123date(time, 0) ;

    print <<EOF ;
$HTTP_1_X $status
Date: $date_header
Content-Type: text/html

<html>
<head><title>$title</title></head>
<body>
<h1>$title</h1>
<h3>$msg</h3>
EOF

    &footer ;
    goto EXIT ;
}
